/*
 * Spectral.java
 *
 * Created on 22. November 2005, 15:53
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package clusterer;
import java.util.*;
import compr.*;
import Jama.*;

/**
 *
 * @author boehm
 */
public class Spectral {
    DB db;
    int numObj;
    int dim;
    double[][] a ;
    double[] diag ;
    double[][] reduced ;
    double[][] eigenvect ;
    double[] eigenval ;
    double[][] representants ;
    Random r = new Random(2) ;
    double varDist ;
    int dim_sub = 12 ;
    int numCl;




    /** Creates a new instance of Spectral */
    public Spectral(int k, DB db) {
        this.db = db;
        //this.sampleSize = db.store.length/10;
        this.dim = db.store[0].coord.length;
        numObj = db.store.length ;
        numCl = k;

    }
    
    public DataObject[] run(){
        // Compute the affinity matrix
        a = new double [numObj][numObj] ;
        double meanDist = 0.0 ;
        varDist = 0.0 ;
        for (int i=0 ; i<numObj ; i++)
            for (int j=0 ; j<numObj ; j++) {
            a[i][j] = dist(db.store[i].coord, db.store[j].coord) ;
            meanDist += a[i][j] ;
            }
        meanDist/=(numObj*numObj) ;
        for (int i=0 ; i<numObj ; i++)
            for (int j=0 ; j<numObj ; j++)
                varDist += Math.pow(a[i][j]-meanDist, 2.0) ;
        varDist/=(numObj*numObj) ;
        for (int i=0 ; i<numObj ; i++)
            for (int j=0 ; j<numObj ; j++)
                a[i][j] = Math.exp(-Math.pow(a[i][j], 2.0)/varDist/2.0) ;
        System.out.println("Affinity Matrix");
        //(new Matrix(a)).print (4,2) ;
        // Step 2: Normalization
        diag = new double [numObj] ;
        for (int i=0 ; i<numObj ; i++)
            for (int j=0 ; j<numObj ; j++)
                diag[i] += a[i][j] ;
        for (int i=0 ; i<numObj ; i++)
            for (int j=0 ; j<numObj ; j++)
                a[i][j] /= Math.sqrt(diag[i]*diag[j]) ;
        (new Matrix(a)).print(6,4) ;
        // Step 3: Eigenvectors
        EigenvalueDecomposition e = new EigenvalueDecomposition( new Matrix(a) ) ;
        Matrix d = e.getD() ;
        Matrix v = e.getV() ;
        eigenval = new double[dim_sub] ;
        boolean [] used = new boolean [numObj] ;
        reduced = new double [numObj][dim_sub] ;
        for (int i=0 ; i<dim_sub ; i++) {
            int largest = -1 ;
            eigenval[i] = -Double.MAX_VALUE ;
            for (int j=0 ; j<numObj ; j++)
                if (!used[j] && d.get(j,j) > eigenval[i]) {
                largest = j ;
                eigenval[i] = d.get(j,j) ;
                }
            used[largest] = true ;
            for (int j=0 ; j<numObj ; j++) {
                reduced[j][i] = v.get(j,largest) ;
            }
        }
        prt("Eigenvalues:", eigenval) ;
        //d.print(6,4) ;

        //rowVector(a[3]).times(matrix(reduced)).print(8,6) ;

        // Step 4: Normalization
        eigenvect = reduced.clone() ;
        for (int i=0 ; i<eigenvect.length ; i++)
            eigenvect[i]=eigenvect[i].clone() ;
        for (int i=0 ; i<numObj ; i++) {
            double sum = 0 ;
            for (int j=0 ; j<dim_sub ; j++)
                sum += Math.pow(reduced [i][j], 2.0) ;
            sum = Math.sqrt(sum) ;
            for (int j=0 ; j<dim_sub ; j++)
                reduced[i][j] /= sum ;
        }
        // Step 5: Clustering
        // Random r = new Random(3) ;
        representants = new double[numCl][] ;
        for (int i=0 ; i<numCl ; i++)
            representants[i] = reduced [r.nextInt(numObj)].clone() ;
        
        boolean converged = false ;
        while (! converged) {
            converged = true ;
            for (int i=0 ; i<numObj ; i++) {
                int minIndex = assignReduced (reduced[i]) ;
                if (minIndex != db.store[i].clusterID)
                    converged = false ;
                db.store[i].clusterID = minIndex ;
            }
            double[]count = new double[numCl] ;
            representants = new double[numCl][dim_sub] ;
            for (int i=0 ; i<numObj ; i++) {
                count [db.store[i].clusterID]++ ;
                
                for (int j=0 ; j<dim_sub ; j++)
                    representants[db.store[i].clusterID][j] += reduced[i][j] ;
            }
            for (int i=0 ; i<numCl ; i++)
                for (int j=0 ; j<dim_sub ; j++)
                    representants [i][j] /= count[i] ;
        }
        for(int i = 0; i < db.store.length; i++)
            db.store[i].clusterID++;
        return db.store;
    }
    
     public void prt (String s, double[][] a) {
        System.out.println() ;
        System.out.println(s) ;
        (new Matrix (a)).print(6,4) ;
    }
    
    public void prt (String s, double[] a) {
        System.out.println() ;
        System.out.print(s) ;
        (new Matrix (a, 1)).print(6,4) ;
    }

    public Matrix matrix(double[][] m) {
        return new Matrix(m) ;
    }
    
    public Matrix columnVector(double[] v) {
        return new Matrix(v, v.length) ;
    }
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
    double dist(double[]a, double[]b) {
        double result = 0.0 ;
        for (int i=0 ; i<a.length ; i++)
            result += Math.pow(a[i]-b[i], 2.0) ;
        return Math.sqrt(result) ;
    }
    
public double[] transformNewObject(double [] x) {
        double[] reduceda = new double[dim_sub] ;
        double[] aa = new double[numObj] ;
        // Step 1: Compute the affinity vector
        double diaga = 0.0 ;
        for (int i=0 ; i<numObj ; i++) {
            aa[i] = Math.exp(-Math.pow(dist(x, db.store[i].coord), 2.0)/varDist/2.0) ;
            diaga += aa[i] ;
        }
        // Step 2: Normalization
        for (int i=0 ; i<numObj ; i++)
            aa[i] /= Math.sqrt(diag[i]*diaga) ;
            
//        prt("aa[i]=", aa) ;
        // Step 3: Transform according to E^{-1} V' aa
        for (int i=0 ; i<dim_sub ; i++)
            for (int j=0 ; j<numObj ; j++)
                reduceda[i] += aa[j]*eigenvect[j][i]/eigenval[i] ;
        // Step 4: Normalize:
        double sum = 0.0 ;
        for (int i=0 ; i<dim_sub ; i++)
            sum += Math.pow (reduceda[i], 2.0) ;
        sum = Math.sqrt (sum) ;
        for (int i=0 ; i<dim_sub ; i++)
            reduceda[i] /= sum ;
        return reduceda ;
    }
    
    public int assignReduced(double[] x) {
        double minDist = Double.MAX_VALUE ;
        int minIndex = -1 ;
        for (int j=0 ; j<numCl ; j++) {
            double h = dist(representants[j], x) ;
            if (h < minDist) {
                minDist = h ;
                minIndex = j ;
            }
        }
        return minIndex;
    }


    
}
