/*
 * KMCost.java
 *
 * Created on 12. Januar 2006, 13:55
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package clusterer;
import java.util.*;
import Jama.*;
import eval.Clustering;
import eval.SortPack;
import eval.Cluster;
import compr.DataObject;
import visu.VisuSimple;
import visu.VisuKMeans;

/**
 *
 * @author boehm
 */
public class TopDown {
    int k;
    DataObject[] data;
    Vector clusters;
    Vector noiseClusters;
    Cluster[] bestClusters;
    Cluster[] bestNoiseClusters;
    double bestCost;
    int iter;
    int numObj;
    int dim;
    int precision;
    int maxClusterID;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    
    /**
     * Creates a new instance of KMCost
     */
    public TopDown(DataObject[] data, int precision) {
        this.data = data;
        dim = data[0].coord.length;
        this.k = k;
        clusters = new Vector();
        noiseClusters = new Vector();
        numObj = data.length;
        this.precision = precision;
        iter = 0;
        maxClusterID = 0;
        bestCost = Double.MAX_VALUE;
    }
    
    public void run(){
        init();
        runAlg();
        bestCost = costModel();
        setBestClusters();
        boolean improve = true;
        
        int it = 0;
        /*
        split();
        setBestClusters();
        assignPointsLocal(bestClusters, bestNoiseClusters, data);
        runAlg();
         */
        
        
        while(improve){
            split();
            //setBestClusters();
            Cluster[] aktClusters = new Cluster[clusters.size()];
            for(int i = 0; i < aktClusters.length; i++)
                aktClusters[i] = new Cluster((Cluster)clusters.elementAt(i));
            Cluster[] aktNoiseClusters = new Cluster[noiseClusters.size()];
            for(int i = 0; i < aktNoiseClusters.length; i++)
                aktNoiseClusters[i] = new Cluster((Cluster)noiseClusters.elementAt(i));
            assignPointsLocal(aktClusters, aktNoiseClusters, data);
            runAlg();
            double aktCost = costModel();
            if(aktCost < bestCost){
               bestCost = aktCost;
                setBestClusters();
            } else
               improve = false;
            it++;
        }
        
        
        System.out.println("bestClusters " + bestClusters.length);
        
        assignPointsLocal(bestClusters, bestNoiseClusters, data);
        
        
        //Mediane zeichnen
        DataObject[] meansCl = new DataObject[bestClusters.length];
        for(int i = 0; i < bestClusters.length; i++){
            DataObject m = new DataObject(bestClusters[i].mean, 0);
            m.clusterID = bestClusters[i].ID;
            meansCl[i] = m;
        }
        
        VisuKMeans vk = new VisuKMeans(data, meansCl);
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
        
    }
    
    
    public void split(){
        int index = chooseSplitCluster();
        Cluster sp = (Cluster)clusters.elementAt(index);
        splitCluster(sp, (Cluster)noiseClusters.elementAt(index));
        
    }
    
    
    
    //die aktuell zu splittenden Cluster sind in Vektor clusters, globaler Durchlauf
    public void runAlg(){
        for(int i = 0; i < clusters.size(); i++){
            Cluster akt = (Cluster)clusters.elementAt(i);
            Cluster aktN = (Cluster)noiseClusters.elementAt(i);
            robustEstimate(akt);
            aktN.mean = akt.mean;
            seperateNoise(akt);
        }
        iter++;
        boolean clusterChanged = true;
        while(clusterChanged && iter < 100){
            iter++;
            determineNoiseVariances();
            determineClusterVariances();
            clusterChanged = assignPoints();
            for(int i = 0; i < clusters.size(); i++){
                Cluster akt = (Cluster)clusters.elementAt(i);
                Cluster aktN = (Cluster)noiseClusters.elementAt(i);
                robustEstimate(akt);
                aktN.mean = akt.mean;
            }
        }
        double c = 0.0;
        for(int i = 0; i < clusters.size(); i++)
            c+= clusterNoiseCost((Cluster)clusters.elementAt(i),(Cluster)noiseClusters.elementAt(i));
        //System.out.println("cost " + c);
        
        
        
/*
        VisuSimple vc = new VisuSimple(data);
        vc.setSize(600,600);
        vc.setLocation(200,0);
        vc.setVisible(true);
 */
        System.out.println(iter + " iterations.");
    }
    
    //lokaler Durchlauf fr 2 Cluster und 2 NoiseCluster nach dem Split
    private void runLocal(Cluster[] c, Cluster[] nc, DataObject[] obj){
        int locIter = 0;
        DataObject[] allObj = data;
        //euklidisch zuweisen
        for(int i = 0; i < obj.length; i++){
            double dist1 = obj[i].distance(c[0].mean);
            double dist2 = obj[i].distance(c[1].mean);
            if(dist1 < dist2)
                obj[i].clusterID = c[0].ID;
            else
                obj[i].clusterID = c[1].ID;
        }
        data = obj;
        for(int i = 0; i < c.length; i++){
            robustEstimate(c[i]);
            nc[i].mean = c[i].mean;
            seperateNoise(c[i]);
        }
        locIter++;
        boolean clusterChanged = true;
        while(clusterChanged &&locIter < 200){
            locIter++;
            determineNoiseVariance(nc[0]);
            determineClusterVariance(c[0]);
            determineNoiseVariance(nc[1]);
            determineClusterVariance(c[1]);
            clusterChanged = assignPointsLocal(c, nc, obj);
            for(int i = 0; i < c.length; i++){
                robustEstimate(c[i]);
                nc[i].mean = c[i].mean;
            }
        }
        data = allObj;
        System.out.println(locIter + " local iterations");
        
        DataObject[] meansCl = new DataObject[c.length];
        for(int i = 0; i < c.length; i++){
            DataObject m = new DataObject(c[i].mean, 0);
            m.clusterID = c[i].ID;
            meansCl[i] = m;
        }
         /*
        VisuKMeans vk = new VisuKMeans(obj, meansCl);
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
         */
    }
    
    private void init(){
        double[] median = median(data);
        Cluster f = new Cluster(median, 0, dim);
        maxClusterID++;
        Cluster fn = new Cluster(median, 1, dim);
        maxClusterID++;
        clusters.addElement(f);
        noiseClusters.addElement(fn);
    }
    
    
    private void robustEstimate(Cluster c){
        //System.out.println("Cluster " + c.ID);
        DataObject[] obj = getObjects(c);
        //if(obj.length > 5){
        double[] median = median(obj);
        c.mean = median;
        double[][] covarm = covarByMedian(data, median);
        Matrix covar = new Matrix(covarm);
        EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
        Matrix vectors = eig.getV();
        Matrix values = eig.getD();
        //System.out.println("EW");
        //values.print(5, 3);
        //System.out.println("EV");
        //vectors.print(5, 3);
        c.dir = vectors.getArrayCopy();
        double maxEigen = -Double.MAX_VALUE ;
        for (int i=0 ; i<dim ; i++)
            maxEigen = Math.max(maxEigen, values.get(i,i));
        for (int i=0 ; i<dim ; i++)
            values.set(i,i,1/Math.max(values.get(i,i), maxEigen*0.001)) ;
        c.invDir =  (vectors.times(values).times(vectors.transpose())).getArrayCopy() ;
        //}
    }
    
    
    private void seperateNoise(Cluster c){
        DataObject[] obj = getObjects(c);
        //if(obj.length > 5){
        double[] dist = new double[obj.length];
        //System.out.println("Cluster " + c.ID + " " + obj.length);
        Matrix m = new Matrix(c.invDir);
        for(int i = 0; i < obj.length; i++){
            Matrix x = rowVector(obj[i].coord).minus(rowVector(c.mean));
            dist[i]=x.times(m).times(x.transpose()).get(0,0) ;
        }
        SortPack [] distcopy = SortPack.array(dist) ;
        int count = 0;
        for(int i = (int)Math.ceil(0.5 * obj.length); i < obj.length; i++){
            int index = obj[distcopy[i].index].number;
            data[index].clusterID++;
            count++;
        }
        //System.out.println("NoiseCluster " + count);
        //}
    }
    
    
    private void determineNoiseVariances(){
        for(int i = 0; i < noiseClusters.size(); i++){
            double addVar = 0.0;
            Cluster akt = (Cluster)noiseClusters.elementAt(i);
            DataObject[] obj = getObjects(akt);
            for(int j = 0; j < dim; j++)
                addVar += var(obj, akt.mean, j);
            for(int j = 0; j < dim; j++){
                akt.var[j] = addVar;
            }
           // System.out.println("Cluster " + akt.ID + " var " + addVar);
        }
    }
    
    private void determineClusterVariances(){
        for(int i = 0; i < clusters.size(); i++){
            Cluster akt = (Cluster)clusters.elementAt(i);
            DataObject[] obj = getObjects(akt);
            DataObject[] red = new DataObject[obj.length];
            Matrix m = new Matrix(akt.dir);
            for(int l = 0; l < red.length; l++){
                double[] redCoord = new double[dim];
                Matrix x = rowVector(obj[l].coord);
                Matrix aktm = x.times(m);
                for(int j = 0; j < dim; j++){
                    redCoord[j] = aktm.get(0,j);
                }
                DataObject redObj = new DataObject(redCoord, obj[l].number);
                redObj.clusterID = obj[l].clusterID;
                red[l] = redObj;
            }
            double[] redMeanCoord = new double[dim];
            Matrix mx = rowVector(akt.mean);
            Matrix aktm = mx.times(m);
            for(int j = 0; j < dim; j++){
                redMeanCoord[j] = aktm.get(0,j);
            }
            for(int j = 0; j < dim; j++){
                akt.var[j] = var(red, redMeanCoord, j);
                //System.out.println("Cluster " + akt.ID + " coord " + j + " var " + akt.var[j]);
            }
        }
    }
    
    private void determineNoiseVariance(Cluster c){
        double addVar = 0.0;
        DataObject[] obj = getObjects(c);
        for(int j = 0; j < dim; j++)
            addVar += var(obj, c.mean, j);
        for(int j = 0; j < dim; j++){
            c.var[j] = addVar;
        }
        //System.out.println("Cluster " + c.ID + " var " + addVar);
    }
    
    
    private void determineClusterVariance(Cluster c){
        DataObject[] obj = getObjects(c);
        //if(obj.length > 5){
        DataObject[] red = new DataObject[obj.length];
        Matrix m = new Matrix(c.dir);
        for(int l = 0; l < red.length; l++){
            double[] redCoord = new double[dim];
            Matrix x = rowVector(obj[l].coord);
            Matrix aktm = x.times(m);
            for(int j = 0; j < dim; j++){
                redCoord[j] = aktm.get(0,j);
            }
            DataObject redObj = new DataObject(redCoord, obj[l].number);
            redObj.clusterID = obj[l].clusterID;
            red[l] = redObj;
        }
        double[] redMeanCoord = new double[dim];
        Matrix mx = rowVector(c.mean);
        Matrix aktm = mx.times(m);
        for(int j = 0; j < dim; j++){
            redMeanCoord[j] = aktm.get(0,j);
        }
        for(int j = 0; j < dim; j++){
            c.var[j] = var(red, redMeanCoord, j);
            //System.out.println("Cluster " + c.ID + " coord " + j + " var " + c.var[j]);
        }
        //}
    }
    
    
    //nehme Gaussverteilung an, bestimme auch Kosten der Cluster
    private boolean assignPoints(){
        boolean clusterChanged = false;
        for(int i = 0; i < clusters.size(); i++){
            ((Cluster)clusters.elementAt(i)).cost = 0.0;
            ((Cluster)noiseClusters.elementAt(i)).cost = 0.0;
        }
        for(int i = 0; i < numObj; i++){
            double minCost = Double.MAX_VALUE;
            int minIndex = clusters.size()+1;
            int minPos = clusters.size()+1;
            for(int j = 0; j < clusters.size(); j++){
                Cluster a = (Cluster)clusters.elementAt(j);
                double aktCost = cost(data[i], a);
                if(aktCost < minCost){
                    minCost = aktCost;
                    minIndex = a.ID;
                    minPos = j;
                }
            }
            for(int j = 0; j < noiseClusters.size(); j++){
                Cluster b = (Cluster)noiseClusters.elementAt(j);
                double aktCost = cost(data[i], b);
                if(aktCost < minCost){
                    minCost = aktCost;
                    minIndex = b.ID;
                    minPos = j;
                }
            }
            if(minIndex != data[i].clusterID){
                data[i].clusterID = minIndex;
                clusterChanged = true;
                
            }
            if(minIndex % 2 == 0)
                ((Cluster)clusters.elementAt(minPos)).cost += minCost;
            else
                ((Cluster)noiseClusters.elementAt(minPos)).cost += minCost;
        }
        
        for(int i = 0; i < clusters.size(); i++){
            ((Cluster)clusters.elementAt(i)).cost += paramCost((Cluster)clusters.elementAt(i));
            ((Cluster)noiseClusters.elementAt(i)).cost += paramCost((Cluster)noiseClusters.elementAt(i));
        }
        
        return clusterChanged;
    }
    
    
    //nehme Gaussverteilung an, bestimme auch Kosten der Cluster
    private boolean assignPointsLocal(Cluster[] c, Cluster[] nc, DataObject[] obj){
        boolean clusterChanged = false;
        for(int i = 0; i < c.length; i++){
            if(c[i] != null)
                c[i].cost = 0.0;
            if(nc[i] != null)
                nc[i].cost = 0.0;
        }
        for(int i = 0; i < obj.length; i++){
            double minCost = Double.MAX_VALUE;
            int minIndex = c.length+1;
            int minPos = c.length+1;
            for(int j = 0; j < c.length; j++){
                double aktCost = cost(obj[i], c[j]);
                if(aktCost < minCost){
                    minCost = aktCost;
                    minIndex = c[j].ID;
                    minPos = j;
                }
            }
            for(int j = 0; j < nc.length; j++){
                double aktCost = cost(data[i], nc[j]);
                if(aktCost < minCost){
                    minCost = aktCost;
                    minIndex = nc[j].ID;
                    minPos = j;
                }
            }
            if(minIndex != obj[i].clusterID){
                obj[i].clusterID = minIndex;
                clusterChanged = true;
                if(minIndex % 2 == 0)
                    c[minPos].cost += minCost;
                else
                    nc[minPos].cost += minCost;
            }
        }
        for(int i = 0; i < c.length; i++){
            if(c[i] != null)
                c[i].cost += paramCost(c[i]);
            if(nc[i] != null)
                nc[i].cost += paramCost(nc[i]);
        }
        return clusterChanged;
    }
    
    
    private double cost(DataObject o, Cluster c){
        double sumCost = 0.0;
        //cluster
        if(c.dir != null){
            Matrix m = new Matrix(c.dir);
            Matrix x = rowVector(o.coord);
            double[] redMeanCoord = new double[dim];
            Matrix mx = rowVector(c.mean);
            Matrix akt = mx.times(m);
            for(int j = 0; j < dim; j++){
                redMeanCoord[j] = akt.get(0,j);
            }
            Matrix aktc = x.times(m);
            for(int j = 0; j < dim; j++){
                double red = aktc.get(0,j);
                sumCost += gaussCost(red, redMeanCoord[j], c.var[j], precision);
            }
        }
        //noise-cluster
        else{
            for(int j = 0; j < dim; j++)
                sumCost += gaussCost(o.coord[j], c.mean[j], c.var[j], precision);
        }
        return sumCost;
    }
    
    //Index von zu splittendem Cluster
    private int chooseSplitCluster(){
        double maxCost = 0.0;
        int max = clusters.size()+1;
        for(int i = 0; i < clusters.size(); i++){
            Cluster akt = (Cluster)clusters.elementAt(i);
            if(akt.cost/(double)akt.numObj > maxCost);
            maxCost = akt.cost;
            max = i;
        }
        System.out.println("split cluster " + max);
        return max;
    }
    
    //Cluster splitten, neuen Cluster erzeugen mit maxID und noiseCluster, Mittelpunkt ist Punkt, der am teuersten ist.
    private void splitCluster(Cluster c, Cluster n){
        DataObject[] obj = getObjects(c);
        DataObject[] noiseObj = getObjects(n);
        DataObject[] all = new DataObject[obj.length + noiseObj.length];
        for(int i = 0; i < obj.length; i++){
            all[i] = new DataObject(obj[i]);
            all[i].number = i;
        }
        for(int i = obj.length; i < all.length; i++){
            all[i] = new DataObject(noiseObj[i-obj.length]);
            all[i].number = i;
        }
        double[] newRep = mostExpensive(all, c);
        Cluster clustNew = new Cluster(newRep, maxClusterID, dim);
        maxClusterID++;
        Cluster noiseNew = new Cluster(newRep, maxClusterID, dim);
        maxClusterID++;
        Cluster[] clust = new Cluster[2];
        clust[0] = c;
        clust[1] = clustNew;
        Cluster[] noise = new Cluster[2];
        noise[0] = n;
        noise[1] = noiseNew;
        runLocal(clust, noise, all);
        //Kostenabfrage hier
        clusters.add(clustNew);
        noiseClusters.add(noiseNew);
    }
    
    private double[] mostExpensive(DataObject[] obj, Cluster c){
        Random r = new Random();
        double[] exp = new double[dim];
        double maxCost = 0.0;
        for(int i = 0; i < obj.length; i++){
            double aktCost = cost(obj[i], c);
            if (aktCost > maxCost)
                maxCost = aktCost;
            exp = obj[i].coord;
        }
        //return exp;
        return obj[r.nextInt(obj.length)].coord;
    }
    
    private double costModel(){
        double cost = 0.0;
        int numCluster = clusters.size() + noiseClusters.size();
        for(int i = 0; i < clusters.size(); i++){
            Cluster akt = (Cluster)clusters.elementAt(i);
            cost += akt.cost;
        }
        for(int i = 0; i < noiseClusters.size(); i++){
            cost += ((Cluster)noiseClusters.elementAt(i)).cost;
        }
        System.out.println("cost model " + cost);
        return cost;
    }
    
    //Hilfsmethoden/////////////////////////////////////////////////////////////////////////////////////////////
    
    private void setBestClusters(){
        Cluster[] bi = new Cluster[clusters.size()];
        for(int i = 0; i < bi.length; i++)
            bi[i] = new Cluster((Cluster)clusters.elementAt(i));
        Cluster[] bni = new Cluster[noiseClusters.size()];
        for(int i = 0; i < bi.length; i++)
            bni[i] = new Cluster((Cluster)noiseClusters.elementAt(i));
        bestClusters = bi;
        bestNoiseClusters = bni;
        
    }
    
    
    //Parameterkosten
    private double paramCost(Cluster c){
        double cost = 0.0;
        if(c.dir != null)
            return dim*32*2 + dim*dim * 32;
        else
            return dim*32*2;
    }
    
    
    //Kosten von Cluster und NoiseCluster Paar
    private double clusterNoiseCost(Cluster c, Cluster noise){
        double clusterCost = 0.0;
        double noiseCost = 0.0;
        DataObject[] clusterObj = getObjects(c);
        DataObject[] noiseObj = getObjects(noise);
        for(int i = 0; i < clusterObj.length; i++)
            clusterCost+= cost(clusterObj[i], c);
        for(int i = 0; i < noiseObj.length; i++)
            noiseCost+= cost(noiseObj[i], noise);
        /*
        System.out.println("clusterCost " + clusterCost);
        System.out.println("noiseCost " + noiseCost);
        System.out.println("gesamt " + (clusterCost + noiseCost));
         */
        return clusterCost + noiseCost;
    }
    
    
    private double var(DataObject[] o, double[] mean, int numCoord){
        double stdv = 0.0;
        for(int i = 0; i < o.length; i++)
            stdv += (o[i].coord[numCoord] - mean[numCoord]) * (o[i].coord[numCoord] - mean[numCoord]);
        return stdv/o.length;
    }
    
    
    private double gaussCost(double x, double mean, double var, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double va = Math.max(var, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*va)+precision*ln10+Math.pow(x-mean,2)/2/va)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    
    public int countClusterMembers(int clusterID){
        int counter = 0;
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == clusterID)
                counter++;
        /*
        if(clusterID > 0)
            clusters[clusterID/2].numObj = counter;
        else
            clusters[clusterID].numObj = counter;
         */
        return counter;
    }
    
    public DataObject[] getObjects(Cluster c){
        DataObject[] members = new DataObject[countClusterMembers(c.ID)];
        int counter = 0;
        for(int j = 0; j < data.length; j++)
            if(data[j].clusterID == c.ID){
            members[counter] = data[j];
            counter++;
            }
        return members;
    }
    
    public double [] median(DataObject[] members) {
        int numObj= members.length ;
        int dim = members[0].coord.length ;
        double[] result = new double[dim];
        double [] sortarray = new double[numObj] ;
        for (int i=0 ; i<dim ; i++) {
            double median = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sortarray[j]=members[j].coord[i] ;
            Arrays.sort(sortarray) ;
            if (numObj==1)
                result[i]=sortarray[0] ;
            else if (numObj%2 == 0)
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
            else
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
        }
        return result ;
    }
    
    public double[][] covarByMedian(DataObject[] data, double[] mean) {
        int numObj= data.length ;
        int dim = data[0].coord.length ;
        double [] sortarray = new double[numObj] ;
        double[][] cov = new double[dim][dim];
        if (numObj <15){
            for (int i=0 ; i<dim ; i++)
                cov[i][i]=1.0 ;
            return cov ;
        }
        for(int l1 = 0; l1 < dim; l1++) {
            for(int l2 = 0; l2 < dim; l2++) {
                for(int j = 0; j < numObj; j++)
                    sortarray[j] = (data[j].coord[l1]-mean[l1])*(data[j].coord[l2]-mean[l2]) ;
                
                Arrays.sort(sortarray) ;
                if (numObj%2 == 0)
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3
                            ;
            }
            
        }
        double maxdiff = 0.0 ;
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                sum += Math.abs(cov[i][j]) ;
            double diff = 1.1*sum - 2.1*cov[i][i] ;
            if (diff>maxdiff)
                maxdiff = diff ;
        }
        if (maxdiff > 0.0)
            for (int i=0 ; i<dim ; i++)
                cov[i][i] += maxdiff ;
        
        return cov ;
    }
    
    
    private double distance(DataObject d, double[] mean){
        double sum = 0.0;
        for (int i = 0; i < d.coord.length; i++){
            double dist = ((d.coord[i] - mean[i])*(d.coord[i] - mean[i]));
            sum = sum + dist;
        }
        return Math.sqrt(sum);
    }
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
}

