/*
 * TopDownKM.java
 *
 * Created on 18. April 2006, 15:19
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package clusterer;
import clusterer.*;
import java.util.*;
import Jama.*;
import eval.Clustering;
import eval.SortPack;
//import eval.IcaCluster;
import eval.ClusteringEvaluation;
import compr.DataObject;
import compr.DB;
import filterCluster.FilterKM;
import ica.IcaCost;
import visu.VisuSimple;
import visu.VisuInfo;
import visu.VisuKMeans;
import visu.Visu2D;
import io.IO;

/**
 * Split all the objects belonging to one cluster and the corresponding noise cluster.
 * @author claudia
 */
public class TopDownIca {
    DataObject[] data;
    Vector<IcaCluster> clusters; //current clusters
    Vector<IcaCluster> noiseClusters;
    IcaCluster[] bestClusters; //cheapest clusters
    IcaCluster[] bestNoiseClusters;
    int[] aktClusterID; //current cluster IDs;
    DB db;
    double bestCost;
    int iter;
    int numObj;
    int dim;
    int maxClusterID;      //Max ID of the clusters (not noise!)
    int numTry = 0;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    
    
    /** Creates a new instance of TopDownKM */
    public TopDownIca(DataObject[] data) {
        this.data = data;
        db = new DB(data);
        numObj = data.length;
        iter = 0;
        dim = data[0].coord.length;
        maxClusterID = 0;
        bestCost = 0.0;
        clusters = new Vector<IcaCluster>();
        noiseClusters = new Vector<IcaCluster>();
    }
    
    
    
    
    /**
     * Returns the filtered clusters with the best cost. Rotated or not, splitted in cluster and noise cluster or not.
     * @param data DataObject[]: objects to be filtered
     * @return IcaCluster[2]: result[0]: filtered cluster, result[1]: filtered noise, can be empty (numObj = 0).
     *COSTS: Parameter costs.
     */
    private IcaCluster[] getBestFilteredClusters(DataObject[] o, boolean verbose){
        //if(verbose)
        System.out.println("getBestFilteredClusters: numObj " + o.length);
        if(o.length == 1){
            if(verbose)
                System.out.println("Singleton cluster");
            IcaCluster[] result = new IcaCluster[2];
            result[0] = new IcaCluster(0);
            result[0].cost = dim * 32 * 2;
            result[0].numObj = 1;
            result[1] = new IcaCluster(1);
            //clusterID
            int[] ids = new int[1];
            ids[0] = 0;
            aktClusterID = ids;
            return result;
        }
        IcaCost ic = new IcaCost(o);
        IcaCluster[] result = ic.fastRobustGenGauss();
        System.out.println("coding cost cluster " + result[0].cost);
        System.out.println("coding cost noise " + result[1].cost);
        //add parameter costs
        result[0].cost += dim * 32 * 2 + dim * dim * 32; //mean + rotation matrix for cluster
        //result[1].cost += dim * 32 * 2;
        System.out.println("coding cost cluster with parameter costs " + result[0].cost);
        System.out.println("coding cost noise with parameter costs " + result[1].cost);
        // compute ID costs to decide if noise should be filtered out
        double costID = 0.0;
        double costIDCluster = result[0].numObj * Math.log((double)o.length/(double)result[0].numObj)/Math.log(2);
        double costIDNoise = result[1].numObj * Math.log((double)o.length/(double)result[1].numObj)/Math.log(2);
        System.out.println("costID cluster " + costIDCluster);
        System.out.println("costID noise " + costIDNoise);
        costID += costIDCluster;
        costID += costIDNoise;
        // determine cost for all objects in cluster
        //double allObjectsInCluster =ic.getMergedCost(o);
        double allObjectsInCluster =Double.MAX_VALUE;
        System.out.println("all objects in cluster: " + allObjectsInCluster);
        allObjectsInCluster += dim * 32 * 2 + dim * dim * 32; //add parameter costs
        System.out.println("all objects in cluster with parameter costs: " + allObjectsInCluster);
        //TEST
        //allObjectsInCluster += 100;
        //TEST
        // determine overall costs for filtered
        double costFiltered = 0.0;
        costFiltered += result[0].cost;
        costFiltered += result[1].cost;
        //TEST
        //costFiltered += costID;
        //TEST
        if(verbose)
            System.out.println("Cost all objects in cluster: " + allObjectsInCluster + " cost filtered: " + costFiltered);
        if(costFiltered < allObjectsInCluster){
            int[] ids = ic.getClusterID();
            aktClusterID = ids;
            return result;
        } else{ //all objects in cluster
            int[] ids = new int[o.length];
            for(int i = 0; i < ids.length; i++)
                ids[i] = 0;
            aktClusterID = ids;
            result[0] = new IcaCluster();
            result[0].ID = 0;
            result[0].cost = allObjectsInCluster;
            result[0].numObj = o.length;
            result[1] = new IcaCluster(1);
            result[1].numObj = 0;
            return result;
        }
    }
    
    
    public void testRun(){
        IcaCluster[] result = getBestFilteredClusters(data, true);
        VisuSimple vk = new VisuSimple(data);
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
    }
    
    public void run(){
        IcaCluster[] result = getBestFilteredClusters(data, false);
        //adjust clusterIDs
        for(int i = 0; i < data.length; i++)
            data[i].clusterID = aktClusterID[i];
        bestClusters = new IcaCluster[1];
        bestClusters[0] = result[0];
        bestNoiseClusters = new IcaCluster[1];
        bestNoiseClusters[0] = result[1];
        double costFiltered = currentCost();
//        //TEST
//          double costFiltered = currentCostWithoutIDCosts();
//          //TEST
        double minCost = costFiltered;
        System.out.println("filtered data set " + costFiltered);
        int[] filteredID = new int[data.length];
        for(int i = 0; i < data.length; i++)
            filteredID[i] = data[i].clusterID;
        boolean improvement = true;
        //improvement = splitCluster();
        int aktTry = 0;
        IcaCluster[] lastLocalMinClusters = new IcaCluster[1];        //save Clusters and ClusterIDs of the last local minimum
        //lastLocalMinClusters[0] = bestClusters[0];
        IcaCluster[] lastLocalMinNoise = new IcaCluster[1];
        //lastLocalMinNoise[0] = bestNoiseClusters[0];
        int[] lastLocalMinIDs = new int[numObj];
        /*for(int i = 0; i < numObj; i++)
            lastLocalMinIDs[i] = data[i].clusterID;
         */
        while(improvement || aktTry < numTry){
            //improvement = splitCluster(aktTry);
            System.out.println("-------------------------------------");
            double costUnsplit = currentCost();
//            //TEST
//             double costUnsplit = currentCostWithoutIDCosts();
//            //TEST
            int numClUnsplit = numClustersNumNoiseClusters()[0];
            int numNUnsplit = numClustersNumNoiseClusters()[1];
            System.out.println("SYMMETRIC SPLIT NUMBER: " + iter + " CURRENT COST " + costUnsplit);
            improvement = symmetricSplit(aktTry);
            if(improvement){
                minCost = currentCost();
//                //TEST
//                  minCost = currentCostWithoutIDCosts();
//                //TEST
                System.out.println("GLOBAL IMPROVEMENT: " + " CURRENT COST " + minCost + " COST OLD: " + costUnsplit);
                lastLocalMinClusters = bestClusters;
                lastLocalMinNoise = bestNoiseClusters;
                for(int i = 0; i < numObj; i++)
                    lastLocalMinIDs[i] = data[i].clusterID;
                aktTry = 0;
            } else{
                aktTry++;
                System.out.println("NO GLOBAL IMPROVEMENT: CURRENT COST " + currentCost()/*currentCostWithoutIDCosts()*/ + " COST OLD: " + costUnsplit + " TRYING FOR ANOTHER " + (numTry - aktTry) + " TIMES");
            }
            System.out.println("OLD: " + numClUnsplit + " clusters, " + numNUnsplit + " noiseClusters.");
            System.out.println("NEW: " + numClustersNumNoiseClusters()[0] + " clusters " + numClustersNumNoiseClusters()[1] + " noiseClusters.");
            System.out.println("-------------------------------------");
            iter++;
        }
        System.out.println("done. " + iter + " iterations");
        
        //adjust clusters and IDs as from the last local minimum
        if(lastLocalMinClusters.length > 1 || lastLocalMinNoise.length > 1){
            bestClusters = lastLocalMinClusters;
            bestNoiseClusters = lastLocalMinNoise;
            for(int i = 0; i < data.length; i++)
                data[i].clusterID = lastLocalMinIDs[i];
            System.out.println("BEST: " + bestClusters.length + " clusters ");
        }
        //boolean impr = finalSplit();
        //mergeClusters();
        simpleGreedyMerge(0);
        minCost = currentCost();
//        //TEST
//          minCost = currentCostWithoutIDCosts();
//        //TEST
        
        //testMerge();
        //FinalAssignment f = new FinalAssignment(data, bestClusters, 3);
        //f.assign();
        //TEST TEST
//        if(costFiltered < minCost){
//            minCost = costFiltered;
//            for(int i = 0; i < data.length; i++)
//                data[i].clusterID = filteredID[i];
//        }
        //TEST
        System.out.println("BEST COST: " + minCost);
        /*
        //print data of specified clusters
        IO ea = new IO();
        ea.printClusters(data, 4);
        ea.printClusters(data, 2);
        ea.printClusters(data, 0);
        ea.printClusters(data, 10);
         */
        
//evaluate clustering result
//        int[] clusterIDs = new int[bestClusters.length];
//        for(int i = 0; i < bestClusters.length; i++)
//            clusterIDs[i] = bestClusters[i].ID;
//        ClusteringEvaluation e = new ClusteringEvaluation(data, clusterIDs);
//        e.evaluate();
        
        VisuInfo vi = new VisuInfo(data, "result");
        vi.setSize(600,600);
        vi.setLocation(200,0);
        vi.setVisible(true);
        
    }
    
    
    private int[] numClustersNumNoiseClusters(){
        int numCl = 0;
        int numN = 0;
        for(int i = 0; i < bestClusters.length; i++)
            if(bestClusters[i].numObj > 0)
                numCl++;
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                numN++;
        int[] result = new int[2];
        result[0] = numCl;
        result[1] = numN;
        return result;
    }
    
    private boolean symmetricSplit(int aktTry){
        //aktTry = numTry;
        IcaCluster[] currentClusters = bestClusters.clone();
        boolean[] improvement = new boolean[currentClusters.length];
        for(int i = 0; i < currentClusters.length; i++){
            improvement[i] = splitCluster(false, i);
        }
        boolean impr = false;
        for(int i = 0; i < improvement.length; i++)
            if(improvement[i])
                impr = true;
        if(!impr && aktTry < numTry){
            for(int i = 0; i < currentClusters.length; i++)
                improvement[i] = splitCluster(true, i);
        }
        return impr;
    }
    
///TEST
    private void checkIDs(){
        Vector ids = new Vector();
        for(int i = 0; i < data.length; i++)
            if(!ids.contains(data[i].clusterID))
                ids.add(data[i].clusterID);
        System.out.println("numIDS " + ids.size());
    }
//TEST
    
    /**
     * Determines the next cluster to split, splits the cluster and returns if improvment of cost results from this split.
     * @param aktTry aktTry: number of trys after minimum cost has been reached
     * @return boolean: improvement of cost
     */
    private boolean splitCluster(boolean splitAnyway, int splitIndex){
        if(bestClusters[splitIndex].numObj == 1)
            return false;
        int[] currentID1 = new int[numObj]; //clusterIDs for cluster and noise cluster if filtered
        int[] currentID2;
        boolean improvement = false;
        boolean filtered1 = false; //k-means cluster 1 filtered?
        boolean filtered2 = false;
        //int splitIndex = chooseSplitCluster();
        //get all objects of the cluster to split
        double costOld = bestClusters[splitIndex].cost + bestNoiseClusters[splitIndex].cost;
        DataObject[] cobj = db.getObjects(bestClusters[splitIndex].ID); // hier obj kopieren und ClusterID auf 0 setzen!
        DataObject[] nobj = db.getObjects(bestNoiseClusters[splitIndex].ID);
        int numAll = cobj.length+nobj.length;
        DataObject[] all = new DataObject[numAll];
        for(int i = 0; i < numAll; i++)
            if(i < cobj.length)
                all[i] = cobj[i];
            else
                all[i] = nobj[i-cobj.length];
        DB kmDb = new DB(all);
        KMeans k = new KMeans(2, kmDb);
        DataObject[] newIDs = k.run(); //ClusterIDs 1 and 2
        
        //objects assigned to cluster 1 by kMeans
        DataObject[] id_1 = kmDb.getObjects(1);
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_1 = new int[id_1.length];
        int[] internalIndex_1 = new int[id_1.length];
        for(int i = 0; i < id_1.length; i++){
            numbers_1[i] = id_1[i].number;
            internalIndex_1[i] = i;
        }
        HashMap h_1 = new HashMap(id_1.length);
        for(int i = 0; i < id_1.length; i++)
            h_1.put(internalIndex_1[i], numbers_1[i]);
        for(int i = 0; i < id_1.length; i++)
            id_1[i].number = internalIndex_1[i];
        //run KMCost for KMeans-cluster 1
        //mit Verwendung von getBestFilteredClusters
        IcaCluster[] clusters1 = getBestFilteredClusters(id_1, true);
        currentID1 = aktClusterID;
        clusters1[0].ID = bestClusters[splitIndex].ID;
        clusters1[1].ID = bestNoiseClusters[splitIndex].ID;
        //objects assigned to clusterID 2 by kMeans
        DataObject[] id_2 = kmDb.getObjects(2);
        //change numbers of objects to internal numbers for KMCost
        int[] numbers_2 = new int[id_2.length];
        int[] internalIndex_2 = new int[id_2.length];
        for(int i = 0; i < id_2.length; i++){
            numbers_2[i] = id_2[i].number;
            internalIndex_2[i] = i;
        }
        HashMap h_2 = new HashMap(id_2.length);
        for(int i = 0; i < id_2.length; i++)
            h_2.put(internalIndex_2[i], numbers_2[i]);
        for(int i = 0; i < id_2.length; i++)
            id_2[i].number = internalIndex_2[i];
        IcaCluster[] clusters2 = getBestFilteredClusters(id_2, true);
        currentID2 = aktClusterID;
        clusters2[0].ID = maxClusterID + 2;
        clusters2[1].ID = maxClusterID + 3;
        //update current clusters
        for(int i = 0; i < bestClusters.length; i++)
            if(i != splitIndex)
                clusters.add(bestClusters[i]);
            else
                clusters.add(clusters1[0]);
        clusters.add(clusters2[0]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(i != splitIndex)
                noiseClusters.add(bestNoiseClusters[i]);
            else
                noiseClusters.add(clusters1[1]);
        noiseClusters.add(clusters2[1]);
        //boolean improvement = true;
        //boolean improvement = evaluateCost(splitIndex, cl, noise);
        if(splitAnyway)
            improvement = true;
        else
            improvement = improvement(true);
        if(improvement){
            //set new clusters and new noiseClusters as best clusters.
            IcaCluster[] newClusters = new IcaCluster[clusters.size()];
            IcaCluster[] newNoiseClusters = new IcaCluster[clusters.size()];
            for(int i = 0; i < clusters.size(); i++){
                newClusters[i] = (IcaCluster)clusters.elementAt(i);
                int aktID = newClusters[i].ID;
                for(int j = 0; j < noiseClusters.size(); j++){
                    if(((IcaCluster)noiseClusters.elementAt(j)).ID == aktID+1)
                        newNoiseClusters[i] = (IcaCluster)noiseClusters.elementAt(j);
                }
                if(newNoiseClusters[i] == null){
                    newNoiseClusters[i] = new IcaCluster();
                    newNoiseClusters[i].ID = aktID+1;
                }
            }
            for(int i = 0; i < noiseClusters.size(); i++)
                newNoiseClusters[i] = (IcaCluster)noiseClusters.elementAt(i);
            bestClusters = newClusters;
            bestNoiseClusters = newNoiseClusters;
            //reset bestCost
            bestCost = 0.0;
            for(int i = 0; i < bestClusters.length; i++)
                bestCost += bestClusters[i].cost;
            for(int i = 0; i < bestNoiseClusters.length; i++)
                bestCost += bestNoiseClusters[i].cost;
            
            maxClusterID = maxClusterID+2;
            int splittedID = bestClusters[splitIndex].ID;
            
            for(int i = 0; i < currentID1.length; i++){
                int index = ((Integer)h_1.get(i)).intValue();
                if(currentID1[i] == 0)
                    data[index].clusterID = splittedID;
                else
                    data[index].clusterID = splittedID+1;
            }
            for(int i = 0; i < currentID2.length; i++){
                int index = ((Integer)h_2.get(i)).intValue();
                if(currentID2[i] == 0)
                    data[index].clusterID = maxClusterID;
                else
                    data[index].clusterID = maxClusterID + 1;
            }
        }
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        if(improvement)
            return true;
        else{
            return false;
        }
    }
    
    /**
     * Try once to split all clusters and all noise clusters individually. Called after top-down has reached first cost minimum and before merging.
     * @return boolean: improvement for at least one cluster or noiseCluster
     */
    private boolean finalSplit(){
        boolean impr = false;
        Vector<IcaCluster> nonEmpty = new Vector<IcaCluster>();
        for(int i = 0; i < bestClusters.length; i++)
            if(bestClusters[i].numObj > 0)
                nonEmpty.add(bestClusters[i]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0)
                nonEmpty.add(bestNoiseClusters[i]);
        boolean[] improvement = new boolean[nonEmpty.size()];
        IcaCluster[] actClusters = new IcaCluster[nonEmpty.size()];
        for(int i = 0; i < actClusters.length; i++)
            actClusters[i] = (IcaCluster)nonEmpty.elementAt(i);
        //relable noiseClusters and generate dummy noiseClusters to use the method splitCluster()
        int clusterCount = 0;
        for(int i = 0; i < actClusters.length; i++){
            int oldClusterID = actClusters[i].ID;
            if(oldClusterID != clusterCount){
                actClusters[i].ID = clusterCount;
                adjustClusterIDs(oldClusterID, clusterCount);
            }
            clusterCount = clusterCount+2;
        }
        maxClusterID = clusterCount;
        /*
        VisuSimple vk = new VisuSimple(data);
        vk.setSize(600,600);
        vk.setLocation(200,0);
        vk.setVisible(true);
         */
        
        IcaCluster[] actNoise = new IcaCluster[actClusters.length];
        for(int i = 0; i < actNoise.length; i++)
            actNoise[i] = new IcaCluster(actClusters[i].ID + 1);
        bestClusters = actClusters;
        bestNoiseClusters = actNoise;
        for(int i = 0; i < improvement.length; i++){
            improvement[i] = splitCluster(false, i);
            if(improvement[i])
                impr = true;
        }
        if(impr){
            System.out.println("IMPROVEMENT DUE TO FINAL SPLIT. NEW: " + bestClusters.length + " clusters");
        }
        return impr;
    }
    
/////TEST/////////////////////////////////////////////////////////
//    private void testMerge(){
//        Vector<IcaCluster> aktClusters = new Vector<IcaCluster>(); //current clusters and non-empty noiseClusters
//        clusters.removeAllElements();
//        noiseClusters.removeAllElements();
//        for(int i = 0; i < bestClusters.length; i++)
//            aktClusters.add(bestClusters[i]);
//        for(int i = 0; i < bestNoiseClusters.length; i++)
//            if(bestNoiseClusters[i].numObj > 0)
//                aktClusters.add(bestNoiseClusters[i]);
//        clusters = aktClusters;
//        //all clusters and noiseClusters to be merged in bestClusters, dummy noiseCluster.
//        IcaCluster[] nonEmpty = new IcaCluster[aktClusters.size()];
//        for(int i = 0; i < nonEmpty.length; i++)
//            nonEmpty[i] = (IcaCluster)aktClusters.elementAt(i);
//        bestClusters = nonEmpty;
//        IcaCluster[] newNoise = new IcaCluster[1];
//        newNoise[0] = new IcaCluster(1);
//        newNoise[0].numObj = 0;
//        bestNoiseClusters = newNoise; //dummy noisecluster
//        IcaCluster m = merge(bestClusters[1], bestClusters[2]);
//        clusters.remove(bestClusters[1]);
//        clusters.remove(bestClusters[2]);
//        clusters.add(m);
//        boolean imp = improvement(true); //ok
//        setNewBestClusters();
//        adjustClusterIDs(4, 2);
//        /*
//        IcaCluster m1 = merge(bestClusters[3], bestClusters[4]);
//        clusters.remove(bestClusters[3]);
//        clusters.remove(bestClusters[4]);
//        clusters.add(m1);
//        boolean impr = improvement(true); //ok
//        setNewBestClusters();
//        adjustClusterIDs(5, 1);
//        IcaCluster m2 = merge(bestClusters[0], bestClusters[2]);
//        clusters.remove(bestClusters[0]);
//        clusters.remove(bestClusters[2]);
//        clusters.add(m2);
//        boolean imprr = improvement(true); //ok
//
//         */
//        System.out.println("mist");
//    }
/////TEST
    
    private void setNewBestClusters(){
        IcaCluster[] newBest = new IcaCluster[clusters.size()];
        IcaCluster[] newBestNoise = new IcaCluster[clusters.size()];
        for(int i = 0; i < clusters.size(); i++){
            newBest[i] = (IcaCluster)clusters.elementAt(i);
            newBestNoise[i] = new IcaCluster(-1);    //dummy noise fr merge
        }
        bestNoiseClusters = newBestNoise;
        bestClusters = newBest;
    }
    
    private void adjustClusterIDs(int oldID, int newID){
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == oldID)
                data[i].clusterID = newID;
        
    }
    
//    private void mergeClusters(){
//        int numRandom = 1000;
//        Random r = new Random(2);
//        IcaCluster c;
//        IcaCluster d;
//        Vector<IcaCluster> aktClusters = new Vector<IcaCluster>(); //current clusters and non-empty noiseClusters
//        clusters.removeAllElements();
//        noiseClusters.removeAllElements();
//        for(int i = 0; i < bestClusters.length; i++)
//            aktClusters.add(bestClusters[i]);
//        for(int i = 0; i < bestNoiseClusters.length; i++)
//            if(bestNoiseClusters[i].numObj > 0)
//                aktClusters.add(bestNoiseClusters[i]);
//        clusters = aktClusters;
//        //all clusters and noiseClusters to be merged in bestClusters, dummy noiseCluster.
//        IcaCluster[] nonEmpty = new IcaCluster[aktClusters.size()];
//        for(int i = 0; i < nonEmpty.length; i++)
//            nonEmpty[i] = (IcaCluster)aktClusters.elementAt(i);
//        bestClusters = nonEmpty;
//        IcaCluster[] newNoise = new IcaCluster[1];
//        newNoise[0] = new IcaCluster(1);
//        newNoise[0].numObj = 0;
//        bestNoiseClusters = newNoise; //dummy noisecluster
//        for(int i = 0; i < numRandom; i++){
//            if(bestClusters.length > 1){
//                //merge random clusters and noiseClusters
//                int nextMerge1 = r.nextInt(bestClusters.length);
//                int nextMerge2 = r.nextInt(bestClusters.length);
//                while(nextMerge2 == nextMerge1)
//                    nextMerge2 = r.nextInt(bestClusters.length);
//                //if(bestClusters[nextMerge2].ID == bestClusters[nextMerge1].ID+1)       //do not try to merge cluster with its own noise cluster.
//                //  nextMerge2++;
//                //objects with this clusterID need to be relabeled, since the merged cluster gets the minimumID of c and d.
//                c = bestClusters[nextMerge1];
//                d = bestClusters[nextMerge2];
//                int maxID = Math.max(c.ID, d.ID);
//                int minID = Math.min(c.ID, d.ID);
//                IcaCluster m = merge(c, d);
//                clusters.remove(c);
//                clusters.remove(d);
//                clusters.add(m);
//                if(improvement(false)){
//                    //adjust clusterIDs
//                    adjustClusterIDs(maxID, minID);
//                    //set new clusters as best clusters
//                    setNewBestClusters();
//                } else{
//                    clusters.remove(m);
//                    clusters.add(c);
//                    clusters.add(d);
//                }
//            }
//            //clusters.removeAllElements();
//            //noiseClusters.removeAllElements();
//        }
//    }
    
    public void simpleGreedyMerge(int numTry){
        System.out.println("Merging clusters...");
        Vector<IcaCluster> aktClusters = new Vector<IcaCluster>(); //current clusters and non-empty noiseClusters
        clusters.removeAllElements();
        noiseClusters.removeAllElements();
        for(int i = 0; i < bestClusters.length; i++)
            aktClusters.add(bestClusters[i]);
        for(int i = 0; i < bestNoiseClusters.length; i++)
            if(bestNoiseClusters[i].numObj > 0){
            bestNoiseClusters[i].cost += dim * 32 * 2 + dim * dim * 32;
            aktClusters.add(bestNoiseClusters[i]);
            }
        clusters = aktClusters;
        bestCost = currentCost();
        boolean improvement = true;
        int aktTry = -1;
        //save Clusters and IDs of the last cost minimum
        int[] lastLocalMinIDs = new int[numObj];
        for(int i = 0; i < numObj; i++)
            lastLocalMinIDs[i] = data[i].clusterID;
        IcaCluster[] lastLocalMinClusters = new IcaCluster[clusters.size()];
        for(int i = 0; i < lastLocalMinClusters.length; i++)
            lastLocalMinClusters[i] = (IcaCluster)clusters.elementAt(i);
        while(improvement && clusters.size() > 1){
            IcaCluster[] startClusters = new IcaCluster[clusters.size()];
            for(int i = 0; i < startClusters.length; i++)
                startClusters[i] = (IcaCluster)clusters.elementAt(i);
            //determine the clusters with the maximum saved Cost
            double maxSaved = -Double.MAX_VALUE;
            IcaCluster[] toMerge = new IcaCluster[3];
            for (int i = 0; i < clusters.size(); i++){
                for(int j = 0; j < clusters.size(); j ++){
                    if(i != j){
                        IcaCluster a = (IcaCluster)clusters.elementAt(i);
                        IcaCluster b = (IcaCluster)clusters.elementAt(j);                
                        IcaCluster m = merge(a, b);
                        clusters.remove(a);
                        clusters.remove(b);
                        clusters.add(m);
                        double aktSaved = savedCost(true);
                        System.out.println("try to merge clusters " + a.ID + " and " + b.ID + " resulting cluster " + m.ID + " savedCost: " + aktSaved);
                        if(aktSaved > maxSaved){
                            toMerge[0] = a;
                            toMerge[1] = b;
                            toMerge[2] = m;
                            maxSaved = aktSaved;
                        }
                        clusters.removeAllElements();
                        for(int k = 0; k < startClusters.length; k++)
                            clusters.addElement(startClusters[k]);
                    }
                }
            }
            clusters.removeAllElements();
            for(int k = 0; k < startClusters.length; k++)
                clusters.addElement(startClusters[k]);
            if(maxSaved > 0){
                System.out.println("merging clusters " + toMerge[0].ID + " " +  toMerge[1].ID + " saved cost: " + maxSaved);
                int maxID = Math.max(toMerge[0].ID, toMerge[1].ID);
                int minID = Math.min(toMerge[0].ID, toMerge[1].ID);
                adjustClusterIDs(maxID, minID);
                clusters.remove(toMerge[0]);
                clusters.remove(toMerge[1]);
                clusters.add(toMerge[2]);
                setNewBestClusters();
                bestCost = currentCost();
                aktTry = 0;
                //save IDs
                for(int i = 0; i < numObj; i++)
                    lastLocalMinIDs[i] = data[i].clusterID;
                lastLocalMinClusters = bestClusters;
            } else{
                improvement = false;
                if(aktTry < numTry){
                    System.out.println("merging clusters " + toMerge[0].ID + " " +  toMerge[1].ID + " saved cost: " + maxSaved);
                    int maxID = Math.max(toMerge[0].ID, toMerge[1].ID);
                    int minID = Math.min(toMerge[0].ID, toMerge[1].ID);
                    adjustClusterIDs(maxID, minID);
                    clusters.remove(toMerge[0]);
                    clusters.remove(toMerge[1]);
                    clusters.add(toMerge[2]);
                    aktTry++;
                    //TEST
                    setNewBestClusters();
                    bestCost = currentCost();
                }
            }
        }
        for(int i = 0; i < data.length; i++)
            data[i].clusterID = lastLocalMinIDs[i];
        bestClusters = lastLocalMinClusters;
    }
    
////nicht fertig
//    public void greedyMerge(){
//        Vector<IcaCluster> aktClusters = new Vector<IcaCluster>(); //current clusters and non-empty noiseClusters
//        clusters.removeAllElements();
//        noiseClusters.removeAllElements();
//        for(int i = 0; i < bestClusters.length; i++)
//            aktClusters.add(bestClusters[i]);
//        for(int i = 0; i < bestNoiseClusters.length; i++)
//            if(bestNoiseClusters[i].numObj > 0)
//                aktClusters.add(bestNoiseClusters[i]);
//        clusters = aktClusters;
//        boolean improvement = true;
//        //determine cost table
//        IcaCluster[] intermediate = new IcaCluster[aktClusters.size()];
//        double[][] savedCost = new double[aktClusters.size()][aktClusters.size()];
//        for (int i = 0; i < aktClusters.size(); i++){
//            for(int j = 0; j < aktClusters.size(); j ++){
//                if(i != j){
//                    for(int k = 0; k < intermediate.length; k++)
//                        intermediate[k] = (IcaCluster)aktClusters.elementAt(k);
//                    IcaCluster a = (IcaCluster)aktClusters.elementAt(i);
//                    IcaCluster b = (IcaCluster)aktClusters.elementAt(j);
//                    IcaCluster m = merge(a, b);
//                    aktClusters.remove(a);
//                    aktClusters.remove(b);
//                    aktClusters.add(m);
//                    savedCost[i][j] = savedCost(false);
//                    aktClusters.removeAllElements();
//                    for(int k = 0; k < intermediate.length; k++)
//                        aktClusters.add(intermediate[k]);
//                } else
//                    savedCost[i][j] = 0.0;
//            }
//        }
//        while(improvement && aktClusters.size() > 1){
//            //determine the index of the clusters to be merged next, i.e. the index with maximum saved cost.
//            double maxSaved = -Double.MAX_VALUE;
//            int i_index = -1;
//            int j_index = -1;
//            for(int i = 0; i < savedCost.length; i++){
//                for(int j = 0; j < savedCost.length; j++){
//                    if(i != j && savedCost[i][j] > maxSaved){
//                        maxSaved = savedCost[i][j];
//                        i_index = i;
//                        j_index = j;
//                    }
//                }
//            }
//            if(maxSaved < 0)
//                improvement = false;
//            //merge these clusters and update aktClusters and the cost table: insert new cluster at the position of minIndex in aktClusters and in the cost table
//            IcaCluster a = (IcaCluster)(aktClusters.elementAt(i_index));
//            IcaCluster b = (IcaCluster)(aktClusters.elementAt(j_index));
//            IcaCluster m = merge(a, b);
//            aktClusters.remove(a);
//            aktClusters.remove(b);
//            aktClusters.add(Math.min(i_index, j_index), m);
//            //determine cost of the new cluster to all other clusters
//            double[] mCosts = new double[aktClusters.size()];
//        }
//    }
    
    private IcaCluster merge(IcaCluster c, IcaCluster d){
        boolean verbose = false;
        if(verbose)
            System.out.println("merge clusters: " + c.ID + " " + d.ID);
        //determine best model: rotated or not and TODO: best distribution
        DataObject[] mObj = getMergedObjects(c, d);
        int minID = Math.min(c.ID, d.ID);
        int maxID = Math.max(c.ID, d.ID);
        if(minID == maxID)
            System.err.println("trying to merge clusters with the same id: " + minID);
        IcaCost icaC = new IcaCost();
        IcaCluster m = icaC.mergedCluster(mObj);
        m.ID = minID;
        m.numObj = mObj.length;
        m.cost += dim * 32 * 2 + dim * dim * 32; //parameter costs
        return m;
    }
    
    
    private DataObject[] getMergedObjects(IcaCluster c, IcaCluster d){
        //DataObject[] members = new DataObject[c.numObj + d.numObj];
        DataObject[] members = new DataObject[countObj(c.ID, d.ID)];
        int counter = 0;
        for(int j = 0; j < data.length; j++)
            if(data[j].clusterID == c.ID || data[j].clusterID == d.ID){
            members[counter] = data[j];
            counter++;
            }
        return members;
    }
    
    public int countObj(int cId, int dId){
        int counter = 0;
        for(int i = 0; i < data.length; i++)
            if(data[i].clusterID == cId || data[i].clusterID == dId)
                counter++;
        return counter;
    }
    
    //TEST
    private double currentCostWithoutIDCosts(){
        boolean verbose = true;
        if(verbose)
            System.out.println("currentCostWithoutIDCosts");
        double cost = 0.0;
        for(int i = 0; i < bestClusters.length; i++){
            IcaCluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("IcaCluster " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                //cost += costID;
                cost += bestClusters[i].cost;
            }
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            IcaCluster akt = bestNoiseClusters[i];
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Noise " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                //cost += costID;
                cost += bestNoiseClusters[i].cost;
            }
        }
        return cost;
    }
    
    /**
     * Returns the current cost of bestClusters and bestNoiseClusters including ID cost
     * @return double: cost
     */
    private double currentCost(){
        boolean verbose = true;
        if(verbose)
            System.out.println("currentCost");
        double cost = 0.0;
        for(int i = 0; i < bestClusters.length; i++){
            IcaCluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("IcaCluster " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                cost += costID;
                cost += bestClusters[i].cost;
            }
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            IcaCluster akt = bestNoiseClusters[i];
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Noise " + akt.ID + " numObj " + akt.numObj + " cost " + akt.cost + " CostID " + costID);
                cost += costID;
                cost += bestNoiseClusters[i].cost;
            }
        }
        return cost;
    }
    
    /**
     * Checks if the costs are lower after one split. Cost for clusterID are counted here.
     * @return boolean: improvement of cost
     */
    private boolean improvement(boolean verbose){
        double costNew = 0.0;
        double costOld = 0.0;
        double costIDNew = 0.0;
        double costIDOld = 0.0;
        int clusterCountNew = 0;
        int clusterCountOld = 0;
        if (verbose)
            System.out.println("NEW:");
        for(int i = 0; i < clusters.size(); i++){
            IcaCluster akt = (IcaCluster)clusters.elementAt(i);
            if(clusters.elementAt(i).numObj > 0){
                clusterCountNew ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                costNew += akt.cost;
            } else
                if (verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < noiseClusters.size(); i++){
            IcaCluster akt = (IcaCluster)noiseClusters.elementAt(i);
            if(noiseClusters.elementAt(i).numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                clusterCountNew ++;
                costNew += akt.cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        if(verbose)
            System.out.println("OLD:");
        for(int i = 0; i < bestClusters.length; i++){
            IcaCluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                clusterCountOld ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            IcaCluster akt = bestNoiseClusters[i];
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                clusterCountOld ++;
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestNoiseClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        if(costNew < costOld){
            if(verbose)
                System.out.println("improvement : cost old: " + costOld + " " + clusterCountOld + " clusters, cost new: " + " " + costNew + " " + clusterCountNew + " clusters.");
            return true;
        }
        
        else{
            if(verbose)
                System.out.println("no improvement : cost old: " + costOld + " " + clusterCountOld + " clusters, cost new: " + " " + costNew + " " + clusterCountNew + " clusters.");
            return false;
        }
    }
    
    
    private double savedCost(boolean verbose){
        double costNew = 0.0;
        double costOld = 0.0;
        double costIDNew = 0.0;
        double costIDOld = 0.0;
        int clusterCountNew = 0;
        int clusterCountOld = 0;
        double savedCost;
        if (verbose)
            System.out.println("NEW:");
        for(int i = 0; i < clusters.size(); i++){
            IcaCluster akt = (IcaCluster)clusters.elementAt(i);
            if(clusters.elementAt(i).numObj > 0){
                clusterCountNew ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                costNew += akt.cost;
            } else
                if (verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < noiseClusters.size(); i++){
            IcaCluster akt = (IcaCluster)noiseClusters.elementAt(i);
            if(noiseClusters.elementAt(i).numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costNew += costID;
                clusterCountNew ++;
                costNew += akt.cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
        if(verbose)
            System.out.println("OLD:");
        for(int i = 0; i < bestClusters.length; i++){
            IcaCluster akt = bestClusters[i] ;
            if(akt.numObj > 0){
                clusterCountOld ++;
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                if(verbose)
                    System.out.println("cluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty cluster " + akt.ID + " cost " + akt.cost);
        }
        for(int i = 0; i < bestNoiseClusters.length; i++){
            IcaCluster akt = bestNoiseClusters[i];
            if(akt.numObj > 0){
                double costID = akt.numObj * Math.log((double)numObj/(double)akt.numObj)/Math.log(2);
                costOld += costID;
                clusterCountOld ++;
                if(verbose)
                    System.out.println("Noisecluster " + akt.ID + " cost " + akt.cost + " costID: " + costID);
                costOld += bestNoiseClusters[i].cost;
            } else
                if(verbose)
                    System.out.println("Empty noise cluster " + akt.ID + " cost " + akt.cost);
        }
//        if(costNew < costOld){
//            savedCost = Math.abs(costOld - costNew);
//            return savedCost;
//        } else
        return costOld-costNew;
    }
    
    
//Index von zu splittendem IcaCluster
    private int chooseSplitCluster(){
        double maxCost = 0.0;
        int max = bestClusters.length;
        for(int i = 0; i < bestClusters.length; i++){
            double aktCost = bestClusters[i].cost + bestNoiseClusters[i].cost;
            double aktNum = bestClusters[i].numObj + bestNoiseClusters[i].numObj;
            double costPerObject = aktCost/aktNum;
            if(costPerObject > maxCost){
                maxCost = costPerObject;
                max = i;
            }
        }
        System.out.println("split cluster " + bestClusters[max].ID);
        return max;
    }

   
    
   
//var of coordinate numCoord
    private double var(DataObject[] o, double[] mean, int numCoord){
        double var = 0.0;
        for(int i = 0; i < o.length; i++)
            var += (o[i].coord[numCoord] - mean[numCoord]) * (o[i].coord[numCoord] - mean[numCoord]);
        return var/(double)o.length;
    }
    
    private double variance(DataObject[] o, double[] mean){
        double var = 0.0;
        for(int i = 0; i < o.length; i++)
            var += o[i].distance(mean) * o[i].distance(mean);
        return var/(double)o.length;
    }
    
    private double gaussCost(double x, double mean, double var, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double va = Math.max(var, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*va)+precision*ln10+Math.pow(x-mean,2)/2/va)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    private double laplaceCost(double x, double mean, double var, int precision){
        double stdv = Math.sqrt(var);
        double b = stdv/Math.sqrt(2);
        if (b<Math.pow(10, -precision)/10)
            return 0 ;
        double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
        double prob = Math.max(Double.MIN_VALUE, Math.min(1.0, pdf * Math.pow(10, -precision)));
        //return 100000;
        return Math.log(1.0/prob)/Math.log(2);
    }
    
    private double[] mean(DataObject[] o){
        double[] mean = new double[o[0].coord.length];
        for(int i = 0; i < o.length; i++)
            for(int j = 0; j < mean.length; j++){
            mean[j] += o[i].coord[j];
            }
        for(int j = 0; j < mean.length; j++)
            mean[j] /= o.length;
        return mean;
    }
    
    public double [] median(DataObject[] members) {
        int numObj= members.length ;
        int dim = members[0].coord.length ;
        double[] result = new double[dim];
        double [] sortarray = new double[numObj] ;
        for (int i=0 ; i<dim ; i++) {
            double median = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sortarray[j]=members[j].coord[i] ;
            Arrays.sort(sortarray) ;
            if (numObj==1)
                result[i]=sortarray[0] ;
            else if (numObj%2 == 0)
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
            else
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
        }
        return result ;
    }
    
    public double[][] covarByMedian(DataObject[] data, double[] mean) {
        int numObj= data.length ;
        int dim = data[0].coord.length ;
        double [] sortarray = new double[numObj] ;
        double[][] cov = new double[dim][dim];
        if (numObj <15){
            for (int i=0 ; i<dim ; i++)
                cov[i][i]=1.0 ;
            return cov ;
        }
        for(int l1 = 0; l1 < dim; l1++) {
            for(int l2 = 0; l2 < dim; l2++) {
                for(int j = 0; j < numObj; j++)
                    sortarray[j] = (data[j].coord[l1]-mean[l1])*(data[j].coord[l2]-mean[l2]) ;
                
                Arrays.sort(sortarray) ;
                if (numObj%2 == 0)
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3
                            ;
            }
            
        }
        double maxdiff = 0.0 ;
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                sum += Math.abs(cov[i][j]) ;
            double diff = 1.1*sum - 2.1*cov[i][i] ;
            if (diff>maxdiff)
                maxdiff = diff ;
        }
        if (maxdiff > 0.0)
            for (int i=0 ; i<dim ; i++)
                cov[i][i] += maxdiff ;
        
        return cov ;
    }
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
}

