package de.lmu.entropyoutlier.algorithm;

//import D;
import Jama.EigenvalueDecomposition;
import Jama.Matrix;
import java.util.Random;

public class ICA
{
  public int ID;
  public double[][] ev;
  public double wscCost;
  boolean verbose = false;
  public double[] shift;
  public double[][] transform;
  public Matrix transformMatrix;
  public double[][] whitened;
  public Matrix whithenedMatrix;
  double[][] centered;
  public Matrix mw;
  public Matrix transposedMW;
  public Matrix transformTIMESmw;
  double log2 = Math.log(2.0D);
  int dim;
  int i;
  int j;
  int k;
  int m;
  int numObj;
  int alliter;
  int iter;
  double sum;
  double eRight;
  double scalar;
  double g;
  double[] diag;
  double[] old_w;
  double[] wUnorth;
  double[] eLeft;
  double[][] result;
  double[][] w;
  Matrix centeredMatrix;
  Matrix c;
  Matrix evm;
  Matrix ewm;
  Matrix res;
  EigenvalueDecomposition e;
  Random r = new Random();

  boolean firstGuess = true;

  int maxIterations = 100;
  double scalarTanH;
  double res2;
  double xsq;
  double x;

  public ICA(int dim)
  {
    this.dim = dim;
    this.transform = new double[dim][dim];
    this.transformMatrix = new Matrix(dim, dim);
    this.shift = new double[dim];
    this.mw = new Matrix(dim, dim);
    this.w = new double[dim][dim];
  }

  public void reset()
  {
    for (this.k = 0; this.k < this.dim; this.k += 1) {
      for (this.i = 0; this.i < this.dim; this.i += 1)
        this.w[this.k][this.i] = this.r.nextDouble();
      this.sum = 0.0D;
      for (this.i = 0; this.i < this.dim; this.i += 1)
        this.sum += this.w[this.k][this.i] * this.w[this.k][this.i];
      for (this.i = 0; this.i < this.dim; this.i += 1)
        this.w[this.k][this.i] /= Math.sqrt(this.sum);
    }
    this.firstGuess = true;
    this.maxIterations = 1000;
  }

  public void doICA(double[][] data, int coordinatesUsedUptoIndex)
  {
    this.numObj = coordinatesUsedUptoIndex;

    this.centered = new double[this.numObj][this.dim];
    this.centeredMatrix = new Matrix(this.numObj, this.dim);
    this.ev = ((double[][])null);
    this.wscCost = 0.0D;

    for (this.i = 0; this.i < this.dim; this.i += 1) {
      this.shift[this.i] = 0.0D;
      for (this.j = 0; this.j < this.numObj; this.j += 1)
        this.shift[this.i] += data[this.j][this.i];
      this.shift[this.i] /= this.numObj;
      for (this.j = 0; this.j < this.numObj; this.j += 1) {
        this.centeredMatrix.set(this.j, this.i, this.centered[this.j][this.i] = data[this.j][this.i] - this.shift[this.i]);
      }

    }

    this.c = this.centeredMatrix.transpose().times(this.centeredMatrix).times(1.0D / this.numObj);
    this.e = new EigenvalueDecomposition(this.c);
    this.evm = this.e.getV();
    this.ewm = this.e.getD();
    this.ev = this.ewm.getArrayCopy();
    this.diag = new double[this.dim];
    for (this.i = 0; this.i < this.dim; this.i += 1) {
      this.diag[this.i] = Math.sqrt(Math.min(Math.max(this.ewm.get(this.i, this.i), 1.0E-006), 1000000.0));
      //this.wscCost += Math.log(this.diag[this.i]) / this.log2;
      this.wscCost += Math.abs(Math.log(1/(this.diag[this.i]))) / this.log2;
    }
    //System.out.println("M=" + wscCost);

    for (this.i = 0; this.i < this.dim; this.i += 1) {
      for (this.j = 0; this.j < this.dim; this.j += 1) {
        this.transformMatrix.set(this.i, this.j, this.transform[this.i][this.j] = this.evm.get(this.i, this.j) / this.diag[this.j]);
      }
    }

    this.whitened = (this.whithenedMatrix = this.centeredMatrix.times(this.transformMatrix)).getArrayCopy();

    for (this.k = 0; this.k < this.dim; this.k += 1) {
      this.iter = 0;
      do {
        this.old_w = ((double[])this.w[this.k].clone());
        this.eLeft = new double[this.dim];
        this.eRight = 0.0D;
        for (int i = 0; i < this.numObj; ++i) {
          this.scalar = 0.0D;
          for (this.j = 0; this.j < this.dim; this.j += 1)
            this.scalar += this.w[this.k][this.j] * this.whitened[i][this.j];
          this.g = ((this.scalarTanH = Math.tanh(this.scalar)) / this.numObj);
          for (this.j = 0; this.j < this.dim; this.j += 1)
            this.eLeft[this.j] += this.g * this.whitened[i][this.j];
          this.eRight += (1.0D - (this.scalarTanH * this.scalarTanH)) / this.numObj;
        }
        for (this.j = 0; this.j < this.dim; this.j += 1) {
          this.w[this.k][this.j] = (this.eLeft[this.j] - (this.eRight * this.w[this.k][this.j]));
        }
        this.wUnorth = ((double[])this.w[this.k].clone());
        for (this.m = 0; this.m < this.k; this.m += 1) {
          this.scalar = 0.0D;
          for (this.j = 0; this.j < this.dim; this.j += 1)
            this.scalar += this.wUnorth[this.j] * this.w[this.m][this.j];
          for (this.j = 0; this.j < this.dim; this.j += 1)
            this.w[this.k][this.j] -= this.scalar * this.w[this.m][this.j];
        }
        this.sum = 0.0D;
        for (this.i = 0; this.i < this.dim; this.i += 1)
          this.sum += this.w[this.k][this.i] * this.w[this.k][this.i];
        for (this.i = 0; this.i < this.dim; this.i += 1)
          this.w[this.k][this.i] /= Math.sqrt(this.sum);
        this.sum = 0.0D;
        for (this.i = 0; this.i < this.dim; this.i += 1)
          this.sum += this.w[this.k][this.i] * this.old_w[this.i];
        this.iter += 1; }
      while ((Math.abs(this.sum) < 0.99999999D) && (this.iter < this.maxIterations));

      for (this.i = 0; this.i < this.dim; this.i += 1) {
        this.mw.set(this.k, this.i, this.w[this.k][this.i]);
      }

    }

    this.transposedMW = this.mw.transpose();

    this.transformTIMESmw = this.transformMatrix.times(this.transposedMW);

    this.maxIterations = 100;
    this.firstGuess = false;
  }
}