/*
 * FilterOld.java
 *
 * Created on October 4, 2006, 10:11 AM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package filterCluster;

/**
 *
 * @author Administrator
 */
import java.util.*;
import compr.DataObject;
import eval.Cluster;
import eval.SortPack;
import visu.VisuSimple;
import Jama.*;


/**
 * Filter cluster as in eval.clustering.
 * clusterIDs: 0: cluster, 1: noise.
 */
public class FilterOld {
    DataObject[] data;
    int precision;
    int numObj;
    int dim;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    
    
    /** Creates a new instance of FilterOld */
    public FilterOld(DataObject[] data, int precision) {
        this.data = data;
        this.precision = precision;
        numObj = data.length;
        dim = data[0].coord.length;
        for(int i = 0; i < data.length; i++)
            data[i].clusterID = 0;
    }
    
    public void run() {
        long startTime = System.currentTimeMillis();
        filterCluster();
        long endTime = System.currentTimeMillis();
        double runtime = (double)(endTime - startTime)/1000;
        System.out.println("duration " + runtime + " s.");
        Evaluation e = new Evaluation(data);
        e.evaluateFiltering();
        
    }
    
    public void filterCluster() {
        Matrix[] vv = new Matrix [10] ;
        vv[0] = Matrix.identity(dim,dim) ;
        Matrix[] ctr = new Matrix [10] ;
        Matrix [] vdvt = {
            Matrix.identity(dim, dim) ,
            getMahalanobis(data, 1.0, true, ctr, vv, 1) ,
            getMahalanobis(data, 0.5, true, ctr, vv, 2) ,
            getMahalanobis(data, 1.0, false, ctr, vv, 3) ,
            getMahalanobis(data, 0.5, false, ctr, vv, 4)
        } ;
        ctr[0] = ctr[3] ;
        double minCost = Double.MAX_VALUE ;
        double minCostUnsplit = Double.MAX_VALUE ;
        int minCase = -1 ;
        int minCut = -1 ;
        
        double [] lb = data[0].coord.clone() ;
        double [] ub = data[0].coord.clone() ;
        for (int i=1 ; i<numObj ; i++) {
            for (int j=0 ; j<dim ; j++) {
                lb[j] = Math.min(lb[j], data[i].coord[j]) ;
                ub[j] = Math.max(ub[j], data[i].coord[j]) ;
            }
        }
        for (int testCase=0 ; testCase<vdvt.length ; testCase ++) {
            //System.out.println("=============================================") ;
            //System.out.println("TestCase "+testCase) ;
            double [] distances = new double [numObj] ;
            for (int i=0 ; i<numObj ; i++) {
                Matrix x = rowVector(data[i].coord).minus(ctr[testCase]) ;
                // left out here
                
                distances[i]=x.times(vdvt[testCase]).times(x.transpose()).get(0,0) ;
            }
            SortPack [] distcopy = SortPack.array(distances) ;
            double [][] reduced = new double [numObj][] ;
            for (int i=0 ; i<numObj ; i++)
                reduced[i] =
                        rowVector(data[distcopy[i].index].coord).times(vv[testCase]).getColumnPackedCopy()
                        ;
            double [] meanCum = new double [dim] ;
            double [] mean = new double [dim] ;
            double [] varMeanCum = new double [dim] ;
            double [] varMean = new double [dim] ;
            double [] var = new double [dim] ;
            double [] lbCore = reduced[0].clone() ;
            double [] ubCore = reduced[0].clone() ;
            double [] cGaussCost = new double [dim] ;
            double [] cLaplaceCost = new double [dim] ;
            double cOverallCost = Double.MAX_VALUE ;
            for (int cut = 0 ; cut < numObj ; cut ++) {
                // update now models for cluster
                cOverallCost = 0 ;
                for (int j=0 ; j<dim ; j++) {
                    meanCum [j] += reduced[cut][j] ;
                    mean [j] = meanCum [j] / (cut+1) ;
                    varMeanCum [j] += reduced[cut][j] * reduced[cut][j] ;
                    var [j] = varMeanCum [j] / (cut/*+1*/) - mean[j]*mean[j] ;
                    // SOME SANITY CHECK NEEDED HERE !!!!
                    lbCore[j] = Math.min(lbCore[j], reduced[cut][j]) ;
                    ubCore[j] = Math.max(ubCore[j], reduced[cut][j]) ;
                    cGaussCost[j]=0 ;
                    cLaplaceCost[j]=0 ;
                    if(cut > 0)
                        for (int i=0 ; i<=cut ; i++) {
                        cGaussCost[j] += gaussCost(reduced[i][j],
                                mean[j], Math.sqrt(var[j]), precision) ;
                        cLaplaceCost[j] += laplaceCost(reduced[i][j],
                                mean[j], Math.sqrt(var[j]), precision) ;
                        }
                    
                    else {
                        cGaussCost[j] = Double.MAX_VALUE;
                        cLaplaceCost[j] = Double.MAX_VALUE;
                    }
                    cOverallCost += Math.min(cLaplaceCost[j], Math.min(cGaussCost[j],
                            (cut+1)*uniformCost(0, lbCore[j], ubCore[j], precision) ) )
                            + (numObj-cut-1)*uniformCost(0, lb[j], ub[j], precision)  ;
                }
                //Kostenangabe
//                if(testCase == 0)
//                    System.out.println("cost " + cOverallCost);
                if (cOverallCost<minCost){
                    minCost=cOverallCost ;
                    minCase = testCase ;
                    minCut = cut ;
                }
            }
            if (cOverallCost < minCostUnsplit)
                // Assume that all points are *cluster members*
                // The actual solution must outperform this situation
                minCostUnsplit = cOverallCost ;
        }
        
        // TEST TEST TEST minCostUnsplit: Assume that all points are
        // *outliers*
        minCostUnsplit = 0 ;
        for (int j=0 ; j<dim ; j++)
            minCostUnsplit += numObj*uniformCost(0, lb[j], ub[j], precision) ;
        
        int testCase = minCase ;
        double [] distances = new double [numObj] ;
        for (int i=0 ; i<numObj ; i++) {
            Matrix x = rowVector(data[i].coord).minus(ctr[testCase]) ;
            
            distances[i]=x.times(vdvt[testCase]).times(x.transpose()).get(0,0) ;
        }
        SortPack [] distcopy = SortPack.array(distances) ;
        for (int i=minCut+1 ; i<numObj ; i++)
            data[distcopy[i].index].clusterID = 1 ;
    }
    
    public Matrix getMahalanobis(DataObject[] data, double fraction,
            boolean robust, Matrix [] ctr, Matrix[] vv, int vNum) {
        // determines the matrix for mahalonobis distance either robust or not
        // based on a fraction of the data set which is selected by Euclidean nearest neighbors
        int numObj = data.length ;
        int numRed = (int)(numObj*fraction) ;
        double[] center = robust
                ? median(data)
                : mean(data) ;
        double[] distances = new double[numObj] ;
        for (int i=0 ; i<numObj ; i++)
            distances[i] = distance(center, data[i].coord);
        SortPack [] distcopy = SortPack.array(distances) ;
        DataObject [] dataRed = new DataObject [numRed] ;
        for (int i=0 ; i<numRed ; i++)
            dataRed[i] = data[distcopy[i].index] ;
        center = robust
                ? median(dataRed)
                : mean(dataRed) ;
        ctr[vNum] = rowVector(center) ;
        double[][] cov = robust
                ? covarByMedian(dataRed, center)
                : covar(dataRed, center) ;
        EigenvalueDecomposition e = new EigenvalueDecomposition(new Matrix(cov)) ;
        vv[vNum] = e.getV() ;
        Matrix d = e.getD() ;
        //d.print(6, 4);
        double maxEigen = -Double.MAX_VALUE ;
        for (int i=0 ; i<dim ; i++)
            maxEigen = Math.max(maxEigen, d.get(i,i)) ;
        for (int i=0 ; i<dim ; i++)
            d.set(i,i,1/Math.max(d.get(i,i), maxEigen*0.001)) ;
        return vv[vNum].times(d).times(vv[vNum].transpose()) ;
    }
    
    private double[][] getRobustEV(DataObject[] data){
        double[] center = median(data);
        double[][] cov = covarByMedian(data, center);
        EigenvalueDecomposition e = new EigenvalueDecomposition(new Matrix(cov)) ;
        return e.getV().getArrayCopy();
    }
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
    private double laplaceCost(double x, double mean, double std, int precision){
        double b = std/Math.sqrt(2);
        if (b<Math.pow(10, -precision)/10)
            return 0 ;
        double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
        double prob = Math.max(Double.MIN_VALUE, Math.min(1.0, pdf * Math.pow(10, -precision)));
        //return 100000;
        return Math.log(1.0/prob)/Math.log(2);
    }
    
    
    private double gaussCost(double x, double mean, double std, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
        double var = Math.max(std*std, Double.MIN_VALUE) ;
        return Math.max(0.0,(0.5*Math.log(2*Math.PI*var)+precision*ln10+Math.pow(x-mean,2)/2/var)/ln2 );
        
        //return Math.log(1/prob)/Math.log(2);
    }
    
    private double uniformCost(double x, double lb, double ub, int precision){
        double prob ;
        if (Math.abs(ub-lb) < Math.pow(10, -precision))
            prob = 1 ;
        else
            prob = Math.pow(10, -precision) / Math.abs(ub - lb) ;
        return Math.log(1/prob)/Math.log(2);
    }
    
    public double[] mean(DataObject[] o){
        double[] mean = new double[o[0].coord.length];
        for(int i = 0; i < o.length; i++)
            for(int j = 0; j < mean.length; j++){
            mean[j] += o[i].coord[j];
            }
        for(int j = 0; j < mean.length; j++)
            mean[j] /= o.length;
        return mean;
    }
    
    public double [] median(DataObject[] members) {
        int numObj= members.length ;
        int dim = members[0].coord.length ;
        double[] result = new double[dim];
        double [] sortarray = new double[numObj] ;
        for (int i=0 ; i<dim ; i++) {
            double median = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sortarray[j]=members[j].coord[i] ;
            Arrays.sort(sortarray) ;
            if (numObj==1)
                result[i]=sortarray[0] ;
            else if (numObj%2 == 0)
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
            else
                result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
        }
        return result ;
    }
    
    private double distance(double[] mean, double[] d){
        double sum = 0.0;
        for (int i = 0; i < d.length; i++){
            double dist = ((d[i] - mean[i])*(d[i] - mean[i]));
            sum = sum + dist;
        }
        return Math.sqrt(sum);
    }
    
    public double[][] covarByMedian(DataObject[] data, double[] mean) {
        int numObj= data.length ;
        int dim = data[0].coord.length ;
        double [] sortarray = new double[numObj] ;
        double[][] cov = new double[dim][dim];
        if (numObj <5){
            for (int i=0 ; i<dim ; i++)
                cov[i][i]=1.0 ;
            return cov ;
        }
        for(int l1 = 0; l1 < dim; l1++) {
            for(int l2 = 0; l2 < dim; l2++) {
                for(int j = 0; j < numObj; j++)
                    sortarray[j] = (data[j].coord[l1]-mean[l1])*(data[j].coord[l2]-mean[l2]) ;
                
                Arrays.sort(sortarray) ;
                if (numObj%2 == 0)
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3
                            ;
            }
            
        }
        double maxdiff = 0.0 ;
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                sum += Math.abs(cov[i][j]) ;
            double diff = 1.1*sum - 2.1*cov[i][i] ;
            if (diff>maxdiff)
                maxdiff = diff ;
        }
        if (maxdiff > 0.0)
            for (int i=0 ; i<dim ; i++)
                cov[i][i] += maxdiff ;
        
        return cov ;
    }
    
    private double[][] covar(DataObject[] data, double[] mean){
        double[][] cov = new double[dim][dim];
        for(int l = 0; l < data.length; l++)
            for(int i = 0; i < dim; i++)
                for(int j = 0; j < dim; j++)
                    cov[i][j]+= (data[l].coord[j]-mean[j])*(data[l].coord[i]-mean[i]);
        for(int i = 0; i < dim; i++)
            for(int j = 0; j < dim; j++)
                cov[i][j]/= data.length;
        return cov;
    }
    
}
