/*
 * FilterSort.java
 *
 * Created on November 6, 2006, 4:24 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package filterCluster;
import java.util.Arrays;
import java.util.Vector;
import Jama.*;
import eval.SortPack;
import compr.DataObject;
import visu.VisuInfo;

/**
 * Filter cluster by sorting according to coding cost.
 * @author Administrator
 */
public class FilterSort {
    int numObj;
    int dim;
    DataObject[] data;
    DataObject[] cluster;
    DataObject[] noise;  //current partitoning
    boolean[] clusterpt; // best partitioning: true if cluster point
    int clusterCount;
    int noiseCount;
    double currentOverallCost;    //save the current overall Cost
    double minCost; //minimum cost
    int precision;
    static int gauss = 0;
    static int uniform = 1;
    static int laplace = 2;
    static double ln10 = Math.log(10);
    static double ln2 = Math.log(2);
    boolean verbose = false;
    
    
    /**
     * Creates a new instance of FilterSort
     *
     * @param data The data to be filtered
     * @param precision precision of coding
     */
    public FilterSort(DataObject[] data, int precision, boolean rotated, boolean finalFilter){
        this.data = data;
        this.precision = precision;
        numObj = data.length;
        dim = data[0].coord.length;
        clusterCount = 0;
        noiseCount = 0;
        minCost = Double.MAX_VALUE;
        clusterpt = new boolean[numObj];
        if(rotated & !finalFilter)
            data = robustRotate(data, false);
    }
    
    
    
    public void filter(){
        cluster = data;
        clusterCount = data.length;
        double[] center = median(data);
        double[] costs = new double[data.length]; //costs[0]: alles Cluster... costs[n-1] alles Noise
        int c = 0;
        while (cluster.length >= 1){
            if(verbose)
                System.out.println(clusterCount + " clusterPts " + noiseCount + " noisePts. Cost: " + currentOverallCost);
            DataObject o = maximumCost(cluster, center);
            costs[c] = currentOverallCost;
            c++;
            updatePartitioning(o);
            if(cluster.length > 1)
                center = median(cluster);
        }
        
        setClusterIDs();
//        
//        Evaluation e = new Evaluation(data);
//        e.evaluateFiltering();    
 }
    
    
    /**
     * Method for refining the final clustering result. Initialize filtering with median, subspace orientation of the clusterPoints. ClusterID 0: cluster, 1: noise.
     */
    public void finalFilter(boolean rotated){
        DataObject[] clusterPts = clusterPts();
        double[] center = median(clusterPts);
        //double[] center = mean(clusterPts);
        if(rotated){
            double[][] dir = robustDirections(clusterPts, center);
             //double[][] dir = directions(clusterPts, center);
            data = rotateToDirections(data, dir, center);
//          //TEST
//            VisuInfo vi = new VisuInfo(data, "result");
//            vi.setSize(600,600);
//            vi.setLocation(200,0);
//            vi.setVisible(true);
//            //TEST
            //rotate center
//            double[] centerRot = new double[dim];
//            for(int i = 0; i < dim; i++)
//                for(int j = 0; j < dim; j++)
//                    centerRot[i] += center[i] * dir[j][i];
//            center = centerRot; 
            
//            clusterPts = clusterPts();
//            center = mean(clusterPts);
            
            center[0] = 0.0;
            center[1] = 0.0;
        }
        cluster = data;
        clusterCount = data.length;
        double[] costs = new double[data.length]; //costs[0]: alles Cluster... costs[n-1] alles Noise
        int c = 0;
        while (cluster.length >= 1){
            if(verbose)
                System.out.println(/*clusterCount + " clusterPts " + noiseCount + " noisePts. Cost: " +*/  currentOverallCost);
            DataObject o = maximumCost(cluster, center);
            costs[c] = currentOverallCost;
            c++;
            updatePartitioning(o);
        }
        setClusterIDs();        
        }
    
        
        private DataObject[] clusterPts(){
            Vector<DataObject> cluster = new Vector<DataObject>();
            for(int i = 0; i < data.length; i++)
                if(data[i].clusterID == 0)
                    cluster.addElement(data[i]);
            DataObject[] d = new DataObject[cluster.size()];
            for(int i = 0; i < cluster.size(); i++)
                d[i] = (DataObject)cluster.elementAt(i);
            return d;
            
        }
        
        
        public boolean[] getClusterIDs(){
            return this.clusterpt;
        }
        
        public DataObject[] getData(){
            return this.data;
        }
        
        public double getCost(){
            return this.minCost;
        }
        
        private void setClusterIDs(){
            for(int i = 0; i < data.length; i++)
                if(clusterpt[i])
                    data[i].clusterID = 0;
                else
                    data[i].clusterID = 1; //noise point
        }
        
        /**
         * update the arrays cluster and noise by removing the object o from the cluster.
         * @param o object to be removed
         */
        private void updatePartitioning(DataObject o){
            clusterCount--;
            noiseCount++;
            DataObject[] clusterNew = new DataObject[clusterCount];
            DataObject[] noiseNew = new DataObject[noiseCount];
            int c = 0;
            for(int i = 0; i < cluster.length; i++)
                if(cluster[i].number != o.number){
                clusterNew[c] = cluster[i];
                c++;
                }
            if(noise != null)
                for(int i = 0; i < noise.length; i++)
                    noiseNew[i] = noise[i];
            noiseNew[noiseNew.length-1] = o;
            cluster = clusterNew;
            noise = noiseNew;
        }
        
        /**
         * Determine coding costs for the current noise points using uniform distribution.
         * @return coding cost (all noise objects and all coordinates)
         */
        private double noiseCost(){
            if(noise != null){
                double[][] bounds = new double[dim][2];
                for(int i = 0; i < dim; i++){
                    double[] b = bounds(noise, i);
                    bounds[i][0] = b[0]; //lower bound of coordinate i
                    bounds[i][1] = b[1]; //upper bound of coordinate i
                }
                double noiseCost = 0.0;
                for(int i = 0; i < noise.length; i++)
                    for(int j = 0; j < dim; j++)
                        noiseCost +=  uniformCost(noise[i].coord[j], bounds[j][0], bounds[j][1], precision);
                return noiseCost;
            } else return 0.0;
        }
        
        /**
         * Determine the point with the highest coding cost to leave out this point.
         * @param clusterPts the points currently in the cluster
         * @param center the center of this point set: mean/median
         * @return the point with the highest cost
         */
        public DataObject maximumCost(DataObject[] clusterPts, double[] center){
            //determine coding scheme: pick for each dimension the best distribution function: NOT ALLOWED: UNIFORM CLUSTER CODING
            //determine parameters
            double[] stdv = new double[dim];
            for(int i = 0; i < dim; i++)
                stdv[i] = stdv(clusterPts, center, i);
            double[][] bounds = new double[dim][2];
            for(int i = 0; i < dim; i++){
                double[] b = bounds(clusterPts, i);
                bounds[i][0] = b[0]; //lower bound of coordinate i
                bounds[i][1] = b[1]; //upper bound of coordinate i
            }
            double[] gaussCost = new double[dim];
            double[] laplaceCost = new double[dim];
            double[] uniformCost = new double[dim];
            int[] bestCoding = new int[dim];  //save best coding cost for each dimension
            for(int i = 0; i < dim; i++){
                for(int j = 0; j < clusterPts.length; j++){
                    //TEST: Filter only with laplacian coding
                    gaussCost[i] += gaussCost(clusterPts[j].coord[i], center[i], stdv[i], precision);
                    laplaceCost[i] += laplaceCost(clusterPts[j].coord[i], center[i], stdv[i], precision);
                    //uniformCost[i] += uniformCost(clusterPts[j].coord[i], bounds[i][0], bounds[i][1], precision);
                }
            }
            for(int i = 0; i < dim; i++)
                uniformCost[i] = Double.MAX_VALUE;
//            ///TEST
//             for(int i = 0; i < dim; i++)
//                gaussCost[i] = Double.MAX_VALUE;
//            ///TEST
            double overallCost = 0.0;
            for(int i = 0; i < dim; i++){
                if(gaussCost[i] <= uniformCost[i] && gaussCost[i] <= laplaceCost[i]){
                    bestCoding[i] = gauss;
                    overallCost += gaussCost[i];
                }
                if(laplaceCost[i] <= gaussCost[i] && laplaceCost[i] <= uniformCost[i]){
                    bestCoding[i] = laplace;
                    overallCost += laplaceCost[i];
                }
                if(uniformCost[i] <= gaussCost[i] && uniformCost[i] <= laplaceCost[i]){
                    bestCoding[i] = uniform;
                    overallCost += uniformCost[i];
                }
            }
            currentOverallCost = overallCost + noiseCost();
            //save IDs of cluster and noise points
            if(currentOverallCost < minCost){
                minCost = currentOverallCost;
                for(int i = 0; i < clusterpt.length; i++)
                    clusterpt[i] = false;
                for(int i = 0; i < cluster.length; i++)
                    clusterpt[cluster[i].number] = true;
            }
            //detemine the index of the point with the maximum coding cost according to this scheme
            double maxCost = 0.0;
            int maxIndex = 0;
            for(int i = 0; i < clusterPts.length; i++){
                double currentCost = 0.0;
                for(int j = 0; j < dim; j++){
                    if(bestCoding[j] == gauss)
                        currentCost += gaussCost(clusterPts[i].coord[j], center[j], stdv[j], precision);
                    if(bestCoding[j] == laplace)
                        currentCost += laplaceCost(clusterPts[i].coord[j], center[j], stdv[j], precision);
                    if(bestCoding[j] == uniform)
                        currentCost += uniformCost(clusterPts[i].coord[j], bounds[j][0], bounds[j][1], precision);
                }
                if(currentCost > maxCost){
                    maxCost = currentCost;
                    maxIndex = i;
                }
            }
//            if(verbose)
//                System.out.println("object with maximal coding cost " + clusterPts[maxIndex].number);
            return clusterPts[maxIndex];
            
        }
        
        public double[] mean(DataObject[] o){
            double[] mean = new double[o[0].coord.length];
            for(int i = 0; i < o.length; i++)
                for(int j = 0; j < mean.length; j++){
                mean[j] += o[i].coord[j];
                }
            for(int j = 0; j < mean.length; j++)
                mean[j] /= o.length;
            return mean;
        }
        
        private double [] median(DataObject[] o) {
            int numObj= o.length ;
            int dim = o[0].coord.length ;
            double[] result = new double[dim];
            double [] sortarray = new double[numObj] ;
            for (int i=0 ; i<dim ; i++) {
                double median = 0.0 ;
                for (int j=0 ; j<numObj ; j++)
                    sortarray[j]=o[j].coord[i] ;
                Arrays.sort(sortarray) ;
                if (numObj==1)
                    result[i]=sortarray[0] ;
                else if (numObj%2 == 0)
                    result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    result[i]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
            }
            return result ;
        }
        
        private double[][] covar(DataObject[] data, double[] mean){
            double[][] cov = new double[dim][dim];
            for(int l = 0; l < data.length; l++)
                for(int i = 0; i < dim; i++)
                    for(int j = 0; j < dim; j++)
                        cov[i][j]+= (data[l].coord[j]-mean[j])*(data[l].coord[i]-mean[i]);
            for(int i = 0; i < dim; i++)
                for(int j = 0; j < dim; j++)
                    cov[i][j]/= data.length;
            return cov;
        }
        
        public double[][] covarByMedian(DataObject[] data, double[] median) {
            int numObj= data.length ;
            int dim = data[0].coord.length ;
            double [] sortarray = new double[numObj] ;
            double[][] cov = new double[dim][dim];
            if (numObj <15){
                for (int i=0 ; i<dim ; i++)
                    cov[i][i]=1.0 ;
                return cov ;
            }
            for(int l1 = 0; l1 < dim; l1++) {
                for(int l2 = 0; l2 < dim; l2++) {
                    for(int j = 0; j < numObj; j++)
                        sortarray[j] = (data[j].coord[l1]-median[l1])*(data[j].coord[l2]-median[l2]) ;
                    Arrays.sort(sortarray) ;
                    if (numObj%2 == 0)
                        cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                    else
                        cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3;
                }
            }
            double maxdiff = 0.0 ;
            for (int i=0 ; i<dim ; i++){
                double sum = 0.0 ;
                for (int j=0 ; j<dim ; j++)
                    sum += Math.abs(cov[i][j]) ;
                double diff = 1.1*sum - 2.1*cov[i][i] ;
                if (diff>maxdiff)
                    maxdiff = diff ;
            }
            if (maxdiff > 0.0)
                for (int i=0 ; i<dim ; i++)
                    cov[i][i] += maxdiff ;
            return cov ;
        }
        
        private DataObject[] rotate(DataObject[] obj, boolean verbose){
            DataObject[] result = new DataObject[obj.length];
            double[] mean = mean(obj);
            double[][] covarm = covar(data, mean);
            Matrix covar = new Matrix(covarm);
            EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
            Matrix vectors = eig.getV();
            Matrix values = eig.getD();
            if(verbose){
                System.out.println("EW");
                values.print(5, 3);
                System.out.println("EV");
                vectors.print(5, 3);
            }
            double[][] v = vectors.getArrayCopy();
            double[][] transCoord = new double[obj.length][dim];
            for(int i = 0; i < obj.length; i++){
                for(int j = 0; j < dim; j++){
                    for(int k = 0; k < dim; k++){
                        transCoord[i][k] += (obj[i].coord[j]-mean[j]) * v[j][k];
                    }
                }
            }
            for(int i = 0; i < obj.length; i++){
                double[] co = new double[dim];
                for(int j = 0; j < dim; j++)
                    co[j] = transCoord[i][j];
                result[i] = new DataObject(co, obj[i].number, obj[i].clusterID);
            }
            return result;
        }
        
        private DataObject[] robustRotate(DataObject[] obj, boolean verbose){
            DataObject[] result = new DataObject[obj.length];
            double[] median = median(obj);
            double[][] covarm = covarByMedian(data, median);
            Matrix covar = new Matrix(covarm);
            EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
            Matrix vectors = eig.getV();
            Matrix values = eig.getD();
            if(verbose){
                System.out.println("EW");
                values.print(5, 3);
                System.out.println("EV");
                vectors.print(5, 3);
            }
            double[][] v = vectors.getArrayCopy();
            double[][] transCoord = new double[obj.length][dim];
            for(int i = 0; i < obj.length; i++){
                for(int j = 0; j < dim; j++){
                    for(int k = 0; k < dim; k++){
                        transCoord[i][k] += (obj[i].coord[j]-median[j]) * v[j][k];
                    }
                }
            }
            for(int i = 0; i < obj.length; i++){
                double[] co = new double[dim];
                for(int j = 0; j < dim; j++)
                    co[j] = transCoord[i][j];
                result[i] = new DataObject(co, obj[i].number, obj[i].clusterID);
            }
            return result;
        }
        
        private double[][] robustDirections(DataObject[] d, double[] median){
            double[][] covarm = covarByMedian(data, median);
            Matrix covar = new Matrix(covarm);
            EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
            Matrix vectors = eig.getV();
            return vectors.getArrayCopy();
        }
        
         private double[][] directions(DataObject[] d, double[] mean){
            double[][] covarm = covar(data, mean);
            Matrix covar = new Matrix(covarm);
            EigenvalueDecomposition eig = new EigenvalueDecomposition(covar);
            Matrix vectors = eig.getV();
            return vectors.getArrayCopy();
        }
        
        private DataObject[] rotateToDirections(DataObject[] obj, double[][] v, double[] median){
            DataObject[] result = new DataObject[obj.length];
            double[][] transCoord = new double[obj.length][dim];
            for(int i = 0; i < obj.length; i++){
                for(int j = 0; j < dim; j++){
                    for(int k = 0; k < dim; k++){
                        transCoord[i][k] += (obj[i].coord[j]-median[j]) * v[j][k];
                    }
                }
            }
            for(int i = 0; i < obj.length; i++){
                double[] co = new double[dim];
                for(int j = 0; j < dim; j++)
                    co[j] = transCoord[i][j];
                result[i] = new DataObject(co, obj[i].number, obj[i].clusterID);
            }
            return result;
        }
        
        private double laplaceCost(double x, double mean, double std, int precision){
            double b = std/Math.sqrt(2);
            if (b<Math.pow(10, -precision)/10)
                return 0 ;
            double pdf = 1/(2*b)*Math.exp(-Math.abs(x-mean)/b);
            double prob = Math.max(Double.MIN_VALUE, Math.min(1.0, pdf * Math.pow(10, -precision)));
            //return 100000;
            return Math.log(1.0/prob)/Math.log(2);
        }
        
        private double gaussCost(double x, double mean, double std, int precision){
       /* double pdf = Math.exp(-(x-mean)*(x-mean)/2.0/(std * std))/Math.sqrt(2 * Math.PI * std * std);
        double prob = pdf * Math.pow(10, -precision) ;
        if (std<4*Math.pow(10, -precision))
            System.err.println("Attention, resolution too large") ;*/
            double var = Math.max(std*std, Double.MIN_VALUE) ;
            return Math.max(0.0,(0.5*Math.log(2*Math.PI*var)+precision*ln10+Math.pow(x-mean,2)/2/var)/ln2 );
            
            //return Math.log(1/prob)/Math.log(2);
        }
        
        private double uniformCost(double x, double lb, double ub, int precision){
            double prob ;
            if (Math.abs(ub-lb) < Math.pow(10, -precision))
                prob = 1 ;
            else
                prob = Math.pow(10, -precision) / Math.abs(ub - lb) ;
            return Math.log(1/prob)/Math.log(2);
        }
        
        //stdv of coordinate numCoord
        private double stdv(DataObject[] o, double[] mean, int numCoord){
            double stdv = 0.0;
            for(int i = 0; i < o.length; i++)
                stdv += (o[i].coord[numCoord] - mean[numCoord]) * (o[i].coord[numCoord] - mean[numCoord]);
            return Math.sqrt(stdv/o.length);
        }
        
        //bounds[0]: lower bound, bounds[1]: upper bound
        private double[] bounds(DataObject[] o, int numCoord){
            int dim = o[0].coord.length;
            double[] bounds = new double[2];
            for(int i = 0; i < bounds.length; i++){
                bounds[0] = Double.MAX_VALUE;
                bounds[1] = -Double.MAX_VALUE;
            }
            for(int i = 0; i < o.length; i++){
                for(int j = 0; j < dim; j++){
                    if(o[i].coord[numCoord] < bounds[0])
                        bounds[0] = o[i].coord[j];
                    if(o[i].coord[numCoord] > bounds[1])
                        bounds[1] = o[i].coord[j];
                }
            }
            return bounds;
        }
        
    }
