package io;
import java.io.*;
import java.util.*;
import compr.*;
import weka.core.*;
import clusterer.IcaCluster;

/* IO.java
 *
 *
 */

/**
 *
 * @author  Claudia
 */


/**Class for IO. Reads input file, writes output.*/

public class IO {
    /**path input file*/
    private String path;
    /**stores data read from input file*/
    private Vector data;
    /**dimensionality*/
    private int dim;
    /**number of data objects*/
    private int numDataObjects;
    
    
    /** Creates a new instance of Class IO.
     * @param dir directory
     * @param dim dimensionality
     */
    public IO(String dir, int dim) {
        this.path = dir;
        this.dim = dim;
        this.data = new Vector();
    }
    
    public IO(String dir, int dim, int numDataObjects) {
        this.path = dir;
        this.dim = dim;
        this.numDataObjects = numDataObjects;
        this.data = new Vector();
    }
    
    public IO(String dir) {
        this.path = dir;
        this.data = new Vector();
    }
    
    public IO(int dim){
        this.dim = dim;
    }
    
    public IO(){
    }
    
    public void arffToText(String dir){
        //structure of file: attributes, classIndex
        try {
            Instances inst = new Instances(new BufferedReader(new FileReader(dir)));
            inst.setClassIndex(inst.numAttributes()-1);
            File target = new File(dir.substring(0, dir.length()-5) + ".txt");
            FileOutputStream output = new FileOutputStream(target);
            PrintStream writer = new PrintStream(output);
            for(int i = 0; i < inst.numInstances(); i++){
                //print data: classID
                writer.print((int)inst.instance(i).value(inst.numAttributes()-1) + " ");
                //coordinates
                for(int j = 0; j < inst.numAttributes()-2; j++)
                    writer.print(inst.instance(i).value(j) + " ");
                writer.println(inst.instance(i).value(inst.numAttributes()-2));
            }
        } catch (FileNotFoundException ex) {
            ex.printStackTrace();
        } catch (IOException ex) {
            ex.printStackTrace();
        }
    }
    
     public void writeLof(DataObject[] data, double[] lof){
         try{
            File target = new File("LOF_nba_30_"+ data.length + ".txt");
            FileOutputStream output = new FileOutputStream(target);
            PrintStream writer = new PrintStream(output);
            for (int a = 0; a < data.length; a++){
                writer.print(data[a].classID + " ");
                writer.print(data[a].clusterID + " ");
                writer.print(lof[a] + " ");
                for (int b = 0; b < data[a].getCoord().length; b++){
                    writer.print(data[a].getCoord()[b]);
                    writer.print(" ");
                }
                writer.println();
            }
        } catch (FileNotFoundException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }        
    }
    
    
    public void writeArff(DataObject[] d, int numCl, String dir){
        int dim = d[0].coord.length;
        try{
            File target = new File(dir + ".arff");
            FileOutputStream output = new FileOutputStream(target);
            PrintStream writer = new PrintStream(output);
            writer.println("@relation test");
            writer.println();
            for(int i = 0; i < dim; i++)
                writer.println("@attribute " + "a" + i + " real");
            //writer.print("@attribute class {");
//            for(int i = 0; i < numCl-1; i++)
//                writer.print((i+1) + ",");
//            writer.print(numCl);
            //writer.println("}");
            writer.println();
            writer.println("@data");
            //int numV = 5;
            //for(int k = 0; k < numV; k++){
            for (int i = 0; i < d.length; i++){
                for (int j = 0; j < dim-1; j++)
                    writer.print(d[i].getCoord()[j]+ ",");
                writer.println(d[i].getCoord()[dim-1]);
                //writer.println(d[i].classID);
            }
            //}
        } catch (FileNotFoundException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }
    }
    
    
    
    /**Reads ascii-File. Stores data into Array of data Objects DataObject[].*/
    public DataObject[] readFile(){
        
        System.out.println("reading file " + this.path);
        try{
            File source = new File(this.path);
            StreamTokenizer st1 = new StreamTokenizer(new BufferedReader(new FileReader(path)));
            st1.parseNumbers();
            st1.eolIsSignificant(false);
            while( st1.ttype != StreamTokenizer.TT_EOF){  //Daten aus File in Vector einlesen
                st1.nextToken();
                double nextNum = (double)(st1.nval);
                Double nextElt = new Double(nextNum) ;
                this.data.addElement(nextElt);
            }
        } catch (IOException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }
        
        DataObject[] store = new DataObject[this.data.size()/(this.dim)];
        this.numDataObjects = store.length;
        System.out.println(store.length +  " data objects.");
        for (int i = 0; i < store.length; i++){
            double[] ko = new double[dim];           //coordinatenarray anlegen
            for (int j = 0; j < (dim); j++){              //coordinatenarray mit Daten aus Vektor f�llen
                ko[j] = ((Double)data.elementAt(0)).doubleValue();
                data.remove(0);
            }
            DataObject create= new DataObject(ko, i);
            store[i] = create;                 //und in store schreiben
            
        }
        return store;
    }
    
    /**Reads classified data. Stores data into Array of data Objects DataObject[].*/
    public DataObject[] readClassClusterID(){
        
        System.out.println("reading file...");
        try{
            File source = new File(this.path);
            StreamTokenizer st1 = new StreamTokenizer(new BufferedReader(new FileReader(path)));
            st1.parseNumbers();
            st1.eolIsSignificant(false);
            while( st1.ttype != StreamTokenizer.TT_EOF){  //Daten aus File in Vector einlesen
                st1.nextToken();
                if(st1.ttype == st1.TT_NUMBER){
                    double nextNum = (double)(st1.nval);
                    Double nextElt = new Double(nextNum) ;
                    this.data.addElement(nextElt);
                }
            }
        } catch (IOException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }
        
        DataObject[] store = new DataObject[this.data.size()/(this.dim+2)];
        this.numDataObjects = store.length;
        System.out.println(store.length +  " data objects.");
        for (int i = 0; i < store.length; i++){
            int classID = ((Double)data.elementAt(0)).intValue();
            data.remove(0);
            int clusterID = ((Double)data.elementAt(0)).intValue();
            data.remove(0);
            double[] ko = new double[dim];           //coordinatenarray anlegen
            for (int j = 0; j < (dim); j++){              //coordinatenarray mit Daten aus Vektor f�llen
                ko[j] = ((Double)data.elementAt(0)).doubleValue();
                data.remove(0);
            }
            DataObject create= new DataObject(ko, i, classID, clusterID);
            store[i] = create;                 //und in store schreiben
            
        }
        return store;
    }
    
    public DataObject[] readClassFile(){
        
        System.out.println("reading file...");
        try{
            File source = new File(this.path);
            StreamTokenizer st1 = new StreamTokenizer(new BufferedReader(new FileReader(path)));
            st1.parseNumbers();
            st1.eolIsSignificant(false);
            while( st1.ttype != StreamTokenizer.TT_EOF){  //Daten aus File in Vector einlesen
                st1.nextToken();
                if(st1.ttype == st1.TT_NUMBER){
                    double nextNum = (double)(st1.nval);
                    Double nextElt = new Double(nextNum) ;
                    this.data.addElement(nextElt);
                }
            }
        } catch (IOException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }
        
        DataObject[] store = new DataObject[this.data.size()/(this.dim+1)];
        this.numDataObjects = store.length;
        System.out.println(store.length +  " data objects.");
        for (int i = 0; i < store.length; i++){
            int classID = ((Double)data.elementAt(0)).intValue();
            data.remove(0);
            double[] ko = new double[dim];           //coordinatenarray anlegen
            for (int j = 0; j < (dim); j++){              //coordinatenarray mit Daten aus Vektor f�llen
                ko[j] = ((Double)data.elementAt(0)).doubleValue();
                data.remove(0);
            }
            DataObject create= new DataObject(ko, i, classID);
            store[i] = create;                 //und in store schreiben
            
        }
        return store;
    }
    
    public DataObject[] readClusteredFile(){
        System.out.println("reading file...");
        try{
            File source = new File(this.path);
            StreamTokenizer st1 = new StreamTokenizer(new BufferedReader(new FileReader(path)));
            st1.parseNumbers();
            st1.eolIsSignificant(false);
            while( st1.ttype != StreamTokenizer.TT_EOF){  //Daten aus File in Vector einlesen
                st1.nextToken();
                if(st1.ttype == st1.TT_NUMBER){
                    double nextNum = (double)(st1.nval);
                    Double nextElt = new Double(nextNum) ;
                    this.data.addElement(nextElt);
                }
            }
        } catch (IOException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }
        
        DataObject[] store = new DataObject[this.data.size()/(this.dim+1)];
        this.numDataObjects = store.length;
        System.out.println(store.length +  " data objects.");
        for (int i = 0; i < store.length; i++){
            int clusterID = ((Double)data.elementAt(0)).intValue();
            data.remove(0);
            double[] ko = new double[dim];           //coordinatenarray anlegen
            for (int j = 0; j < (dim); j++){              //coordinatenarray mit Daten aus Vektor f�llen
                ko[j] = ((Double)data.elementAt(0)).doubleValue();
                data.remove(0);
            }
            DataObject create= new DataObject(ko, i);
            create.clusterID = clusterID;
            store[i] = create;                 //und in store schreiben
        }
        return store;
    }
    
    
    //read Cat retina Data: 513d
    public DataObject[] readRetina(){
        System.out.println("reading file...");
        int countDim = 0;
        int countObj = 0;
        DataObject[] data = new DataObject[5000];
        Vector aktFeatures = new Vector();
        try{
            File source = new File(this.path);
            StreamTokenizer st1 = new StreamTokenizer(new BufferedReader(new FileReader(path)));
            st1.parseNumbers();
            st1.eolIsSignificant(true);
            while(st1.ttype != StreamTokenizer.TT_EOF && countObj < 5000){
                while(countDim < 513){
                    st1.nextToken();
                    double nextNum = (double)(st1.nval);
                    Double nextElt = new Double(nextNum) ;
                    aktFeatures.addElement(nextElt);
                    countDim++;
                }
                double[] coord = new double[countDim];
                for(int i = 0; i < coord.length; i++)
                    coord[i] = ((Double)aktFeatures.elementAt(i)).doubleValue();
                DataObject akt = new DataObject(coord, countObj);
                data[countObj] = akt;
                countObj++;
                aktFeatures.removeAllElements();
                countDim = 0;
            }
        } catch (IOException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
            
        }
        return data;
    }
    
    
    public DataObject[] readRetinaClass(){
        int imageSize = 96; //96 tiles per image
        int[] labels = readClassLabels("retinaLabels.txt");
        DataObject[] uData = readFile();  //unlabeled data
        int counter = 0; //counting tiles
        int labelIndex = 0;
        for(int i = 0; i < uData.length; i++){
            uData[i].classID = labels[labelIndex];
            counter++;
            if(counter > imageSize-1){
                labelIndex++;
                counter = 0;
            }
        }
        System.out.println(uData.length + " data objects");
        printData(uData);
        return uData;
    }
    
    
    private int[] readClassLabels(String lpath){
        System.out.println("reading class labels...");
        Vector ldata = new Vector();
        try{
            File source = new File(lpath);
            StreamTokenizer st1 = new StreamTokenizer(new BufferedReader(new FileReader(lpath)));
            st1.parseNumbers();
            st1.eolIsSignificant(false);
            while( st1.ttype != StreamTokenizer.TT_EOF){  //Daten aus File in Vector einlesen
                st1.nextToken();
                //System.out.println(st1.ttype);
                if(st1.ttype == -2){
                    int nextNum = (int)(st1.nval);
                    Integer nextElt = new Integer(nextNum) ;
                    ldata.addElement(nextElt);
                }
            }
        } catch (IOException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }
        
        int[] labels = new int[ldata.size()];
        System.out.println(labels.length +  " labels read.");
        for (int i = 0; i < labels.length; i++){
            labels[i] = ((Integer)ldata.elementAt(i)).intValue();
            //data.remove(i);
        }
        
        return labels;
    }
    
    //handwritten digits
    public DataObject[] readDigits(){
        int[] labels = readClassLabels("labels.txt");
        int numObj = 1000;
        System.out.println("numObj " + numObj);
        System.out.println("reading file...");
        int countDim = 0;
        int countObj = 0;
        DataObject[] data = new DataObject[numObj];
        Vector aktFeatures = new Vector();
        try{
            File source = new File(this.path);
            StreamTokenizer st1 = new StreamTokenizer(new BufferedReader(new FileReader(path)));
            st1.parseNumbers();
            st1.eolIsSignificant(true);
            while(st1.ttype != StreamTokenizer.TT_EOF && countObj < numObj){
                while(countDim < dim){
                    st1.nextToken();
                    double nextNum = (double)(st1.nval);
                    Double nextElt = new Double(nextNum) ;
                    aktFeatures.addElement(nextElt);
                    countDim++;
                }
                double[] coord = new double[countDim];
                for(int i = 0; i < coord.length; i++)
                    coord[i] = ((Double)aktFeatures.elementAt(i)).doubleValue();
                DataObject akt = new DataObject(coord, countObj, labels[countObj]+1);
                data[countObj] = akt;
                countObj++;
                aktFeatures.removeAllElements();
                countDim = 0;
            }
        } catch (IOException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
            
        }
        return data;
    }
    
    //handwritten digits
    public DataObject[] readPenDigits(){
        System.out.println("reading file...");
        int countDim = 0;
        int countObj = 0;
        DataObject[] data = new DataObject[numDataObjects];
        Vector aktFeatures = new Vector();
        try{
            File source = new File(this.path);
            StreamTokenizer st1 = new StreamTokenizer(new BufferedReader(new FileReader(path)));
            st1.parseNumbers();
            st1.eolIsSignificant(true);
            while(st1.ttype != StreamTokenizer.TT_EOF && countObj < numDataObjects){
                while(countDim < dim+1){
                    st1.nextToken();
                    double nextNum = (double)(st1.nval);
                    Double nextElt = new Double(nextNum) ;
                    aktFeatures.addElement(nextElt);
                    countDim++;
                }
                int label = ((Double)aktFeatures.elementAt(0)).intValue();
                double[] coord = new double[dim];
                for(int i = 0; i < coord.length; i++)
                    coord[i] = ((Double)aktFeatures.elementAt(i+1)).doubleValue();
                DataObject akt = new DataObject(coord, countObj, label+1);
                data[countObj] = akt;
                countObj++;
                aktFeatures.removeAllElements();
                countDim = 0;
            }
        } catch (IOException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
            
        }
        return data;
    }
    
    public IcaCluster[] readResult(int numCl, int dim, String rPath){
        IcaCluster[] res = new IcaCluster[numCl];
        for(int i = 0; i < res.length; i++){
            res[i] = new IcaCluster();
            res[i].realCenter = new double[dim];
            res[i].realRotation = new double[dim][dim];
            res[i].beta = new double[dim];
        }
        //format of .txt. file
        //clusterID: size 1
        //mean: size dim
        //rotation: size dim*dim
        //beta: size dim
        //numObj: size 1
        //read file in vector
        Vector<Double> v = new Vector<Double>();
        System.out.println("READ RESULT " + " numCl: " + numCl + " dim: " + dim + " rPath: " + rPath);
        try{
            File source = new File(rPath);
            StreamTokenizer st1 = new StreamTokenizer(new BufferedReader(new FileReader(rPath)));
            st1.parseNumbers();
            st1.eolIsSignificant(false);
            while( st1.ttype != StreamTokenizer.TT_EOF){
                st1.nextToken();
                if(st1.ttype == st1.TT_NUMBER){
                    double nextNum = (double)(st1.nval);
                    Double nextElt = new Double(nextNum) ;
                    v.addElement(nextElt);
                }
            }
        } catch (IOException e){
            System.err.println("Error: " + e);
        }
        //store data in res
        for(int i = 0; i < res.length; i++){
            res[i].ID = v.elementAt(0).intValue();
            v.remove(0);
            for(int j = 0; j < dim; j++){
                res[i].realCenter[j] = v.elementAt(0);
                v.remove(0);
            }
            for(int j = 0; j < dim; j++){
                for(int k = 0; k < dim; k++){
                    res[i].realRotation[j][k] = v.elementAt(0);
                    if(res[i].realRotation[j][k] != 0.0)
                        res[i].realRotation[j][k] = 1.0/res[i].realRotation[j][k];
                    v.remove(0);
                }
            }
            for(int j = 0; j < dim; j++){
                res[i].beta[j] = v.elementAt(0);
                v.remove(0);
            }
            res[i].numObj = v.elementAt(0).intValue();
            v.remove(0);
        }
        return res;
    }
    
    /**creates a file named path + "_" + size + ".txt" containing size samplepoints of DB.
     * @param db DB
     * @param size the number of sample points
     */
    public void createSampleDat(DB db, int size){
        
        try{
            File target = new File(this.path + "_" + size + ".txt");
            FileOutputStream output = new FileOutputStream(target);
            PrintStream writer = new PrintStream(output);
            Random r = new Random();
            DataObject[] data = db.getData();
            for (int a = 0; a < size; a++){
                int s = r.nextInt(db.store.length);
                //writer.print(data[s].classID+ " ");
                //writer.print(" ");
                for (int b = 0; b < data[s].getCoord().length; b++){
                    writer.print(data[s].getCoord()[b]+" ");
                    //writer.print(" ");
                }
                writer.println();
            }
        } catch (FileNotFoundException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }
    }
    
    
    public DataObject[] sample(DB db, int size){
        DataObject[] sample = new DataObject[size];
        Random r = new Random(10);
        for (int a = 0; a < size; a++){
            int s = r.nextInt(db.store.length);
            sample[a] = db.store[s];
        }
        return sample;
    }
    
    
    public void createLaplaceData(int size, int dim, double a, double b){
        Random r = new Random();
        DataObject[] obj = new DataObject[size];
        for(int i = 0; i < size; i++){
            double[] co = new double[dim];
            for(int j = 0; j < dim; j++)
                co[j] = nextLaplacian(a, b, r);
            DataObject n = new DataObject(co, i);
            obj[i] = n;
        }
        printData(obj);
    }
    
    
    /**
     * Create a dataset with one laplacian cluster and noise
     * @param size size of cluster
     * @param noiseSize number of noise objects
     * @param dim dimensionality
     * @param a
     * @param b
     */
    //IN WORK
    public void createLaplaceDataNoise(int size, int noiseSize, int dim, double a, double b){
        Random r = new Random(1);
        DataObject[] obj = new DataObject[size];
        //min and max of cluster points
        double[] min = new double[dim];
        for(int i = 0; i < dim; i++)
            min[i] = Double.MAX_VALUE;
        double[] max = new double[dim];
        for(int i = 0; i < dim; i++)
            max[i] = -Double.MAX_VALUE;
        //generate cluster points
        for(int i = 0; i < size; i++){
            double[] co = new double[dim];
            for(int j = 0; j < dim; j++){
                co[j] = nextLaplacian(a, b, r);
                if(co[j] < min[j])
                    min[j] = co[j];
                if(co[j] > max[j])
                    max[j] = co[j];
            }
            DataObject n = new DataObject(co, i);
            n.classID = 0;
            obj[i] = n;
        }
        //scale clusterpoints to 20% of attribute range
        
        //scale cluster points in [0..1], for coodinates > 0
        DataObject[] objNew = scaleData(obj, 5);
        //fill up with uniform noise
        DataObject[] res = new DataObject[size + noiseSize];
        for(int i = 0; i < res.length; i++)
            if(i < size)
                res[i] = objNew[i];
            else{
            double[] co = new double[dim];
            for(int j = 0; j < dim; j++){
                double rd = r.nextDouble();
                if(rd < 0.001)
                    rd = 0.001;
                co[j] = rd;
            }
            DataObject n = new DataObject(co, i);
            n.classID = 1;
            res[i] = n;
            }
        printData(res);
    }
    
    public void rotatedLaplacian(int size, int dim, double a, double b){
        double[] rotVector = new double[dim];
        Random r = new Random();
        DataObject[] obj = new DataObject[size];
        rotVector[0] = 0.5;
        rotVector[1] = 0.1;
        for(int i = 0; i < size; i++){
            double[] co = new double[dim];
            for(int j = 0; j < dim; j++){
                co[j] = nextLaplacian(a, b, r); //* rotVector[j];
                if(j == 1 && co[j] > 50.0){
                    if(co[j] > 50)
                        co[0]+= (co[j]-50.0);
                    else
                        co[0]-= (50.0 - co[j]);
                }
                // if(j==1 && co[j]<50){
                //  if(co[0]< 50)
                // co[0]-= (co[j]-50.0);
                
                //}
            }
            DataObject n = new DataObject(co, i);
            obj[i] = n;
        }
        printData(obj);
    }
    
    //classID, clusterID, Koordinaten
    public void printResult(DataObject[] data){
        try{
            File target = new File("RESULT_"+ data.length + ".txt");
            FileOutputStream output = new FileOutputStream(target);
            PrintStream writer = new PrintStream(output);
            for (int a = 0; a < data.length; a++){
                writer.print(data[a].classID + " ");
                writer.print(data[a].clusterID + " ");
                for (int b = 0; b < data[a].getCoord().length; b++){
                    writer.print(data[a].getCoord()[b]);
                    writer.print(" ");
                }
                writer.println();
            }
        } catch (FileNotFoundException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }
    }
    
    public void printClusters(DataObject[] data, int id1){
        try{
            File target = new File("CLUSTER" + id1 + ".txt");
            FileOutputStream output = new FileOutputStream(target);
            PrintStream writer = new PrintStream(output);
            for (int a = 0; a < data.length; a++){
                if(data[a].clusterID == id1){
                    for (int b = 0; b < data[a].getCoord().length; b++){
                        writer.print(data[a].getCoord()[b]);
                        writer.print(" ");
                    }
                    writer.println();
                }
            }
        } catch (FileNotFoundException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }
        
    }
    
    public void printData(DataObject[] data){
        try{
            File target = new File("NEWDATA_"+ data.length + ".txt");
            FileOutputStream output = new FileOutputStream(target);
            PrintStream writer = new PrintStream(output);
            for (int a = 0; a < data.length; a++){
                //writer.print(data[a].classID + " ");
                for (int b = 0; b < data[a].getCoord().length; b++){
                    writer.print(data[a].getCoord()[b]);
                    writer.print(" ");
                }
                writer.println();
            }
        } catch (FileNotFoundException e){            //Es gibt kein solches File
            System.err.println("Error: " + e);
        }
    }
    
    
    public static double nextLaplacian(double a, double b, Random r) {
        double y = r.nextDouble() ;
        if (y<0.5)  return a + b * Math.log(2*y) ;
        else        return a - b * Math.log(2-2*y) ;
    }
    
    //IN WORK
    private DataObject[] scaleData(DataObject[] d, int range){
        for(int i = 0; i < dim; i++){
            double min = Double.MAX_VALUE;
            for(int j = 0; j < d.length; j++)
                if(d[j].coord[i] < min)
                    min = d[j].coord[i];
            for(int j = 0; j < d.length; j++)
                if(min < 0)
                    d[j].coord[i]+= Math.abs(min);
                else
                    d[j].coord[i]-=min;
            double max = 0.0;
            for(int j = 0; j < d.length; j++)
                if(d[j].coord[i] > max)
                    max = d[j].coord[i];
            for(int j = 0; j < d.length; j++)
                d[j].coord[i] = d[j].coord[i]/max;
            //System.out.println("Koordinate " + i + " max: " + max + " min: " + min);
        }
        return d;
    }
}




