/*
 * GaussianFilter.java
 *
 * Created on 11. September 2006, 11:26
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

package noise;
import compr.DataObject;
import utils.Statistics;
import visu.VisuSimple;
import io.IO;

/**
 *
 * @author boehm
 */
public class GaussianFilter {
    DataObject[] d;
    int numObj;
    int dim;
    double[] mean; //mean of cluster and noise;
    double[] clusterVariance;
    double[] noiseVariance;
    
    
    
    /**
     * Creates a new instance of GaussianFilter 
     */
    public GaussianFilter(DataObject[] d) {
        this.d = d;
        numObj = d.length;
        dim = d[0].coord.length;
        mean = new double[dim];
        clusterVariance = new double[dim];
        noiseVariance = new double[dim];
    }
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        int dim = 2;
        //Verzeichnis Ica-test
         //IO ea = new IO("icaQuadratNoise5.txt", dim);
        //IO ea = new IO("2cluster.txt", dim); //ica uniform am besten.
        //IO ea = new IO("icaTest1.txt", dim); //am besten ist laplacian ica?
        //IO ea = new IO("rotatedLpl.txt", dim); //gut: am besten ist laplacian ica
        //draussen
        //String dir = "cor1noise10.txt"; //noise wird nicht ausgesondert
        //String dir = "cor1asymmout.txt"; //ok 
        //String dir = "cor1asymmoutn.txt"; //noise nicht ausreichend ausgefiltert
        //String dir = "clust50.txt"; // ok
        //String dir = "cor1.txt"; //ok
        //String dir = "cor1g1n1.txt";  //noise wird nicht ausgesondert
        //String dir = "cor1noise.txt"; //ok
        //String dir = "cor1vielnoise.txt"; //ok: unmglich
        //String dir = "cor20noise5.txt"; //ok
        //String dir = "g100.txt"; //ok
        //String dir = "g1000.txt"; //ok
        //String dir = "gauss2000.txt"; //ok
        //String dir = "laplace200.txt"; //? laplace-Verteilung aufnehmen
        String dir = "noise500.txt"; //ok
        //String dir = "outlierTest.txt"; // ok
        //String dir = "t2.txt"; // ok
        IO ea = new IO(dir, dim);
        DataObject[] d = ea.readFile();
        for(int i = 0; i < d.length; i++)
            d[i].clusterID = -1;
        GaussianFilter f = new GaussianFilter(d);
        f.run();
    }
    
    public void run(){
        init();
        boolean clusterChanged = true;
        int iter = 0;
        while(clusterChanged){
            System.out.println("iter " + iter);
            iter++;
            clusterChanged = assignPoints();
            update();
        }
        VisuSimple visu = new VisuSimple(d);
        visu.setSize(600,600);
        visu.setLocation(100,0);
        visu.setVisible(true);
    }
    
    
    public void init(){
        Statistics s = new Statistics(d.length, d[0].coord.length);
        double[] m = s.mean(d);
        double[] sigma = new double[dim];
        double addVar = 0.0;
        for(int i = 0; i < dim; i++){
            sigma[i] = s.sigma(d, m, i);
            addVar += sigma[i];
        }
        //initialize mean and cluster- and noise variances.
        mean = s.median(d);
        double varCluster = addVar/(double)dim;
        double varNoise = addVar * (double)dim;
        for(int i = 0; i < dim; i++){
            clusterVariance[i] = varCluster;
            noiseVariance[i] = varNoise;
        }
    }
    
    
    /**
     * ClusterID 1: cluster, 0: noise.
     */
    private boolean assignPoints(){
        boolean clusterChanged = false;
        for(int i = 0; i < numObj; i++){
            double clusterCost = 0.0;
            for(int j = 0; j < dim; j++)
                clusterCost += gaussianCost(d[i].coord[j], clusterVariance[j], mean[j]);
            double noiseCost = 0.0;
            for(int j = 0; j < dim; j++)
                noiseCost += gaussianCost(d[i].coord[j], noiseVariance[j], mean[j]);
            if(clusterCost < noiseCost){
                if(d[i].clusterID != 1){
                    d[i].clusterID = 1;
                    clusterChanged = true;
                }
            } else{
                if(d[i].clusterID != 0){
                    d[i].clusterID = 0;
                    clusterChanged = true;
                }
            }
        }
        return clusterChanged;
    }
    
    
    /**
     * update mean of cluster, noise cluster has the same mean as the cluster. Update cluster and noise variances.
     */
    private void update(){
        //update mean as mean of the cluster points (clusterID 1)
        double[] meanNew = new double[dim];
        int counter = 0;
        for(int i = 0; i < d.length; i++){
            if(d[i].clusterID == 1){
                for(int j = 0; j < dim; j++){
                    meanNew[j] += d[i].coord[j];
                }
                counter++;
            }
        }
        for(int j = 0; j < dim; j++)
            meanNew[j]/=(double)counter;
        mean = meanNew;
        //update variances
        //cluster points
        DataObject[] clusterPts = new DataObject[counter];
        DataObject[] noisePts = new DataObject[numObj-counter]; //Fehlt: Behandlung: eines von beidem wird leer.
        int clusterCounter = 0;
        int noiseCounter = 0;
        for(int i = 0; i < numObj; i++){
            if(d[i].clusterID == 1){
                clusterPts[clusterCounter] = d[i];
                clusterCounter++;
            } else{
                noisePts[noiseCounter] = d[i];
                noiseCounter++;
            }
        }
        double[] clusterVarianceNew = new double[dim];
        double[] noiseVarianceNew = new double[dim];
        Statistics sCluster = new Statistics(clusterCounter, dim);
        Statistics sNoise = new Statistics(noiseCounter, dim);
        for(int i = 0; i < dim; i++)
            clusterVarianceNew[i] = sCluster.sigma(clusterPts, mean, i);
        for(int i = 0; i < dim; i++)
            noiseVarianceNew[i] = sNoise.sigma(noisePts, mean, i);
    }
    
    
    /**
     * gaussian cost of one coordinate.
     */
    private double gaussianCost(double coord, double sigma, double mean){
        double cost = 0.0;
        double factor = 1/(Math.sqrt(2 * Math.PI * sigma));
        return -Math.log(Math.exp(-(coord - mean)*(coord - mean)/(2 * sigma)) * factor)/Math.log(2.0);
    }
    
}
