package oci;
/*
 * GenGauss.java
 *
 * Created on 11. Mai 2007, 15:39
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

/**
 *
 * @author boehm
 */

import Jama.*;
import umontreal.iro.lecuyer.util.Num ;
import java.util.Random;


public class GenGauss {
    
    boolean verbose = false ;
    Random montecarlo = new Random(17);
    
    /** Creates a new instance of GenGauss */
    public GenGauss() {
    }
    
    double pow(double x, double y) {
        return Math. pow(x, y) ;
    }
    
    double exp(double x) {
        return Math.exp(x) ;
    }
    
    double fabs(double x) {
        return Math.abs(x) ;
    }
    
    double abs(int i, double x) {
        return Math.signum(x) ;
    }
    
    double signum(int i, double x) {
        return 0.0 ;
    }
    
    double lgamma(double x) {
        return Num.lnGamma(x) ;
    }
    
    double Psi(double x) {
        return Num.digamma(x) ;
    }
    
    double Psi(int i, double x) {
        if (i==0)
            return Num.digamma(x) ;
        else if (i==1)
            return Num.trigamma(x) ;
        else System.err.println("HIGHER ORDER DERIVIATIVE REQUIRED") ;
        return 0.0 ;
    }
    
    double log(double x) {
        return Math.log(x) ;
    }
    
    
    double dcm(double x,double mu,double sigma,double beta) {
        double t13;
        double t14;
        double t11;
        double t12;
        double t9;
        double t8;
        
        t8 = 1/(1+beta);
        t9 = pow(exp(lgamma(0.15E1+0.15E1*beta))/exp(lgamma(0.5+0.5*beta)),t8);
        t11 = 1/sigma;
        t12 = (x-mu)*t11;
        t13 = fabs(t12);
        t14 = pow(t13,2*t8);
        return(-2*t9*t14*t8*abs(1,t12)*t11/t13);
    }
    
    public double integral(double [] lb, double [] ub, double [] mu,
            double [][] sigma, double [] beta, int steps) {
        double det = 1.0/(new Matrix(sigma)).det() ;
        if(Double.isInfinite(Math.abs(det)))
            System.err.println("det infty");
        //det -Infty bei cluster 1, 2 von em
        double sum = 0.0 ;
        int dim = lb.length ;
        double [] sample = new double [dim] ;
        for (int i=0 ; i<steps ; i++) {
            for (int j=0 ; j<dim ; j++)
                sample [j] =
                        montecarlo.nextDouble()*(ub[j]-lb[j])+lb[j] ;
            double prod = 1.0 ;
            for (int j1=0 ; j1<dim ; j1++) {
                double sampleR = 0.0 ;
                for (int j2=0 ; j2<dim ; j2++)
                    //sampleR += sigma[j1][j2] * (sample[j2]-mu[j2]) ; //nicht transponiert
                    sampleR += sigma[j2][j1] * (sample[j2]-mu[j2]) ;
                prod *= gg(sampleR, 0.0, 1.0, beta[j1]) ;
                //if(prod < 0.0)
                //System.err.println("prod: " + prod);
            }
            //System.out.println("det: " + det);
            //if(Math.abs(det) > 0.0)
            sum += prod /Math.abs(det);
            //else
            //sum += prod;
        }
        double prod = 1.0 ;
        for (int j=0 ; j<dim ; j++)
            prod *= (ub[j]-lb[j]) ;
        return sum / steps * prod ;
    }
    
    public double kullbackLeiblerDivergence(double [] lb, double [] ub,
            IcaCluster[] m1, IcaCluster[] m2, int steps, int numObj) {
        double[] det_m1 = new double[m1.length];
        double[] det_m2 = new double[m2.length];
        for(int i = 0; i < det_m1.length; i++)
            det_m1[i] = 1.0/(new Matrix(m1[i].realRotation)).det() ;
        for(int i = 0; i < det_m2.length; i++)
            det_m2[i] = 1.0/(new Matrix(m2[i].realRotation)).det() ;
        double sum = 0.0 ;
        int dim = lb.length ;
        double [] sample = new double [dim] ;
        for (int i=0 ; i<steps ; i++) {
            for (int j=0 ; j<dim ; j++)
                sample [j] = montecarlo.nextDouble()*(ub[j]-lb[j])+lb[j] ;
            double s1 = 0.0 ;
            double s2 = 0.0 ;
            for (int j=0 ; j<m1.length ; j++) {
                double prod = 1.0 ;
                for (int j1=0 ; j1<dim ; j1++) {
                    double sampleR = 0.0 ;
                    for (int j2=0 ; j2<dim ; j2++)
                        sampleR += m1[j].realRotation [j1][j2] *
                                (sample[j2]-m1[j].realCenter[j2]) ;
                    prod *= gg(sampleR, 0.0, 1.0, m1[j].beta[j1]) ;
                }
                s1 += prod / Math.abs(det_m1[j]) * m1[j].numObj/numObj ;
            }
            for (int j=0 ; j<m2.length ; j++) {
                double prod = 1.0 ;
                for (int j1=0 ; j1<dim ; j1++) {
                    double sampleR = 0.0 ;
                    for (int j2=0 ; j2<dim ; j2++)
                        sampleR += m2[j].realRotation [j1][j2] *
                                (sample[j2]-m2[j].realCenter[j2]) ;
                    prod *= gg(sampleR, 0.0, 1.0, m2[j].beta[j1]) ;
                }
                s2 += prod / Math.abs(det_m2[j]) * m2[j].numObj/numObj ;
            }
            sum += s1*Math.log(s1/s2) ;
        }
        double prod = 1.0 ;
        for (int j=0 ; j<dim ; j++)
            prod *= (ub[j]-lb[j]) ;
        
        //return sum * prod / steps ;
        return sum / steps * prod ; //wie in integral
        
    }
    
    public double gaussEvaluation(double[] lb, double[] ub, int steps){
        //String dir = "3cluster2dnoise.txt";
//        int numCl = 4;
//        double[] mu0 = {46.077, 25.9095};
//        double[][]sigma0 = {{1/10.9679,0},{0, 1/10.9313}};
//        double[]beta0 = {0,0};
//        double prior0 = 0.2346;
//        double[] mu1 = {30.0432, 24.9414};
//        double[][]sigma1 = {{1/2.587,0},{0, 1/2.2915}};
//        double[]beta1 = {0,0};
//        double prior1 = 0.2719;
//        double[] mu2 = {50.0154, 15.0399};
//        double[][]sigma2 = {{1/0.5874,0},{0, 1/0.716}};
//        double[]beta2 = {0,0};
//        double prior2 = 0.2228;
//        double[] mu3 = {50.0991, 40.1038};
//        double[][]sigma3 = {{1/2.0612,0},{0, 1/1.8977}};
//        double[]beta3 = {0,0};
//        double prior3 = 0.2707;
        
        //String dir = "3cluster2dnoiseErzeugt.txt";
        int numCl = 4;
        double[] mu0 = {49.9623, 14.9371};
        double[][]sigma0 = {{1/0.5435,0},{0, 1/0.6968}};
        double[]beta0 = {0,0};
        double prior0 = 0.217;
        
        double[] mu1 = {49.9462, 39.9725};
        double[][]sigma1 = {{1/2.1208, 0},{0, 1/1.8845}};
        double[]beta1 = {0,0};
        double prior1 = 0.2695;
        
        double[] mu2 = {30.1184, 25.122};
        double[][]sigma2 = {{1/2.3823,0},{0, 1/2.1579}};
        double[]beta2 = {0,0};
        double prior2 = 0.2581;
        
        double[] mu3 = {44.1112, 22.9276};
        double[][]sigma3 = {{1/20.7131,0},{0, 1/20.8172}};
        double[]beta3 = {0,0};
        double prior3 = 0.2553;
        
        double result = 0.0 ;
        result += prior0 * integral(lb,ub,mu0,sigma0,beta0,steps) ;
        result += prior1 * integral(lb,ub,mu1,sigma1,beta1,steps) ;
        result += prior2 * integral(lb,ub,mu2,sigma2,beta2,steps) ;
        result += prior3 * integral(lb,ub,mu3,sigma3,beta3,steps) ;
        return result ;
    }
    
    public double klEvaluation(IcaCluster[] res, double[] lb, double[] ub, int steps, int numObj){
        int dim = 2;
        IcaCluster[] gen = new IcaCluster[4];
        //----------------------------erzeugt--------------------------------------------
        //cluster 8
        gen[0] = new IcaCluster();
        double[] beta0 =  {-0.85,3.8};
        gen[0].beta = beta0;
        double[][] realRotation0 = {{2.5, 2.0},{0.5, 2.0}} ;
        gen[0].realRotation = realRotation0;
        double[] mu0 = {30.0, 25.0} ;
        gen[0].realCenter = mu0;
        gen[0].numObj = 500;
        gen[0].ID = 8;
        
        //cluster 0
        gen[1] = new IcaCluster();
        double[] beta1 = {-0.5,0.5};
        gen[1].beta = beta1;
        double[][] realRotation1 =  {{2.0, 0.0},{0.8, 2.0}} ; ;
        gen[1].realRotation = realRotation1;
        double[] mu1 = {50.0, 40.0};
        gen[1].realCenter = mu1;
        gen[1].numObj = 500;
        gen[1].ID = 0;
        
        //cluster 2
        gen[2] = new IcaCluster();
        double[] beta2 = {4.5,5.3};
        gen[2].beta = beta2;
        double[][] realRotation2 = {{2.0, 1.0},{0.2, 2.0}};
        gen[2].realRotation = realRotation2;
        double[] mu2 = {50.0, 15.0};
        gen[2].realCenter = mu2;
        gen[2].numObj = 500;
        gen[2].ID = 2;
        
        //noise
        gen[3] = new IcaCluster();
        double[] beta3 = {-0.97, -0.97};
        gen[3].beta = beta3;
        double[][] realRotation3 = {{36.43, 0.0},{0.0, 36.43}};
        gen[3].realRotation = realRotation3;
        double[] mu3 = {42.35, 26.60};
        gen[3].realCenter = mu3;
        gen[2].numObj = 50;
        gen[2].ID = 2;
        //---------------------------------erzeugt----------------------------------------
        //---------------------------------gauss--------------------------------------
        //String dir = "3cluster2dnoiseErzeugt.txt";
        IcaCluster[] em = new IcaCluster[4];
        em[0] = new IcaCluster();
        double[] beta_em0 =  {0.0, 0.0};
        em[0].beta = beta_em0;
        double[][] realRotation_em0 =  {{1/0.5435,0},{0, 1/0.6968}};
        em[0].realRotation = realRotation_em0;
        double[] mu_em0 = {49.9623, 14.9371};
        em[0].realCenter = mu_em0;
        em[0].numObj = 390;
        em[0].ID = 2;
        
        em[1] = new IcaCluster();
        double[] beta_em1 =  {0.0, 0.0};
        em[1].beta = beta_em1;
        double[][] realRotation_em1 =  {{1/2.1208, 0},{0, 1/1.8845}};
        em[1].realRotation = realRotation_em1;
        double[] mu_em1 =  {49.9462, 39.9725};
        em[1].realCenter = mu_em1;
        em[1].numObj = 498;
        em[1].ID = 0;
        
        em[2] = new IcaCluster();
        double[] beta_em2 =  {0.0, 0.0};
        em[2].beta = beta_em2;
        double[][] realRotation_em2 = {{1/2.3823,0},{0, 1/2.1579}};
        em[2].realRotation = realRotation_em2;
        double[] mu_em2 =  {30.1184, 25.122};
        em[2].realCenter = mu_em2;
        em[2].numObj = 482;
        em[2].ID = 8;
        
        em[3] = new IcaCluster();
        double[] beta_em3 =  {0.0, 0.0};
        em[3].beta = beta_em3;
        double[][] realRotation_em3 = {{1/20.7131,0},{0, 1/20.8172}};
        em[3].realRotation = realRotation_em3;
        double[] mu_em3 = {44.1112, 22.9276};
        em[3].realCenter = mu_em3;
        em[3].numObj = 430;
        em[3].ID = 1;
        //--------------------------gauss--------------------------------
        //--------------------------RIC----------------------------------
        IcaCluster[] ric = new IcaCluster[5];
        ric[0] = new IcaCluster();
        double[] beta0_ric =  {0.0,0.0};
        ric[0].beta = beta0_ric;
        double[][] realRotation0_ric = {{1.114, -1.35},{0.252, 0.208}} ;
        ric[0].realRotation = realRotation0_ric;
        double[] mu0_ric = {30.0, 25.0} ;
        ric[0].realCenter = mu0_ric;
        ric[0].numObj = 460;
        ric[0].ID = 8;
        
        //cluster 0
        ric[1] = new IcaCluster();
        double[] beta1_ric = {0.0,0.0};
        ric[1].beta = beta1_ric;
        double[][] realRotation1_ric =  {{0.365, -0.5},{0.335, 0.245}} ; ;
        ric[1].realRotation = realRotation1_ric;
        double[] mu1_ric = {50.0, 40.0};
        ric[1].realCenter = mu1_ric;
        ric[1].numObj = 500;
        ric[1].ID = 0;
        
        //cluster 2
        ric[2] = new IcaCluster();
        double[] beta2_ric = {0.0, 0.0};
        ric[2].beta = beta2_ric;
        double[][] realRotation2_ric = {{-2.011, 1.288},{0.763, 1.191}};
        ric[2].realRotation = realRotation2_ric;
        double[] mu2_ric = {50.0, 15.0};
        ric[2].realCenter = mu2_ric;
        ric[2].numObj = 384;
        ric[2].ID = 2;
        
        //cluster 10
        ric[3] = new IcaCluster();
        double[] beta3_ric = {0.0, 0.0};
        ric[3].beta = beta3_ric;
        double[][] realRotation3_ric = {{0.185, -0.193},{0.121, 0.116}};
        ric[3].realRotation = realRotation3_ric;
        double[] mu3_ric = {50.0, 15.0};
        ric[3].realCenter = mu3_ric;
        ric[3].numObj = 132;
        ric[3].ID = 10;
        
        //noise
        ric[4] = new IcaCluster();
        double[] beta4_ric = {0.0, 0.0};
        ric[4].beta = beta4_ric;
        double[][] realRotation4_ric = {{-0.024, -0.034},{0.032, -0.024}};
        ric[4].realRotation = realRotation4_ric;
        double[] mu4_ric = {41.0, 24.8062};
        ric[4].realCenter = mu4_ric;
        ric[4].numObj = 314;
        ric[4].ID = 1;
        //-----------------------------------RIC-------------------------------------
        
        double kld = kullbackLeiblerDivergence(lb, ub, gen, res, steps, numObj);
        //montecarlo = new Random (17) ;
        double kld_em =  kullbackLeiblerDivergence(lb, ub, gen, em, steps, numObj);
        double kld_ric =  kullbackLeiblerDivergence(lb, ub, gen, ric, steps, numObj);
        System.out.println("KL-divergence EM: " + kld_em);
        System.out.println("KL-divergence RIC: " + kld_ric);
        return kld;
        
    }
    
    public void selectivityEstimation(DataObject[] d, IcaCluster[] resRic, int numSamplePoints){
        System.out.println("SELECTIVITY ESTIMATION: numSamplePoints " + numSamplePoints);
        int dim = d[0].coord.length;
        int numObj = d.length;
        double ric = 0.0; //selctivity deviation with ric
        double other = 0.0;//selectivity deviation with other method
        double trueSel = 0.0; //true number of objects in window
        double windowSize = 0.5;
        int steps = 100000 ;
        double[] lb = new double[dim];
        double ub[] = new double[dim];
        Random r = new Random(15);
        //read result file of other method
        IO ea = new IO();
        IcaCluster[] resOther = ea.readResult(2, dim, "EViris2cluster_EM.txt");
        //IcaCluster[] resOther = ea.readResult(5, dim, "EV3cluster2dNoiseErzeugt_EM.txt");
        for(int i = 0; i < numSamplePoints; i++){
            double aktTrue = 0.0;  //true selectivity in window around aktPt
            double aktRic = 0.0; //selectivity of ric around aktPt
            double aktOther = 0.0; //selectivity of other method around aktPt
            int aktPt = r.nextInt(numObj);
            //determine true number of objects in window around aktPt
            for(int j = 0; j < dim; j++){
                lb[j] = d[aktPt].coord[j]-windowSize;
                ub[j] = d[aktPt].coord[j]+windowSize;
            }
            int counter = 0;
            for(int j = 0; j < d.length; j++){
                boolean contained = true;
                for(int k = 0; k < dim; k++){
                    if(d[j].coord[k] > ub[k] || d[j].coord[k] < lb[k])
                        contained = false;
                }
                if(contained)
                    counter++;
            }
            aktTrue = (double)counter/(double)numObj;
            System.out.println("aktTrue: " + aktTrue);
            //determine selectivity of ric
            for(int j = 0; j < resRic.length; j++){
                //System.out.println(resRic[j].ID + " " + integral(lb, ub, resRic[j].realCenter, new Matrix(resRic[j].realRotation).getArrayCopy(), resRic[j].beta, steps));
                aktRic += (double)resRic[j].numObj/(double)numObj * integral(lb, ub, resRic[j].realCenter, new Matrix(resRic[j].realRotation).getArrayCopy(), resRic[j].beta, steps);
            }
            System.out.println("aktRic: " + aktRic);
            //determine selectivity of other method
            for(int j = 0; j < resOther.length; j++){
                //System.out.println(resOther[j].ID + " " + integral(lb, ub, resOther[j].realCenter, new Matrix(resOther[j].realRotation).getArrayCopy(), resOther[j].beta, steps));
                aktOther += (double)resOther[j].numObj/(double)numObj * integral(lb, ub, resOther[j].realCenter, new Matrix(resOther[j].realRotation).getArrayCopy(), resOther[j].beta, steps);
            }
            System.out.println("aktOther: " + aktOther);
            //determine sum of true selectivity
            trueSel += aktTrue;
            //determe deviations
            ric += Math.abs(aktTrue - aktRic);
            other += Math.abs(aktTrue - aktOther);
        }
        //print result
        System.out.println("sum of selectivity: " + trueSel);
        System.out.println("sum of deviations ric: " + ric);
        System.out.println("sum of deviations other: " + other);
    }
    
    
    
    
    
    double dcs(double x,double mu,double sigma,double beta) {
        double t8;
        double t20;
        double t10;
        double t15;
        double t4;
        double t9;
        double t3;
        double t19;
        double t18;
        double t16;
        double t5;
        double t2;
        double t22;
        double t25;
        double t41;
        double t44;
        
        t2 = exp(lgamma(0.15E1+0.15E1*beta));
        t3 = pow(t2,0.5);
        t4 = 1+beta;
        t5 = 1/t4;
        t8 = exp(lgamma(0.5+0.5*beta));
        t9 = pow(t8,-0.15E1);
        t10 = sigma*sigma;
        t15 = pow(t2/t8,t5);
        t16 = x-mu;
        t18 = t16/sigma;
        t19 = fabs(t18);
        t20 = pow(t19,2*t5);
        t22 = exp(-t15*t20);
        t22 = Math.max(t22, 1e10/Double.MAX_VALUE) ; // MODIFIED
        t25 = t4*t4;
        t41 = pow(t2,-0.5);
        t44 = pow(t8,0.15E1);
        return(-(-t3*t5*t9/t10*t22+2*t3/t25*t9/t10/sigma*t15*t20*abs(1,t18)*t16/t19*
                t22)*t41*t4*t44*sigma/t22);
    }
    
    double dcb(double x,double mu,double sigma,double beta) {
        double t58;
        double t1;
        double t3;
        double t2;
        double t4;
        double t31;
        double t55;
        double t8;
        double t9;
        double t10;
        double t7;
        double t11;
        double t20;
        double t12;
        double t14;
        double t13;
        double t21;
        double t19;
        double t18;
        double t17;
        double t22;
        double t26;
        double t25;
        double t5;
        
        t1 = 0.15E1+0.15E1*beta;
        t2 = exp(lgamma(t1));
        t3 = pow(t2,0.5);
        t4 = 1+beta;
        t5 = 1/t4;
        t7 = 0.5+0.5*beta;
        t8 = exp(lgamma(t7));
        t9 = pow(t8,-0.15E1);
        t10 = t3*t5*t9;
        t11 = 1/sigma;
        t12 = 1/t8;
        t13 = t2*t12;
        t14 = pow(t13,t5);
        t17 = fabs((x-mu)*t11);
        t18 = pow(t17,2*t5);
        t19 = t14*t18;
        t20 = exp(-t19);
        t20 = Math.max(t20, 1e10/Double.MAX_VALUE) ; // MODIFIED
        t21 = t11*t20;
        t22 = Psi(t1);
        t25 = t4*t4;
        t26 = 1/t25;
        t31 = Psi(t7);
        t55 = pow(t2,-0.5);
        t58 = pow(t8,0.15E1);
        return(-(0.75*t10*t21*t22-t3*t26*t9*t11*t20-0.75*t10*t21*t31+t10*t11*(-t14*(-
                t26*log(t13)+t5*(0.15E1*t22*t2*t12-0.5*t13*t31)/t2*t8)*t18+2*t19*t26*log(t17))*
                t20)*t55*t4*t58*sigma/t20);
    }
    
    
    /* The options were    : operatorarrow */
    double dcmm(double x,double mu,double sigma,double beta) {
        double t8;
        double t7;
        double t26;
        double t21;
        double t9;
        double t15;
        double t13;
        double t12;
        double t14;
        double t11;
        double t16;
        double t20;
        double t24;
        double t28;
        
        t7 = 1+beta;
        t8 = 1/t7;
        t9 = pow(exp(lgamma(0.15E1+0.15E1*beta))/exp(lgamma(0.5+0.5*beta)),t8);
        t11 = 1/sigma;
        t12 = (x-mu)*t11;
        t13 = fabs(t12);
        t14 = pow(t13,2*t8);
        t15 = t9*t14;
        t16 = t7*t7;
        t20 = pow(abs(1,t12),2);
        t21 = sigma*sigma;
        t24 = t13*t13;
        t26 = t20/t21/t24;
        t28 = t15*t8;
        return(4*t15/t16*t26-2*t28*signum(1,t12)*t11/t13-2*t28*t26);
    }
    
    /* The options were    : operatorarrow */
    double dcms(double x,double mu,double sigma,double beta) {
        double t25;
        double t15;
        double t30;
        double t11;
        double t10;
        double t20;
        double t13;
        double t7;
        double t14;
        double t22;
        double t28;
        double t33;
        double t19;
        double t16;
        double t12;
        double t9;
        double t8;
        
        t7 = 1+beta;
        t8 = 1/t7;
        t9 = pow(exp(lgamma(0.15E1+0.15E1*beta))/exp(lgamma(0.5+0.5*beta)),t8);
        t10 = x-mu;
        t11 = 1/sigma;
        t12 = t10*t11;
        t13 = fabs(t12);
        t14 = pow(t13,2*t8);
        t15 = t9*t14;
        t16 = t7*t7;
        t19 = abs(1,t12);
        t20 = t19*t19;
        t22 = sigma*sigma;
        t25 = t13*t13;
        t28 = t20*t10/t22/sigma/t25;
        t30 = t15*t8;
        t33 = 1/t13;
        return(4*t15/t16*t28-2*t30*signum(1,t12)*t11*t33+2*t30*t19/t22*t33-2*t30*t28)
        ;
    }
    
    /* The options were    : operatorarrow */
    double dcmb(double x,double mu,double sigma,double beta) {
        double t34;
        double t1;
        double t6;
        double t4;
        double t28;
        double t35;
        double t29;
        double t27;
        double t11;
        double t10;
        double t9;
        double t8;
        double t7;
        double t38;
        double t32;
        double t30;
        double t2;
        double t5;
        double t3;
        
        t1 = 0.15E1+0.15E1*beta;
        t2 = exp(lgamma(t1));
        t3 = 0.5+0.5*beta;
        t4 = exp(lgamma(t3));
        t5 = 1/t4;
        t6 = t2*t5;
        t7 = 1+beta;
        t8 = 1/t7;
        t9 = pow(t6,t8);
        t10 = t7*t7;
        t11 = 1/t10;
        t27 = 1/sigma;
        t28 = (x-mu)*t27;
        t29 = fabs(t28);
        t30 = pow(t29,2*t8);
        t32 = abs(1,t28);
        t34 = 1/t29;
        t35 = t27*t34;
        t38 = t9*t30;
        return(-2*t9*(-t11*log(t6)+t8*(0.15E1*Psi(t1)*t2*t5-0.5*t6*Psi(t3))/t2*t4)*
                t30*t8*t32*t35+4*t38/t10/t7*log(t29)*t32*t35+2*t38*t11*t32*t27*t34);
    }
    
    /* The options were    : operatorarrow */
    double dcss(double x,double mu,double sigma,double beta) {
        double t9;
        double t19;
        double t18;
        double t5;
        double t17;
        double t4;
        double t3;
        double t6;
        double t13;
        double t22;
        double t20;
        double t10;
        double t28;
        double t26;
        double t29;
        double t8;
        double t2;
        double t23;
        double t21;
        double t59;
        double t55;
        double t49;
        double t34;
        double t46;
        double t43;
        double t79;
        double t77;
        double t69;
        double t74;
        double t66;
        double t36;
        double t39;
        double t38;
        double t88;
        double t16;
        double t47;
        double t51;
        double t52;
        double t56;
        
        t2 = exp(lgamma(0.15E1+0.15E1*beta));
        t3 = pow(t2,0.5);
        t4 = 1+beta;
        t5 = 1/t4;
        t6 = t3*t5;
        t8 = exp(lgamma(0.5+0.5*beta));
        t9 = pow(t8,-0.15E1);
        t10 = sigma*sigma;
        t13 = t9/t10/sigma;
        t16 = pow(t2/t8,t5);
        t17 = x-mu;
        t18 = 1/sigma;
        t19 = t17*t18;
        t20 = fabs(t19);
        t21 = pow(t20,2*t5);
        t22 = t16*t21;
        t23 = exp(-t22);
        t23 = Math.max(t23, 1e10/Double.MAX_VALUE) ; // MODIFIED
        t26 = t4*t4;
        t28 = t3/t26;
        t29 = t10*t10;
        t34 = abs(1,t19);
        t36 = 1/t20;
        t38 = t17*t36*t23;
        t39 = t21*t34*t38;
        t43 = t3/t26/t4;
        t46 = t9/t29/sigma;
        t47 = t46*t16;
        t49 = t34*t34;
        t51 = t17*t17;
        t52 = t20*t20;
        t55 = t51/t52*t23;
        t56 = t21*t49*t55;
        t59 = t28*t13*t16;
        t66 = t16*t16;
        t69 = t21*t21;
        t74 = pow(t2,-0.5);
        t77 = pow(t8,0.15E1);
        t79 = 1/t23;
        t88 = (-t6*t9/t10*t23+2*t59*t39)*t74;
        return(-(2*t6*t13*t23-8*t28*t9/t29*t16*t39-4*t43*t47*t56+2*t59*t21*signum(1,
                t19)*t38+2*t28*t47*t56+4*t43*t46*t66*t69*t49*t55)*t74*t4*t77*sigma*t79-t88*t4*
                t77*t79+2*t88*t77*t18*t79*t22*t34*t17*t36);
    }
    
    /* The options were    : operatorarrow */
    double dcsb(double x,double mu,double sigma,double beta) {
        double t23;
        double t15;
        double t11;
        double t22;
        double t18;
        double t20;
        double t19;
        double t7;
        double t8;
        double t12;
        double t1;
        double t10;
        double t6;
        double t4;
        double t9;
        double t5;
        double t2;
        double t3;
        double t101;
        double t21;
        double t16;
        double t13;
        double t14;
        double t33;
        double t31;
        double t29;
        double t28;
        double t27;
        double t24;
        double t49;
        double t52;
        double t46;
        double t59;
        double t65;
        double t64;
        double t63;
        double t61;
        double t62;
        double t60;
        double t75;
        double t74;
        double t84;
        double t100;
        double t97;
        double t103;
        double t102;
        double t108;
        double t109;
        
        t1 = 0.15E1+0.15E1*beta;
        t2 = exp(lgamma(t1));
        t3 = pow(t2,0.5);
        t4 = 1+beta;
        t5 = 1/t4;
        t6 = t3*t5;
        t7 = 0.5+0.5*beta;
        t8 = exp(lgamma(t7));
        t9 = pow(t8,-0.15E1);
        t10 = t6*t9;
        t11 = sigma*sigma;
        t12 = 1/t11;
        t13 = 1/t8;
        t14 = t2*t13;
        t15 = pow(t14,t5);
        t16 = x-mu;
        t18 = t16/sigma;
        t19 = fabs(t18);
        t20 = pow(t19,2*t5);
        t21 = t15*t20;
        t22 = exp(-t21);
        t22 = Math.max(t22, 1e10/Double.MAX_VALUE) ; // MODIFIED
        t23 = t12*t22;
        t24 = Psi(t1);
        t27 = t4*t4;
        t28 = 1/t27;
        t29 = t3*t28;
        t31 = t9*t12*t22;
        t33 = Psi(t7);
        t46 = -t28*log(t14)+t5*(0.15E1*t24*t2*t13-0.5*t14*t33)/t2*t8;
        t49 = log(t19);
        t52 = -t15*t46*t20+2*t21*t28*t49;
        t59 = t9/t11/sigma*t15;
        t60 = t29*t59;
        t61 = abs(1,t18);
        t62 = t20*t61;
        t63 = t62*t16;
        t64 = 1/t19;
        t65 = t64*t22;
        t74 = t16*t64*t22;
        t75 = t62*t74;
        t84 = t27*t27;
        t97 = pow(t2,-0.5);
        t100 = pow(t8,0.15E1);
        t101 = t100*sigma;
        t102 = 1/t22;
        t103 = t101*t102;
        t108 = (-t6*t31+2*t60*t75)*t97;
        t109 = t108*t4;
        return(-(-0.75*t10*t23*t24+t29*t31+0.75*t10*t23*t33-t10*t12*t52*t22+0.15E1*
                t60*t63*t65*t24-4*t3/t27/t4*t59*t75-0.15E1*t60*t63*t65*t33+2*t60*t46*t20*t61*
                t74-4*t3/t84*t59*t20*t49*t61*t74+2*t60*t63*t64*t52*t22)*t97*t4*t103+0.75*t109*
                t101*t102*t24-t108*t103-0.75*t109*t101*t102*t33+t109*t101*t102*t52);
    }
    
    /* The options were    : operatorarrow */
    double dcbb(double x,double mu,double sigma,double beta) {
        double t36;
        double t29;
        double t23;
        double t32;
        double t30;
        double t26;
        double t17;
        double t14;
        double t18;
        double t12;
        double t20;
        double t5;
        double t8;
        double t11;
        double t4;
        double t1;
        double t10;
        double t7;
        double t9;
        double t3;
        double t21;
        double t19;
        double t13;
        double t2;
        double t22;
        double t28;
        double t27;
        double t41;
        double t42;
        double t43;
        double t46;
        double t38;
        double t52;
        double t53;
        double t47;
        double t49;
        double t44;
        double t66;
        double t68;
        double t64;
        double t57;
        double t61;
        double t70;
        double t79;
        double t76;
        double t109;
        double t107;
        double t118;
        double t123;
        double t122;
        double t128;
        double t126;
        double t127;
        double t129;
        double t136;
        double t137;
        
        t1 = 0.15E1+0.15E1*beta;
        t2 = exp(lgamma(t1));
        t3 = pow(t2,0.5);
        t4 = 1+beta;
        t5 = 1/t4;
        t7 = 0.5+0.5*beta;
        t8 = exp(lgamma(t7));
        t9 = pow(t8,-0.15E1);
        t10 = t3*t5*t9;
        t11 = 1/sigma;
        t12 = 1/t8;
        t13 = t2*t12;
        t14 = pow(t13,t5);
        t17 = fabs((x-mu)*t11);
        t18 = pow(t17,2*t5);
        t19 = t14*t18;
        t20 = exp(-t19);
        t20 = Math.max(t20, 1e10/Double.MAX_VALUE) ; // MODIFIED
        t21 = t11*t20;
        t22 = Psi(t1);
        t23 = t22*t22;
        t26 = t4*t4;
        t27 = 1/t26;
        t28 = t3*t27;
        t29 = t28*t9;
        t30 = t21*t22;
        t32 = Psi(t7);
        t36 = log(t13);
        t38 = t22*t2;
        t41 = 0.15E1*t38*t12-0.5*t13*t32;
        t42 = t5*t41;
        t43 = 1/t2;
        t44 = t43*t8;
        t46 = -t27*t36+t42*t44;
        t47 = t14*t46;
        t49 = log(t17);
        t52 = -t47*t18+2*t19*t27*t49;
        t53 = t11*t52;
        t57 = Psi(1,t1);
        t61 = 1/t26/t4;
        t64 = t9*t11*t20;
        t66 = t21*t32;
        t68 = t53*t20;
        t70 = t32*t32;
        t76 = Psi(1,t7);
        t79 = t46*t46;
        t107 = t26*t26;
        t109 = t49*t49;
        t118 = t52*t52;
        t122 = 0.5625*t10*t21*t23-0.15E1*t29*t30-0.1125E1*t10*t21*t22*t32+0.15E1*t10*
                t53*t20*t22+0.1125E1*t10*t21*t57+2*t3*t61*t64+0.15E1*t29*t66-2*t29*t68+0.5625*
                t10*t21*t70-0.15E1*t10*t53*t20*t32-0.375*t10*t21*t76+t10*t11*(-t14*t79*t18-t14*
                (2*t61*t36-2*t27*t41*t44+t5*(0.225E1*t57*t2*t12+0.225E1*t23*t2*t12-0.15E1*t38*
                t12*t32+0.25*t13*t70-0.25*t13*t76)*t44-0.15E1*t42*t44*t22+0.5*t42*t43*t32*t8)*
                t18+4*t47*t18*t27*t49-4*t19/t107*t109-4*t19*t61*t49)*t20+t10*t11*t118*t20;
        t123 = pow(t2,-0.5);
        t126 = pow(t8,0.15E1);
        t127 = t126*sigma;
        t128 = 1/t20;
        t129 = t127*t128;
        t136 = (0.75*t10*t30-t28*t64-0.75*t10*t66+t10*t68)*t123;
        t137 = t136*t4;
        return(-t122*t123*t4*t129+0.75*t137*t127*t128*t22-t136*t129-0.75*t137*t127*
                t128*t32+t137*t127*t128*t52);
    }
    
    /* The options were    : operatorarrow */
    double cost(double x,double m,double s,double b) {
        return(-log(gg(x,m,s,b))/log(2));
    }
    
     /* The options were    : operatorarrow */
    double similarity(double x,double m,double s,double b) {
        return gg(x,m,s,b);
    }
    
    double costBIC(double x,double m,double s,double b){
        return(Math.log(gg(x,m,s,b)));
    }
    
    /* The options were    : operatorarrow */
    double gg(double x,double mu,double sigma,double beta) {
        double t2;
        double t10;
        
        t2 = 1/sigma;
        t10 = pow(fabs((x-mu)*t2),2/(1+beta));
        return(omega(beta)*t2*exp(-c(beta)*t10));
    }
    
    /* The options were    : operatorarrow */
    double omega(double beta) {
        double t9;
        double t3;
        
        t3 = pow(exp(lgamma(0.15E1+0.15E1*beta)),0.5);
        t9 = pow(exp(lgamma(0.5+0.5*beta)),-0.15E1);
        return(t3/(1+beta)*t9);
    }
    
    /* The options were    : operatorarrow */
    double c(double beta) {
        return(pow(exp(lgamma(0.15E1+0.15E1*beta))/exp(lgamma(0.5+0.5*beta)),1/(1+
                beta)));
    }
    
    public double costV(double x, double[] msb) {
        return cost(x, msb[0], msb[1], msb[2]);
    }
    
    public double similarityV(double x, double[] msb){
        return similarity(x, msb[0], msb[1], msb[2]);
    }
    
    public double costVBIC(double x, double[] msb){
        return costBIC(x, msb[0], msb[1], msb[2]);
    }
    
    public double determineCost(double [] x, double [][] msb) {
        double sum = 0.0 ;
        for (int i=0 ; i<x.length ; i++)
            sum += costV(x[i], msb[i]) ;
        return sum ;
    }
    
    
    public double [] update(double x, double m, double s, double b) {
        double [] gr = {dcm(x,m,s,b), dcs(x,m,s,b), dcb(x,m,s,b) } ;
        double [][] hessian = {{dcmm(x,m,s, b), dcms(x,m,s,b), dcmb(x,m,s,b)},
        {0, dcss(x,m,s,b), dcsb(x,m,s,b) },
        {0, 0, dcbb(x,m,s, b) } } ;
        hessian [1][0] = hessian [0][1] ;
        hessian [2][0] = hessian [0][2] ;
        hessian [2][1] = hessian [1][2] ;
        Matrix mgr = new Matrix(gr, 3) ;
        Matrix mhessian = new Matrix(hessian) ;
        Matrix result = mhessian.inverse().times(mgr) ;
        return result.getColumnPackedCopy() ;
    }
    
    public double [] getInfo(double x, double m, double s, double b) {
        double [] result = {dcm(x,m,s,b), dcs(x,m,s,b), dcb(x,m,s,b),
        dcmm(x,m,s, b), dcms(x,m,s,b), dcmb(x,m,s,b),
        dcss(x,m,s,b), dcsb(x,m,s,b), dcbb(x,m,s, b) }  ;
        return result ;
    }
    
    public double [] update(double [] x, double m, double s, double b) {
        double [] gr = new double [3] ;
        double [][] hessian = new double[3][3] ;
        for (int i=0 ; i<x.length ; i++) {
            gr[0] += dcm(x[i],m,s,b) ;
            gr[1] += dcs(x[i],m,s,b) ;
            gr[2] += dcb(x[i],m,s,b) ;
            hessian[0][0] += dcmm(x[i],m,s,b) ;
            hessian[0][1] += dcms(x[i],m,s,b) ;
            hessian[0][2] += dcmb(x[i],m,s,b) ;
            hessian[1][1] += dcss(x[i],m,s,b) ;
            hessian[1][2] += dcsb(x[i],m,s,b) ;
            hessian[2][2] += dcbb(x[i],m,s,b) ;
        }
        double factor = 1.0/x.length/5 ;
        gr[0] *= factor ;
        gr[1] *= factor ;
        gr[2] *= factor ;
        hessian[0][0] *= factor ;
        hessian[0][1] *= factor ;
        hessian[0][2] *= factor ;
        hessian[1][1] *= factor ;
        hessian[1][2] *= factor ;
        hessian[2][2] *= factor ;
        hessian[1][0] = hessian[0][1] ;
        hessian[2][0] = hessian[0][2] ;
        hessian[2][1] = hessian[1][2] ;
        Matrix mgr = new Matrix(gr, 3) ;
        Matrix mhessian = new Matrix(hessian) ;
        Matrix result = mhessian.inverse().times(mgr) ;
        return result.getColumnPackedCopy() ;
    }
    
    public double [] update_nohes(double [] x, double m, double s, double b) {
        double [] gr = new double [3] ;
        double [] hessian = new double[3] ;
        for (int i=0 ; i<x.length ; i++) {
            gr[0] += dcm(x[i],m,s,b) ;
            gr[1] += dcs(x[i],m,s,b) ;
            gr[2] += dcb(x[i],m,s,b) ;
            hessian[0] += dcmm(x[i],m,s,b) ;
            hessian[1] += dcss(x[i],m,s,b) ;
            hessian[2] += dcbb(x[i],m,s,b) ;
        }
        double factor = 1.0/x.length/5 ;
        gr[0] *= factor ;
        gr[1] *= factor ;
        gr[2] *= factor ;
        hessian[0] *= factor ;
        hessian[1] *= factor ;
        hessian[2] *= factor ;
        /*
        if (hessian[0][0] <= 1e-8) {
            System.out.println ("mu-derivative negative "+hessian[0][0]) ;
            hessian[0][0] = 1e-8 ;
        }
        if (hessian[1][1] <= 0) {
            System.out.println ("sigma-derivative negative "+hessian[1][1]) ;
            gr [1] = 0.2*s ;
            hessian[1][1] = 1.0 ;
        }
        if (hessian[2][2] <= 0) {
            System.out.println ("beta-derivative negative "+hessian[2][2]) ;
            gr[2] = -0.01 ;
            hessian[2][2] = 1.0 ;
        }
         */
        gr[0]/=Math.abs(hessian[0]) ;
        gr[1]/=Math.abs(hessian[1]) ;
        gr[2]/=Math.abs(hessian[2]) ;
        return gr ;
    }
    
    public double quantile(double [] arr, int k) {
        double [] cost = arr.clone() ;
        int l = 0 ;
        int r = cost.length - 1 ;
        while (r>l) {
            int i = partition(cost, l, r) ;
            if (i >= k) r = i-1 ;
            if (i <= k) l = i+1 ;
        }
        return cost[r] ;
    }
    
    public int partition(double [] cost, int l, int r) {
        int i = l-1 ;
        int j = r ;
        double v = cost[r] ;
        for (;;) {
            while (cost[++i] < v) ;
            while (v < cost[--j]) if (j==l) break ;
            if (i >= j) break ;
            double h = cost[i] ;
            cost[i] = cost[j] ;
            cost[j] = h ;
        }
        double h = cost[i] ;
        cost[i] = cost[r] ;
        cost[r] = h ;
        return i ;
    }
    
    public double [] optimizeMSB(double[] sample) {
        // HERE IS ASSUMPTION OF WHITE SPACE AT VARIOUS POSITIONS !!!
        double empMu = 0.0 ;    // Assume here centered data
        double empSigma = 1.0 ; // Assume here whitened data
        double empBeta = 0.0 ;  // Use GAUSSIAN as starting condition
        // This can be improved later to a beta estimated from data
        double lb = sample[0] ;
        double ub = sample[0] ;
        for (int i=1 ; i<sample.length ; i++) {
            lb = Math.min(lb, sample[i]) ;
            ub = Math.max(ub, sample[i]) ;
        }
        double median = quantile(sample, sample.length/2) ;
        double sigmaLaplace = 0 ;
        for (int i=0 ; i<sample.length ; i++)
            sigmaLaplace += Math.abs(sample[i]-median) ;
        sigmaLaplace /= sample.length ;
        double costGauss   = 1.4189385332 ; // ln(sqrt(2 pi) + 0.5 FOR WHITE SPACE !!!
        double costLaplace = Math.log(2.0*sigmaLaplace) + 1.0 ;
        
        double costUniform = Math.log(ub-lb) ;
        if (costUniform < costGauss && costUniform < costLaplace) {
            empMu    = 0.5*(ub+lb) ;
            empSigma = 0.55*(ub-lb) ; // actually not the true estimate (0.33)
            empBeta  = -0.97 ;
        } else if (costLaplace < costUniform && costLaplace < costGauss) {
            empMu    = median ;
            empSigma = sigmaLaplace*1.4142 ;
            empBeta  = 1.05 ;
        }
        double [] m = new double [3] ;
        m[0] = empMu ;
        m[1] = empSigma ;
        m[2] = empBeta ;
        
        if (verbose)
            System.out.println("-->"+0+": "+m[0]+" "+m[1]+" "+m[2]) ;
        
        mainloop:for (int j=1 ; j<50 ; j++){
            try {
                double z1=0,z2=0,n1=0,n2=0,zs=0,ns=0 ;
                for (int i=0 ; i<sample.length ; i++) {
                    z1 += dcm(sample[i],m[0],m[1],m[2]) ;
                    z2 += dcb(sample[i],m[0],m[1],m[2]) ;
                    n1 += dcmm(sample[i],m[0],m[1],m[2]) ;
                    n2 += dcbb(sample[i],m[0],m[1],m[2]) ;
                    //zs += dcs(sample[i],m[0],m[1],m[2]) ;
                    //ns += dcss(sample[i],m[0],m[1],m[2]) ;
                }
                //n1 = Math.max (1.0, Math.abs (n1)) ;
                //n2 = Math.max (1.0, Math.abs (n2)) ;
                double g1=z1/Math.abs(n1) ;
                double g2=z2/Math.abs(n2) ;
                //double gs=zs/ns ;
                if (g1>=-Double.MAX_VALUE && g1<=Double.MAX_VALUE && g2>=-Double.MAX_VALUE && g2<Double.MAX_VALUE) {
                    m[0] -= g1 ;
                    if (m[2] - g2 < -0.97)
                        g2 = 0.97 + m[2] ;
                    if(g2 > 0.2) g2 = 0.2;
                    if(g2 < -0.2) g2 = -0.2;
                    m[2] -= g2 ;
                    
                    double pppp = 2.0/(1.0+m[2]) ;
                    m[1] = 0 ;
                    for (int i=0 ; i<sample.length ; i++)
                        m[1] += Math.pow(Math.abs(sample[i]-m[0]),pppp) ;
                    m[1] = Math.pow(m[1]/sample.length*pppp*c(m[2]),1/pppp) ;
                    
                    if(m[0] < lb)
                        m[0] = lb;
                    if(m[0] > ub)
                        m[0] = ub;
                    if(m[1] > 0.55 * (ub-lb))
                        m[1] = 0.55 * (ub-lb);
                    
                    
                    //m[1] -= gs ;
                    if (verbose)
                        System.out.println("-->"+j+": "+m[0]+" "+m[1]+" "+m[2]) ;
                    if (Math.abs(g1) < 0.01*m[1] && Math.abs(g2) < 0.01)
                        break mainloop;
                } else {
                    m[1] *= 1.01 ;
                    m[0] += m[1] / 100.0 ;
                    m[2] += 0.01 ;
                    if (verbose)
                        System.out.println("-->"+j+": "+m[0]+" "+m[1]+" "+m[2]+" ERROR") ;
                }
            } catch (Exception e) {
                m[1] *= 1.2 ;
                m[0] += m[1]/100 ;
            }
        }
        return m ;
    }
    
    public double [] optimizeMSB2(double[] sample) {
        
        double empMu = 0.0 ;    // Assume here centered data
        double empSigma = 1.0 ; // Assume here whitened data
        double empBeta = 0.0 ;  // Use GAUSSIAN as starting condition
        // This can be improved later to a beta estimated from data
        double [] m = new double [3] ;
        m[0] = empMu ;
        m[1] = empSigma ;
        m[2] = empBeta ;
        
        for (int j=0 ; j<20 ; j++){
            try {
                double [] g = update_nohes(sample,m[0],m[1],m[2]) ;
                if (g[0] == Double.NaN || g[1] == Double.NaN || g[2] == Double.NaN ||
                        g[0] == Double.NEGATIVE_INFINITY || g[1] == Double.NEGATIVE_INFINITY || g[2] == Double.NEGATIVE_INFINITY ||
                        g[0] == Double.POSITIVE_INFINITY || g[1] == Double.POSITIVE_INFINITY || g[2] == Double.POSITIVE_INFINITY) {
                    m[1] *= 1.2 ;
                    m[0] += m[1] / 100.0 ;
                } else {
                    if (verbose)
                        System.out.println(m[0]+" "+m[1]+" "+m[2]+" "+g[0]+" "+g[1]+" "+g[2]) ;
                    m[0] -= g[0] ;
                    m[1] -= g[1] ;
                    m[2] -= g[2] ;
                    
                    if (m[0]<empMu-3*empSigma){
                        if (verbose) System.out.println("Warning: mu = "+m[0]+" (empirically: "+empMu+" +/- 3 * "+empSigma+")") ;
                        m[0] = empMu-3*empSigma ;
                    }
                    if (m[0]>empMu+3*empSigma){
                        if (verbose) System.out.println("Warning: mu = "+m[0]+" (empirically: "+empMu+" +/- 3 * "+empSigma+")") ;
                        m[0] = empMu+3*empSigma ;
                    }
                    if (m[1]<empSigma/3) {
                        if (verbose) System.out.println("Warning: sigma = "+m[1]+" (empirically: "+empSigma+")") ;
                        m[1] = empSigma/3 ;
                    }
                    if (m[1]>empSigma*3) {
                        if (verbose) System.out.println("Warning: sigma = "+m[1]+" (empirically: "+empSigma+")") ;
                        m[1] = empSigma*3 ;
                    }
                    if (m[2]<-0.95)
                        m[2] = -0.95 ;
                    if (m[2] > 1.5)
                        if (verbose) System.out.println("Warning: beta = "+m[2]) ;
                }
            } catch (Exception e) {
                m[1] *= 1.5 ;
            }
        }
        return m ;
    }
    
}
