/*
 * Ica.java
 *
 * Created on 13. Oktober 2005, 15:46
 *
 */

package oci;
import Jama.*;
import java.util.*;


/**
 *
 * @author boehm
 */
public class Ica {
    public double[][] ev;   //eigenvalues
    public double wscCost;  // Cost of White Space Compensation in Bits
    boolean verbose = false;
    
    
    /** Creates a new instance of Ica */
    public Ica() {
    }
    
    public Matrix matrix(double[][] m) {
        return new Matrix(m) ;
    }
    
    public Matrix columnVector(double[] v) {
        return new Matrix(v, v.length) ;
    }
    
    public Matrix rowVector(double [] v) {
        return new Matrix(v, 1) ;
    }
    
    public double[][] getEV(){
        return this.ev;
    }
    
    public double[][] centerData(double[][] data) {
        int numObj= data.length ;
        int d = data[0].length ;
        double[][] result = new double [numObj][d] ;
        
        for (int i=0 ; i<d ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sum+= data[j][i] ;
            sum /= numObj ;
            for (int j=0 ; j<numObj ; j++)
                result[j][i] = data[j][i] - sum ;
        }
        
        return result ;
    }
    
    public double[][] centerData(double[][] data, double [] center) {
        int numObj= data.length ;
        int d = data[0].length; 
        if (center == null)
            center = new double[d] ;
        double[][] result = new double [numObj][d] ;
        
        for (int i=0 ; i<d ; i++){
            center [i] = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                center[i] += data[j][i] ;
            center[i] /= numObj ;
            for (int j=0 ; j<numObj ; j++)
                result[j][i] = data[j][i] - center[i] ;
        }
        
        return result ;
    }
    
    public double[][] centerByMedian(double[][] data) {
        int numObj= data.length ;
        int dim = data[0].length ;
        double [] sortarray = new double[numObj] ;
        double[][] result = new double [numObj][dim] ;
        for (int i=0 ; i<dim ; i++) {
            double median = 0.0 ;
            for (int j=0 ; j<numObj ; j++)
                sortarray[j]=data[j][i] ;
            Arrays.sort(sortarray) ;
            if (numObj%2 == 0)
                median=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
            else
                median=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
            for (int j=0 ; j<numObj ; j++)
                result[j][i] = data[j][i]-median ;
        }
        return result ;
    }
    
    public Matrix covar(Matrix data) {
        return data.transpose().times(data).times(1.0/data.getRowDimension()) ;
    }
    
    public double[][] covarByMedian(double[][] data) {
        int numObj= data.length ;
        int dim = data[0].length ;
        double [] sortarray = new double[numObj] ;
        double[][] cov = new double[dim][dim];
        for(int l1 = 0; l1 < dim; l1++) {
            for(int l2 = 0; l2 < dim; l2++) {
                for(int j = 0; j < numObj; j++)
                    sortarray[j] = data[j][l1]*data[j][l2] ;
                
                Arrays.sort(sortarray) ;
                if (numObj%2 == 0)
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2])/2 ;
                else
                    cov[l1][l2]=(sortarray[numObj/2-1]+sortarray[numObj/2]+sortarray[numObj/2+1])/3 ;
            }
        }
        double maxdiff = 0.0 ;
        for (int i=0 ; i<dim ; i++){
            double sum = 0.0 ;
            for (int j=0 ; j<dim ; j++)
                sum += Math.abs(cov[i][j]) ;
            double diff = 1.1*sum - 2.1*cov[i][i] ;
            if (diff>maxdiff)
                maxdiff = diff ;
        }
        if (maxdiff > 0.0)
            for (int i=0 ; i<dim ; i++)
                cov[i][i] += maxdiff ;
        return cov ;
    }
    
    public double [][] whitenData(double[][] data, double [][] transform, double [][] retransform) {
        int numObj= data.length ;
        int dim = data[0].length ;
        double[][] result = new double [numObj][dim] ;
        
        Matrix dm = new Matrix(data) ;
        // T E S T    T E S T    T E S T    T E S T    T E S T    T E S T
        Matrix c = covar(dm) ;
        //robust fit
        //Matrix c = new Matrix(covarByMedian(data)) ;
        EigenvalueDecomposition e = new EigenvalueDecomposition(c);
        Matrix evm = e.getV();
        Matrix ewm = e.getD();
        if (verbose) {
            System.out.println("Whiten Data:");
            System.out.println("covariance matrix");
            c.print(5,3) ;
            System.out.println("eigenvectors");
            evm.print(5, 3);
            System.out.println("eigenvalues");
            ewm.print(5, 3);
        }
        ev = ewm.getArrayCopy();
        double [] diag = new double [dim] ;
        for (int i=0 ; i<dim ; i++) {
            diag[i] = Math.sqrt(Math.min(Math.max(ewm.get(i,i), 1.0e-6), 1.0e+6) ) ;
            wscCost += Math.log(diag[i]) / Math.log(2.0) ;
        }
        
        for (int i=0; i<dim ; i++) {
            for (int j=0 ; j<dim ; j++) {
            transform  [i][j] = evm.get(i,j) / diag[j] ;
            retransform[i][j] = evm.get(j,i) * diag[i] ;
            }
        }
        
        Matrix res = dm.times(new Matrix(transform) ) ;
        if (verbose) {
            System.out.println("Some sanity checks for data whitening:") ;
            System.out.println("transform &* retransform: This should be unity:") ;
            (new Matrix(transform)).times(new Matrix(retransform)) .print(10,8) ;
        }
        return res.getArrayCopy() ;
        
    }
    
    public double g_func(double x) {
        return Math.tanh(x) ;
    }
    
    public double g_slope(double x) {
        return 1.0-Math.pow(Math.tanh(x), 2) ;
    }
    
    public double curtosis(double [][] data, double [] w) {
        int numObj= data.length ;
        int dim = data[0].length ;
        double res = 0.0 ;
        double xsq = 0.0 ;
        for (int i=0 ; i<numObj ; i++) {
            double x = 0 ;
            for (int j=0 ; j<dim ; j++)
                x+=data[i][j]*w[j] ;
            res += Math.pow(x,4.0) ;
            xsq += x*x ;
        }
        if (xsq/numObj>1.1 || xsq/numObj<0.9)
            System.out.println("Error xsq="+xsq/numObj) ;
        return res/numObj - 3 ;
    }
    
    private void rotateAndDisplay(double[][] data, double[][] w, int iter){
        Matrix mw = new Matrix(w);
        Matrix mwhitened = new Matrix(data);
         double[][] projectedWhitened = mwhitened.times(mw.transpose()).getArrayCopy();
         String title = new Integer(iter).toString();
        displayData(projectedWhitened, title);
        System.out.println("projectedWhitened:") ;
            mwhitened.times(mw.transpose()).print(6,4) ;
    }
    
    //data: Zeilenvektoren
    /**
     * Standard fastICA. Re-transformation to original space commented out.
     * @param data data as row vetors.
     * @return w as row vectors.
     */
    public double [][] fastICA(double[][] data) {
        //verbose = true;
        // fastICA algorithm acc. to p. 194 with deflationary orthogonalization
        int numObj= data.length ;
        int dim = data[0].length ;
        Random r = new Random();
        
        // T E S T    T E S T    T E S T    T E S T    T E S T    T E S T
        //double[][] data1 = centerByMedian(data) ;
        double[][] data1 = centerData(data) ;
        //displayData(data1, "centered");
        double[][] transform = new double[dim][dim] ;
        double[][] retransform = new double[dim][dim] ;
        
        data1 = whitenData(data1, transform, retransform) ;
        //displayData(data1, "whitened");
        
        if (verbose) {
            Matrix dd = new Matrix(data1) ;
            System.out.println("For whitened data the covariance matrix must also be unity: ...") ;
            covar(dd).print(10, 8) ;
            System.out.println("Daten:") ;
            dd.print(6,4) ;
            System.out.println("-------");
        }
        
        double[][] w = new double[dim][dim] ;
        int alliter = 0 ;
        
        for (int n=0 ; n<dim ; n++) {
            
            for (int i=0; i<dim ; i++)
                w[n][i] = r.nextDouble() ;
            double sum = 0.0 ;
            for (int i=0; i<dim ; i++)
                sum += w[n][i]*w[n][i] ;
            for (int i=0; i<dim ; i++)
                w[n][i] /= Math.sqrt(sum) ;
            
            double [] old_w ;
            double [] wUnorth ;
            
            int iter = 0 ;
            do {
                if (verbose)
                    System.out.println("w = ["+w[n][0]+" "+w[n][1]+"] curtosis = "+curtosis(data1, w[n])) ;
               //if(iter == 1)
                //rotateAndDisplay(data1, w , iter);
                old_w = w[n].clone() ;
                double eLeft[] = new double[dim] ;
                double eRight = 0.0 ;
                for (int i=0 ; i<numObj ; i++) {
                    double scalar = 0.0 ;
                    for (int j=0 ; j<dim ; j++)
                        scalar += w[n][j]*data1[i][j] ;
                    double g = g_func(scalar) / numObj ;
                    for (int j=0 ; j<dim ; j++)
                        eLeft[j] += g * data1[i][j] ;
                    eRight += g_slope(scalar) / numObj ;
                    
                }
                for (int j=0 ; j<dim ; j++)
                    w[n][j] = eLeft[j] - eRight * w[n][j] ;
                
                // deflationary orthogonalize w[n] w.r.t. vectors w[0]..w[n-1]
                
                wUnorth = w[n].clone();
                
                for (int m=0 ; m<n ; m++) {
                    double scalar = 0.0 ;
                    for (int j=0 ; j<dim ; j++)
                        scalar += wUnorth[j] * w[m][j] ;
                    for (int j=0 ; j<dim ; j++)
                        w[n][j] -= scalar * w[m][j] ;
                }
                sum = 0.0 ;
                for (int i=0; i<dim ; i++)
                    sum += w[n][i]*w[n][i] ;
                for (int i=0; i<dim ; i++)
                    w[n][i] /= Math.sqrt(sum) ;
                sum = 0.0 ;
                for (int i=0; i<dim ; i++)
                    sum += w[n][i] * old_w[i] ;
                iter ++ ;
            } while (Math.abs(sum)<0.9999 && iter < 100);
            if (verbose) {
                System.out.println("w = ["+w[n][0]+" "+w[n][1]+"] curtosis = "+curtosis(data1, w[n])) ;
                //if(iter == 1)
                //rotateAndDisplay(data1, w , iter);
                System.out.println("ITERATIONS: "+iter);
                Matrix ww = rowVector(w[n]).times(new Matrix(retransform)) ;
                ww.print(6,4) ;
//              (new Matrix(data)).print(6,4) ;
            }
            alliter += iter ;
        }
        
        // retransform the matrix w to the original (not whitened) data space
        // w' = w' * retransform
        double [][] wn = new double [dim][dim] ;
         
        for (int m=0 ; m<dim ; m++) {
            for (int i=0 ; i<dim ; i++) {
                for (int j=0 ; j<dim ; j++)
                    wn[m][i] += w[m][j] * retransform [j][i] ;
            }
        }
         
        // normalize the result (for the last time...)
         
        for (int m=0 ; m<dim ; m++) {
            double sum = 0.0 ;
            for (int i=0; i<dim ; i++)
                sum += wn[m][i]*wn[m][i] ;
            for (int i=0; i<dim ; i++)
                wn[m][i] /= Math.sqrt(sum) ;
        }
         
         
        if (verbose) {
            System.out.println("SUM OF ITERATIONS: "+alliter) ;
            (new Matrix(wn)).print(6,4) ;
        }
         
        // To transform data such that independent components are located at
        // the coordinate axes, we have to multiply the data with wn^{-1} which
        // does not coincide with wn' (unlike the Eigenvector matrix of PCA)
         
        return wn ;
         
//        if (verbose) {
//            System.out.println("SUM OF ITERATIONS: "+alliter) ;
//            (new Matrix(w)).print(6,4) ;
//        }
//        return w;
        
    }
    
    private void displayData(double[][] d, String title){
        int dim = 2;
        DataObject[] data = new DataObject[d.length];
        for(int i = 0; i < d.length; i++){
            double[] coord = new double[dim];
            for(int j = 0; j < dim; j++)
                coord[j] = d[i][j];
            data[i] = new DataObject(coord, i);
        }
        VisuInfo visu = new VisuInfo(data, title);
        visu.setSize(600,600);
        visu.setLocation(100,0);
        visu.setVisible(true);
    }
    
    /**
     * Standard fastICA. Without whitening and re-transformation to original space.
     * @param data data[][] rows: objects, columns: dimensions
     * @return Matrix W: independent components as row vectors
     */
    public double [][] optimizeEntropy(double[][] data) {
        // fastICA algorithm acc. to p. 194 with deflationary orthogonalization
        int numObj= data.length ;
        int dim = data[0].length ;
        Random r = new Random();
        double[][] w = new double[dim][dim] ;
        int alliter = 0 ;
        for (int n=0 ; n<dim ; n++) {
            for (int i=0; i<dim ; i++)
                w[n][i] = r.nextDouble() ;
            double sum = 0.0 ;
            for (int i=0; i<dim ; i++)
                sum += w[n][i]*w[n][i] ;
            for (int i=0; i<dim ; i++)
                w[n][i] /= Math.sqrt(sum) ;
            double [] old_w ;
            double [] wUnorth ;
            int iter = 0 ;
            do {
                if (verbose)
                    System.out.println("w = ["+w[n][0]+" "+w[n][1]+"] curtosis = "+curtosis(data, w[n])) ;
                old_w = w[n].clone() ;
                double eLeft[] = new double[dim] ;
                double eRight = 0.0 ;
                for (int i=0 ; i<numObj ; i++) {
                    double scalar = 0.0 ;
                    for (int j=0 ; j<dim ; j++)
                        scalar += w[n][j]*data[i][j] ;
                    double g = g_func(scalar) / numObj ;
                    for (int j=0 ; j<dim ; j++)
                        eLeft[j] += g * data[i][j] ;
                    eRight += g_slope(scalar) / numObj ;
                }
                for (int j=0 ; j<dim ; j++)
                    w[n][j] = eLeft[j] - eRight * w[n][j] ;
                // deflationary orthogonalize w[n] w.r.t. vectors w[0]..w[n-1]
                wUnorth = w[n].clone();
                for (int m=0 ; m<n ; m++) {
                    double scalar = 0.0 ;
                    for (int j=0 ; j<dim ; j++)
                        scalar += wUnorth[j] * w[m][j] ;
                    for (int j=0 ; j<dim ; j++)
                        w[n][j] -= scalar * w[m][j] ;
                }
                sum = 0.0 ;
                for (int i=0; i<dim ; i++)
                    sum += w[n][i]*w[n][i] ;
                for (int i=0; i<dim ; i++)
                    w[n][i] /= Math.sqrt(sum) ;
                sum = 0.0 ;
                for (int i=0; i<dim ; i++)
                    sum += w[n][i] * old_w[i] ;
                iter ++ ;
            } while (Math.abs(sum)<0.99999999 && iter<100); // <0.9999 && iter<100
            if (verbose) {
                System.out.println("w = ["+w[n][0]+" "+w[n][1]+"] curtosis = "+curtosis(data, w[n])) ;
                System.out.println("ITERATIONS: "+iter);
            }
            alliter += iter ;
        }
        if (verbose) {
            System.out.println("SUM OF ITERATIONS: "+alliter) ;
            (new Matrix(w)).print(6,4) ;
        }
        
        // To transform data such that independent components are located at
        // the coordinate axes, we have to multiply the data with w^{-1} which
        // does not coincide with wn' (unlike the Eigenvector matrix of PCA)
        
        //(new Matrix (data1)).times((new Matrix (w)).transpose()).print (6,4) ;
        return w ;
    }
}
