package outlier.byEPD.bottomUp;

import pipeline.*;
import util.*;
import weka.clusterers.XMeans;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instances;
import algorithms.*;

import java.util.*;
import java.io.*;

import com.sun.org.apache.xml.internal.security.Init;

import static java.lang.Math.*;

import Jama.*;

/**
 * Calculates the outlier costs as well as the outlier similarity value for each object starting with initialisation of minKNN objects be determining
 * the euclidean distance and the rest using ICA and EPD or only using EPD
 * 
 * @author khaegler
 * @param minKNN
 *            for each object the first minKNN nearest neighbor objects are determined by using the euclidean distance
 * @param dim
 *            the dimension of the input data
 * @param stepSize
 *            the number of nearest neighbor objects determined in one step
 * @param numberObjects
 *            the total number of objects given
 */
public class FindOutlier {
        String clustername = "4Clu";
    //   String filePath ="/Users/khaegler/NetBeans/outlier-oci/trunk/data/synthetic/"+clustername+"/"+clustername+".data";
	// String filePath = "data/synthetic/2Clu.data";
	// String filePath = "data/synthetic/3Clu.data";
	// String filePath = "data/synthetic/1Clu.EPD.data";
	// String filePath = "/Users/niko/Workspace/CURRENT/outlier-oci/data/synthetic/2Clu.data";
	// String filePath = "data/default/syn1d.txt";
        String filePath = "nba.txt";
	public int dim = 4; // the dimension of the input data


	ICA ica = new ICA(dim);
	EPD epd = new EPD(dim);
	
	public int minKNN = 20; // for each object the first minKNN nearest neighbor objects are determined by using the euclidean distance
	public int version = 2;
        public final int STEP_SIZE = 1; // the number of Objects considered in one step
	public int stepSize = STEP_SIZE; // the number of Objects considered in one step
	public int numberObjects; // the number of objects in the initial data set

	public DataObject[] data;

	String extension;
	public FindOutlier() {

		IO ea = new IO(filePath, dim); // SPECIFY DIMENSIONALITY
		data = ea.readFile(); // NOT CLASSIFIED DATA: for kddExample
		extension = "." + STEP_SIZE + "ss." + minKNN + "NN.v"+version;
		
		getAllNN();
	}

	/**
	 * 
	 * @param minKNN
	 *            for each object the first minKNN nearest neighbor objects are determined by using the euclidean distance
	 */
	public FindOutlier(int minKNN, String pathToFile) {

	}

	/**
	 * calculating all the outlier costs as well as all the similarity values for each object o
	 * 
	 * @param data
	 *            a DataObject array containing all objects given in the initial data set
	 * @param dimension
	 *            the dimension of the input data
	 */
	public void getAllNN() {

		int noNNSize, NNSize;
		ArrayList<Integer> NNList, NoNNList;
		int index;
		double costMean, nodeCostPerEPDWeightened, min;
                double[] minimum;
		double[] outlierCost = new double[stepSize];
		int[] costObjects = new int[stepSize];
		double[] coordinates = new double[dim];
		
		// Start the runtime check of the outlier determination
		long startTime = System.currentTimeMillis();


		numberObjects = data.length;

		initializeDataObjects();
                
                int intervalNo = 10;
                stepSizelimit = new double[intervalNo];
                exp = new int[intervalNo];
		for (int i = 0; i<intervalNo; i++){
                    stepSizelimit[i] = (i+1) * (numberObjects / intervalNo);
                    exp[i] = (int) exp(intervalNo-(i+1));
                }
		
		FastVector vec = new FastVector();
		try {

			int k, j = 0;
			// for loop should be going from 0 until numberObjects considering outlier costs and
			// outlier similarity values for all objects
			for (DataObject currentObject : data) {

				System.out.println("Object " + currentObject);
//				System.out.println("Object " + currentObject.NoNNList);

				stepSize = STEP_SIZE;

				epd.reset();
				ica.reset();
				
				// while loop should be going until NoNNList.size() is 0
				while ((noNNSize = currentObject.NoNNList.size()) > 0) {

					// the stepSize is set according to the remaining number of objects in NoNNList
					updateStepSize(noNNSize);
                                    
					// NNList contains now all objects already considered as NN
					NNList = currentObject.NNList;
					NoNNList = currentObject.NoNNList;
					NNSize = NNList.size();

					nnCoordinates = new double[NNSize][];
					index = 0;
					for (Integer neighbors : NNList)
						nnCoordinates[index++] = data[neighbors].coordinates;
                                        //>>>>>>>> ICA plot
//                                        nnCoordinates = new double[data.length][];
//					index = 0;
//					for (DataObject object : data)
//						nnCoordinates[index++] = object.coordinates;
					//<<<<<<<< ICA plot
					
                                        doICA_EPD();
					
					
					// here to increase : 12% of runtime!!!
					
					outlierCost = new double[stepSize];
					costObjects = new int[stepSize];
					
					
					nodeCostPerEPDWeightened = epd.codingCost0; //(epd_afterICA.cost(whiten(currentObject.coordinates, ica)));// + ica.wscCost);;
					
					currentObject.clusterSize2Costs.put(NNSize, nodeCostPerEPDWeightened);
					currentObject.clusterCostDeviation.put(NNSize, (nodeCostPerEPDWeightened - epd.codingCostMin)); // min((y(c,:)) - clustermin(c,:));
					currentObject.clusterCostMin.put(NNSize, epd.codingCostMin);
					
					for (j = 0; j < noNNSize; j++) { 
//						coordinates = whiten(data[NoNNList.get(j)].coordinates, ica);
//						getClusterCost(outlierCost, costObjects, j, epd.cost(coordinates) + ica.wscCost);
						getClusterCost(outlierCost, costObjects, j, epd.cost(whiten(data[NoNNList.get(j)].coordinates, ica)) + ica.wscCost);
					}
					
					j = 0;
					for(int object2Move : costObjects){
						currentObject.NNList.add(NoNNList.get(object2Move));				
						currentObject.NoNNList.remove(object2Move);
						for (k = j++; k < costObjects.length; k++) {
							if (costObjects[k] > object2Move) {
								costObjects[k]--;
							}
						}
					}
				}
				
				/*
				 * for (int j = 0; j < o[i].kNNCost.size(); j++){ int x_coord = j+(minKNN+1); double y_coord = o[i].kNNCost.get(j);
				 * out_oneObject.write(x_coord+"\t"+y_coord+"\n");
				 * //System.out.println(i+" "+o[i].kNNCost.size()+" "+j+" "+x_coord+"\t"+y_coord+"\n"); } out_oneObject.close();
				 */
			}
			// print the results in a tab separated text file

			BufferedWriter out = new BufferedWriter(new FileWriter(filePath + extension + ".clusterCosts.txt"));
			BufferedWriter out5 = new BufferedWriter(new FileWriter(filePath + extension + ".clusterFactor.txt"));
			BufferedWriter out6 = new BufferedWriter(new FileWriter(filePath + extension + ".clusterFactor.arff"));
			out6.append("@relation clusterFaktor\r\n"+
					"@attribute a numeric\r\n"+
					"@data\r\n");
			Integer[] is = data[0].clusterSize2Costs.keySet().toArray(new Integer[data[0].clusterSize2Costs.keySet().size()]);
			Arrays.sort(is);
			for (Integer inti : is) {
				out.append(inti + "\t");
			}
			out.append("\r\n");
			for (DataObject dataI : data) {
				for (Integer inti : is) {
					out.append(dataI.clusterSize2Costs.get(inti) + "\t");
				}
                                minimum = minimum(dataI.clusterCostMin);
				out.append("\r\n");
				out5.append((min = dataI.clusterSize2Costs.get((int)minimum[1])-minimum[0])+"\r\n");
				 vec.addElement(min);
				out6.append(min+"\r\n");

			}

			out.close();
			out5.close();
			out6.close();
			
			out = new BufferedWriter(new FileWriter(filePath + ".objects.txt"));
			for (DataObject d : data) {
				for (double co : d.coordinates)
					out.append(co + "\t");
				out.append("\r\n");
			}
			out.close();

		} catch (IOException e) {
			e.printStackTrace();
		}

		// End the runtime check of the program
		long endTime = System.currentTimeMillis();

		// determine the runtime of the outlier calculation in hours, minutes and seconds
		GregorianCalendar g1 = new GregorianCalendar();
		GregorianCalendar g2 = new GregorianCalendar();
		g1.setTimeInMillis(startTime);
		g2.setTimeInMillis(endTime);

		int hours = g2.get(GregorianCalendar.HOUR_OF_DAY) - g1.get(GregorianCalendar.HOUR_OF_DAY);
		int minutes = g2.get(GregorianCalendar.MINUTE) - g1.get(GregorianCalendar.MINUTE);
		int seconds = g2.get(GregorianCalendar.SECOND) - g1.get(GregorianCalendar.SECOND);
		if (minutes < 0) {
			hours--;
			minutes += 60;
		}
		if (seconds < 0) {
			minutes--;
			seconds += 60;
		}
		System.out.println("Runtime: " + hours + "h " + minutes + "m " + seconds + "s xc");
		System.out.println("total : "+(endTime-startTime)/1000d);
		System.out.println("SUM "+sum);
		

		XMeansDetailed xm = new XMeansDetailed();
		try {
			
			xm.buildClusterer(new Instances(new FileReader(filePath + extension + ".clusterFactor.arff")));
			
			double[] clusterCenter = xm.getClusterCenter();
			int[] clusterAssignment = xm.getClusterAsignment();
			
			int[] clusterID = new int[clusterCenter.length];
			for(int i = 0; i < clusterID.length; i++)
				clusterID[i] = i;
			
			sortArrayAsc(clusterCenter, clusterID);
			
			int[] outlierOrder = new int[clusterAssignment.length];
			
			for(int i = clusterID.length-1; i > -1; i--){
				int order = (clusterID.length - i);
				System.out.println("Outlier of "+order+" order: ");
				for(int j = 0; j < clusterAssignment.length; j++){
					if(clusterAssignment[j] == clusterID[i]){
						System.out.print(data[j].number+", ");
						outlierOrder[j] = order;
					}
				}
				System.out.println("");
				System.out.println("");
				System.out.println("");
			}
			
			BufferedWriter out7 = new BufferedWriter(new FileWriter(filePath + extension + ".outlierOrder.txt"));
			for(int i = 0; i < outlierOrder.length; i++) {
				out7.append(outlierOrder[i]+"\r\n");
			}
			out7.close();
			
			
		} catch (Exception e) {
			e.printStackTrace();
		}	
	}


	/**
	 * fills the first minKNN objects in the o[i].NNList variable with the NN being determined using the EuclideanDistance
	 * 
	 * @param data
	 *            a DataObject array containing all objects given in the initial data set
	 */
	public void initializeDataObjects() {

		double currentMaxDistForNN = Double.MIN_VALUE;
		Integer currentMaxNumber;
		int size2Fill = minKNN + 1;

		double distance;

		for (DataObject object : data) {

			double[] minObjectDistances = new double[size2Fill];
			int[] minObjectIDs = new int[size2Fill];

			// Fill fist NN + 1 with the first dataobject that come our way ...
			for (int i = 0; i < size2Fill; i++) {
				DataObject object2 = data[i];

				minObjectDistances[i] = object.euclideanDistance(object2);
				minObjectIDs[i] = object2.number;

				object.NNList.add(object2.number);
			}

			// sort the first NN
			sortArrayAsc(minObjectDistances, minObjectIDs);
			// save the largest distance
			currentMaxDistForNN = minObjectDistances[minKNN];
			currentMaxNumber = minObjectIDs[minKNN];

			// fill the rest
			for (int i = size2Fill; i < numberObjects; i++) {
				DataObject object2 = data[i];
				distance = object.euclideanDistance(object2);
				if (currentMaxDistForNN > distance) {
					// We found a nearer object: replace the farest NN with current object

					object.NNList.remove(currentMaxNumber);
					object.NoNNList.add(currentMaxNumber);
					
					minObjectDistances[minKNN] = distance;
					minObjectIDs[minKNN] = object2.number;

					sortArrayAsc(minObjectDistances, minObjectIDs);

					currentMaxDistForNN = minObjectDistances[minKNN];
					currentMaxNumber = minObjectIDs[minKNN];

					object.NoNNList.remove(currentMaxNumber);
					object.NNList.add(object2.number);
					
				} else {
					object.NoNNList.add(object2.number);
				}
				
				object.NNList.clear();
				for (int j = 0; j < minObjectIDs.length; j++) {
					object.NNList.add(minObjectIDs[j]);
				}
				
			}
		}
	}

	/**
	 * sorts an array with given double values and an array containing integer values somehow related to the values contained in data in ascending
	 * order
	 * 
	 * @param data
	 *            an array containing double values which are considered for the sorting
	 * @param minObjects
	 *            an array containing integer values somehow related to the values contained in data
	 */
	public void sortArrayAsc(double[] data, DataObject[] minObjects) {
		for (int i = 0; i < data.length; i++) {
			for (int j = 0; j < i; j++) {
				if (data[j] > data[j + 1]) {
					DataObject tmpO = minObjects[j];
					double tmp = data[j];
					minObjects[j] = minObjects[j + 1];
					data[j] = data[j + 1];
					minObjects[j + 1] = tmpO;
					data[j + 1] = tmp;

					j = -1;
				}
			}
		}
		return;
	}
	
	public void sortArrayAsc(double[] data, int[] minObjects) {
		for (int i = 0; i < data.length; i++) {
			for (int j = 0; j < i; j++) {
				if (data[j] > data[j + 1]) {
					int tmpO = minObjects[j];
					double tmp = data[j];
					minObjects[j] = minObjects[j + 1];
					data[j] = data[j + 1];
					minObjects[j + 1] = tmpO;
					data[j + 1] = tmp;

					j = -1;
				}
			}
		}
		return;
	}

	/**
	 * // * determines the stepSize number of smallest cluster costs to the object of interest (objectNumber) // * // * @param clusterCost // * an
	 * array containing the stepSize number of smallest cluster costs // * @param costObjects // * an array containing the integer values of the
	 * objects that are related to the values contained in clusterCost at the same position // * @param j // * the number of the object considered at
	 * the moment // * @param temp_cost // * the cluster cost of the object of interest // * @param objectNumber // * the int number of the object of
	 * interest //
	 */
	 public void getClusterCost(double[] clusterCost, int[] costObjects, int j, double temp_cost) {
		// the last object of the clusterCost array is replaced if the cluster cost of interest is smaller
		if (j >= stepSize) {
			int lastObject = clusterCost.length - 1;
			if (clusterCost[lastObject] > temp_cost) {
				clusterCost[lastObject] = temp_cost;
				costObjects[lastObject] = j;
				sortArrayAsc(clusterCost, costObjects);
			}
			// the first stepSize objects are added as initial clusterCost and costObjects array
		} else if (j < stepSize) {
			clusterCost[j] = temp_cost;
			costObjects[j] = j;
			if (j == (stepSize - 1)) {
				sortArrayAsc(clusterCost, costObjects);
			}
		}
	 }



	 
	/**
	 * Nikola & Katrin: Bringing data into white space, given ICA parameter ...
	 * 
	 * @param dataPoint
	 * @param bestResult
	 * @return
	 */
	 
	public double[] whiten(double[] dataPoint, ICA bestResult) { // 0.4 % of runtime

		
		
		for (int i = 0; i < dim; i++)
			centered[0][i] = dataPoint[i] - bestResult.shift[i];

		mdata = new Matrix(centered);

		// WHITEN: from some whitenData() code from ICA
		whitened = mdata.times(bestResult.transformMatrix);

		// ROTATE from claudia
		whitrot = whitened.times(bestResult.transposedMW).getArrayCopy();
		return whitrot[0];
		
	}
	Matrix mdata,whitened;

//	double sum = 0, sum2 = 0;;
	
	private double[] s2; //s1
	private double[][] nnCoordinates, nnProjectedWhitened, centered = new double[1][dim], whitrot;
	private int currentNumberObjects, j, j2;
	double sumCost_AfterICA, currentCost;
	
	// COST: writes coding cost using ICA and EPD
	public void doICA_EPD() { // 77% runtime
		
		currentNumberObjects = nnCoordinates.length;
		
		if (currentNumberObjects == 0) {
			// empty cluster due to split
			System.err.println("empty");
		}

		long t = System.currentTimeMillis();
		ica.doICA(nnCoordinates, currentNumberObjects);
		
//		ica.init(dim, nnCoordinates);
//		ica.centerData();
//		ica.whitenData();
//		ica.optimizeEntropy();
		
		sum += System.currentTimeMillis() - t;
		
                nnProjectedWhitened = (ica.whithenedMatrix.times(ica.mw.transpose())).getArrayCopy();
                              
		// estimate epd (with rotation) and unrotated ...
		for (j = 0; j < dim; j++) {
//			s1 = new double[currentNumberObjects];
			s2 = new double[currentNumberObjects];
			for (int k = 0; k < currentNumberObjects; k++) {
//				s1[k] = nnCoordinates[k][j];
				s2[k] = nnProjectedWhitened[k][j];
			}
			
			epd.estimate(s2, j);
			
		}
		
		// get Costs for rotated and not rotated version.
//		sumCost_AfterICA = 0d;
//		epd.codingCostObjects = new double[currentNumberObjects];
		
		
		tmp_cost = 0d;
		for (j2 = 0; j2 < dim; j2++) {
			tmp_cost += epd.cost(nnProjectedWhitened[0][j2], j2);
		}
		tmp_cost += ica.wscCost;
		codingCostMin = tmp_cost;
		epd.codingCost0 = tmp_cost;
		
		
		for ( j = 1; j < currentNumberObjects; j++) {
			tmp_cost = 0d;
			for (j2 = 0; j2 < dim; j2++) {
//				currentCost = epd.cost(nnProjectedWhitened[j][j2], j2);
//				sumCost_AfterICA += currentCost;
//				epd.codingCostObjects[j] += currentCost;
				tmp_cost += epd.cost(nnProjectedWhitened[j][j2], j2);
			}
//			epd.codingCostObjects[j] += ica.wscCost;
			tmp_cost += ica.wscCost;
			codingCostMin = min(codingCostMin, tmp_cost);
		}
		epd.codingCostMin = codingCostMin;
		
//		sumCost_AfterICA += ica.wscCost * currentNumberObjects;
		
	}
	
	double sum;
	
	double tmp_cost, codingCostMin;

	public Matrix rowVector(double[] v) {
		return new Matrix(v, 1);
	}

        	
	
	double[] stepSizelimit;
	int[] exp;
	/**
	 * determine the stepSize using the size of the data set and the size of the NoNNList which is the variable size
	 * 
	 * @param noNNsize
	 *            the size of the NoNNList at the moment
	 */
	public void updateStepSize(int noNNsize) {
		// only if the new stepSize is larger than the size of the NoNNList the stepSize object is
		// set to the exponent else it is set to the size of the NoNNList
		// the data set is devided into 6 equal parts the first 1*(numberObjects/6) outliers are calculated using a
		// setSize of one the second 2*(numberObjects/6) outliers are calculated using a stepSize of
		// exp(2) etc.		
		
		if (noNNsize <= stepSize) {
			stepSize = noNNsize;
			return;
                } else if (noNNsize <= stepSizelimit[0]) {
			stepSize = exp[0];
                } else if (noNNsize <= stepSizelimit[1]) {
			stepSize = exp[1];                
                } else if (noNNsize <= stepSizelimit[2]) {
			stepSize = exp[2];                
                } else if (noNNsize <= stepSizelimit[3]) {
			stepSize = exp[3];
                } else if (noNNsize <= stepSizelimit[4]) {
			stepSize = exp[4];
		} else if (noNNsize <= stepSizelimit[5]) {
			stepSize = exp[5];
		} else if (noNNsize <= stepSizelimit[6]) {
			stepSize = exp[6];
		} else if (noNNsize <= stepSizelimit[7]) {
			stepSize = exp[7];
		} else if (noNNsize <= stepSizelimit[8]) {
			stepSize = exp[8];
		} else if (noNNsize <= stepSizelimit[9]) {
			stepSize = exp[9];
		}
		if (noNNsize <= stepSize) {
			stepSize = noNNsize;
		}
	}
	

	private double[] minimum(Hashtable<Integer, Double> clusterCostDeviation) {
		double[] min = {Double.MAX_VALUE, 0}; // min[0] is the clustercost, min[1] NNnumber of the object having the lowest clusterCost

		for (Integer val : clusterCostDeviation.keySet()){
			min[0] = min(min[0],clusterCostDeviation.get(val));
                        if (min[0] == clusterCostDeviation.get(val))
                                min[1] = val;
                                
                        
                }

		return min;
	}

	public static void main(String[] args) {
		new FindOutlier();

	}
	
	

}
