package outlier.byEPD.bottomUp;

import java.util.*;
import java.io.*;

import com.sun.org.apache.xml.internal.security.Init;

import oci.DB;
import oci.DataObject;
import oci.EpdMineo;
import oci.GenGauss;
import oci.IO;
import oci.Ica;
import oci.IcaCluster;
import Jama.*;

/**
 * Calculates the outlier costs as well as the outlier similarity value for each object starting with initialisation of minKNN objects be determining
 * the euclidean distance and the rest using ICA and EPD or only using EPD
 * 
 * @author khaegler
 * @param minKNN
 *            for each object the first minKNN nearest neighbor objects are determined by using the euclidean distance
 * @param dim
 *            the dimension of the input data
 * @param stepSize
 *            the number of nearest neighbor objects determined in one step
 * @param numberObjects
 *            the total number of objects given
 */
public class KNN {


	String filePath;
	
	
	
	public int minKNN; // for each object the first minKNN nearest neighbor objects are determined by using the euclidean distance
	public int dim; // the dimension of the input data
	public final int STEP_SIZE = 1; // the number of Objects considered in one step
	public int stepSize = STEP_SIZE; // the number of Objects considered in one step
	public int numberObjects; // the number of objects in the initial data set
	public boolean clusterCost = true;

	String extension;
	
	
	/**
	 * 
	 * @param minKNN
	 *            for each object the first minKNN nearest neighbor objects are determined by using the euclidean distance
	 */
	public KNN(int minKNN, String pathToFile) {
		this.minKNN = minKNN;
		this.filePath = pathToFile;
		extension = "."+STEP_SIZE+"ss."+minKNN+"NN.v2";
	}

	/**
	 * calculating all the outlier costs as well as all the similarity values for each object o
	 * 
	 * @param o
	 *            a DataObject array containing all objects given in the initial data set
	 * @param dimension
	 *            the dimension of the input data
	 */
	public void getAllNN(DataObject[] o, int dimension) {
		GenGauss gg = new GenGauss();
		// Start the runtime check of the outlier determination
		long startTime = System.currentTimeMillis();

		this.dim = dimension;
		numberObjects = o.length;
		// fills the first minKNN objects in the o[i].NNList variable with the NN being determined using the EuclideanDistance
		setInitialNNs(o);

		// The outlier cost has to be calculated for all nearest neighbors > minKNN
		// the first minKNN points are already determined and the point itself doesn't have to be considered
		// therefore the initial cluster size is minKNN+1
		// the rest of the objects not yet assigned to the NNList are inserted in the o[i].NoNNList variable
		for (int i = 0; i < numberObjects; i++) {
			ArrayList<Integer> noNN = new ArrayList<Integer>(); // will contain the objects not yet assigned to the NNList
			for (int j = 0; j < numberObjects; j++) {
				boolean iNN = false;
				for (int k = 0; k < (minKNN + 1); k++) {
					if (o[j].number == o[i].NNList.get(k)) {
						iNN = true;
						break;
					}
				}
				if (!iNN) {
					// The objects not yet assigned to the NNList
					noNN.add(o[j].number);
				}
			}
			o[i].NoNNList = noNN;
		}

		try {
			// for loop should be going from 0 until numberObjects considering outlier costs and
			// outlier similarity values for all objects
			for (int i = 0; i < numberObjects; i++) {
				/*
				 * String filename_oneObject = "KNN_clustercost"+i+".txt"; BufferedWriter out_oneObject = new BufferedWriter(new
				 * FileWriter(filename_oneObject));
				 */
				
				stepSize = STEP_SIZE;
				
				o[i].kNNCost = new ArrayList<Double>(); // will contain outlier costs of numberObjects-(minKNN+1)
				o[i].kNNSimilarity = new ArrayList<Double>(); // will contain outlier similarities of numberObjects-(minKNN+1)
				// while loop should be going until NoNNList.size() is 0
				while (o[i].NoNNList.size() > 0) {
//					System.out.println("Object " + o[i].number + " " + o[i].NoNNList.size() + " NN left.");

					// int NNid = 0;
					// double clusterCost = Double.MAX_VALUE;

					// the stepSize is set according to the remaining number of objects in NoNNList
//					setStepSize(o[i].NoNNList.size());

					// insert all coord in the coord variable
					double[][] coord = new double[o[i].NNList.size()][];
					for (int k = 0; k < o[i].NNList.size(); k++) {
						coord[k] = o[o[i].NNList.get(k)].coord.clone();
					}
					// NNList contains now all objects already considered as NN
					ArrayList<Integer> NNList =  o[i].NNList;

					IcaCluster resultNoRot = new IcaCluster();
					IcaCluster resultRot = new IcaCluster();
					IcaCluster bestResult = new IcaCluster();
                                        
                           //********** WICHTIG: nicht vergessen im Fall von resultNoRot findet keine Rotation statt, d.h. auch kein whitening
                                        boolean whitening = true;
//					// calculate the cluster cost without using ICA only using EPD
//					resultNoRot = mergedClusterNoRot(NNList, coord);
//					// calculate the cluster cost using ICA and EPD
//					resultRot = mergedCluster(NNList, coord);
//					// only the method producing the lower cluster costs is considered in the following
//					if (resultNoRot.cost < resultRot.cost) {
//						bestResult = resultNoRot;
//                                              whitening = false;
//					} else {
//						bestResult = resultRot;
//					}
					bestResult = mergedCluster(NNList, coord);

					
					

					
					double[] outlierCost = new double[stepSize];
					int[] costObjects = new int[stepSize];
					double[] outlierSimilarities = new double[stepSize];
					int[] similarityObjects = new int[stepSize];
					// find out which objects from the NoNNList have the lowest outlier costs or the highest similarity values
					// and keep the stepSize best ones
					
					double[][] ggparam = new double[dim][3];
					for (int j2 = 0; j2 < dim; j2++) {
						ggparam[j2][0] = bestResult.realCenter[j2];
						ggparam[j2][1] = bestResult.realRotation[j2][j2];
						ggparam[j2][2] = bestResult.beta[j2];
					}
                                        /*
                                        double[] whitened = new double[dim];
					if (whitening){
                                            whitened = whiten(o[i].coord, bestResult);
                                        }else{
                                            whitened = o[i].coord.clone();
                                        }
                                        
                                        
					
					for (int j2 = 0; j2 < dim; j2++) {
////						nodeCostPerEPD += gg.costV(o[i].coord[j2], ggparam[j2]);
//						nodeCostPerEPD += gg.costV(o[i].coord[j2], ggparam[j2]);
						double dimCost = gg.costV(whitened[j2], ggparam[j2]);
//						System.out.println(dimCost);
						nodeCostPerEPDWeightened += dimCost;
					}
//					System.out.println("Obj 0 Whit?: "+whitened[0]+"\t"+whitened[1]);
					// TODO: they should be equal, but are not.
					//System.out.println(nodeCostPerEPDWeightened+"\t"+bestResult.codingCostObjects[0]);
					*/

					double nodeCostPerEPDWeightened = 0;
					nodeCostPerEPDWeightened = bestResult.codingCostObjects[0]; 
					
					o[i].clusterSize2Costs.put(NNList.size(), nodeCostPerEPDWeightened);
					double costMean;
					o[i].clusterSize2ClusterCostMean.put(NNList.size(), (costMean = mean(bestResult.codingCostObjects)));
					o[i].clusterSize2ClusterCostVariance.put(NNList.size(), stddev(bestResult.codingCostObjects, costMean));
					
					for (int j = 0; j < o[i].NoNNList.size(); j++) {
						
						if (clusterCost) {
							double temp_cost = 0;
                                                        double[] whitened = new double[dim];
                                                        if (whitening){
                                                            whitened = whiten(o[o[i].NoNNList.get(j)].coord, bestResult);
                                                        }else{
                                                            whitened = o[o[i].NoNNList.get(j)].coord.clone();
                                                        }
                                                        
							for (int j2 = 0; j2 < dim; j2++) {
								// the three gauss parameters: the mean, standard deviation and beta
								// get costs for adding object j to the cluster
                                                                double aktCost = gg.costV(whitened[j2], ggparam[j2]);
//								double aktCost = gg.costV(o[o[i].NoNNList.get(j)].coord[j2], ggparam[j2]);
								// add the costs of each dimension
								temp_cost += aktCost;
							}
							// check if the object of interest has smaller outlier costs as the
							// ones contained in the outlierCost array
							getClusterCost(outlierCost, costObjects, j, temp_cost);
						}else{
							double temp_similarity = 0;
							for (int j2 = 0; j2 < dim; j2++) {
								double aktSimilarity = gg.similarityV(o[o[i].NoNNList.get(j)].coord[j2], ggparam[j2]);
								// add the costs of each dimension
								temp_similarity += aktSimilarity;
							}
							// check if the object of interest has higher outlier similarity as the
							// ones contained in the outlierCost array
							getSimilarityValue(outlierSimilarities, similarityObjects, j, temp_similarity);
						}
					}

					// insert the outlier cost and similarity in the predefined objects variables
					for (int j = 0; j < stepSize; j++) {
						o[i].kNNCost.add(outlierCost[j]);
						o[i].kNNSimilarity.add(outlierSimilarities[j]);

						if (clusterCost) {
							o[i].NNList.add(o[i].NoNNList.get(costObjects[j]));
							o[i].NoNNList.remove(costObjects[j]);
							for (int k = j; k < costObjects.length; k++) {
								if (costObjects[k] > costObjects[j]) {
									costObjects[k]--;
								}
							}
						} else {
							o[i].NNList.add(o[i].NoNNList.get(similarityObjects[j]));
							o[i].NoNNList.remove(similarityObjects[j]);
							for (int k = j; k < similarityObjects.length; k++) {
								if (similarityObjects[k] > similarityObjects[j]) {
									similarityObjects[k]--;
								}
							}
						}
					}
				}
				/*
				 * for (int j = 0; j < o[i].kNNCost.size(); j++){ int x_coord = j+(minKNN+1); double y_coord = o[i].kNNCost.get(j);
				 * out_oneObject.write(x_coord+"\t"+y_coord+"\n");
				 * //System.out.println(i+" "+o[i].kNNCost.size()+" "+j+" "+x_coord+"\t"+y_coord+"\n"); } out_oneObject.close();
				 */
                                boolean isOutlier = isOutlier(o[i]);
                                o[i].isOutlier = isOutlier;
                                
				System.out.println("Object " + o[i]+"; Outliercount "+o[i].outlierCount);
			}
			// print the results in a tab separated text file
			
			BufferedWriter out = new BufferedWriter(new FileWriter(filePath+extension+".knnCost.txt"));

			int NNsize = 0;
			if (clusterCost) {
				NNsize = o[0].kNNCost.size();
			} else {
				NNsize = o[0].kNNSimilarity.size();
			}
			for (int i = 0; i < NNsize; i++) {
				for (int j = 0; j < numberObjects; j++) {
					int x_coord = i + (minKNN + 1);
					if (clusterCost) {
						double y_coord = o[j].kNNCost.get(i);
						if (j == 0) {
							out.write(x_coord + "\t" + y_coord);
						} else if (j == (numberObjects - 1)) {
							out.write("\t" + y_coord + "\n");
						} else {
							out.write("\t" + y_coord);
						}
					} else {
						double y_coord = o[j].kNNSimilarity.get(i);
						if (j == 0) {
							out.write(x_coord + "\t" + y_coord);
						} else if (j == (numberObjects - 1)) {
							out.write("\t" + y_coord + "\n");
						} else {
							out.write("\t" + y_coord);
						}
					}
				}
			}

			out.close();

			out = new BufferedWriter(new FileWriter(filePath+extension+".clusterCosts.txt"));
			BufferedWriter out2 = new BufferedWriter(new FileWriter(filePath+extension+".clusterMean.txt"));
			BufferedWriter out3 = new BufferedWriter(new FileWriter(filePath+extension+".clusterStddev.txt"));
			Integer[] is = o[0].clusterSize2Costs.keySet().toArray(new Integer[o[0].clusterSize2Costs.keySet().size()]);
			Arrays.sort(is);
			for (Integer inti : is){
				out.append(inti + "\t");
				out2.append(inti + "\t");
				out3.append(inti + "\t");
			}
			out.append("\r\n");
			out2.append("\r\n");
			out3.append("\r\n");
			for (DataObject dataI : o) {
				for (Integer inti : is){
					out.append(dataI.clusterSize2Costs.get(inti) + "\t");
					out2.append(dataI.clusterSize2ClusterCostMean.get(inti) + "\t");
					out3.append(dataI.clusterSize2ClusterCostVariance.get(inti) + "\t");
				}
				out.append("\r\n");
				out2.append("\r\n");
				out3.append("\r\n");
			}

			out.close();
			out2.close();
			out3.close();
			
			out = new BufferedWriter(new FileWriter(filePath+".objects.txt"));
			for(DataObject d : o){
				for(double co : d.coord)
					out.append(co+"\t");
				out.append("\r\n");
			}
			out.close();
			
			
		} catch (IOException e) {
		}

		// End the runtime check of the program
		long endTime = System.currentTimeMillis();

		// determine the runtime of the outlier calculation in hours, minutes and seconds
		GregorianCalendar g1 = new GregorianCalendar();
		GregorianCalendar g2 = new GregorianCalendar();
		g1.setTimeInMillis(startTime);
		g2.setTimeInMillis(endTime);

		int hours = g2.get(GregorianCalendar.HOUR_OF_DAY) - g1.get(GregorianCalendar.HOUR_OF_DAY);
		int minutes = g2.get(GregorianCalendar.MINUTE) - g1.get(GregorianCalendar.MINUTE);
		int seconds = g2.get(GregorianCalendar.SECOND) - g1.get(GregorianCalendar.SECOND);
		if (minutes < 0) {
			hours--;
			minutes += 60;
		}
		if (seconds < 0) {
			minutes--;
			seconds += 60;
		}
		System.out.println("Runtime: " + hours + "h " + minutes + "m " + seconds + "s xc");

		return;
	}

	/**
	 * fills the first minKNN objects in the o[i].NNList variable with the NN being determined using the EuclideanDistance
	 * 
	 * @param o
	 *            a DataObject array containing all objects given in the initial data set
	 */
	public void setInitialNNs(DataObject[] o) {

		DataObject[] q = o.clone();
		for (int i = 0; i < numberObjects; i++) {
			double[] minDistances = new double[minKNN + 1];
			int[] minObjects = new int[minKNN + 1];
			for (int j = 0; j < numberObjects; j++) {
				double dist = o[i].euclideanDistance(q[j]);
				// if dist is smaller than the last element in minDistances (the object having
				// the largest distance to the object i) it is replaced with dist
				if (j >= (minKNN + 1)) {
					int lastObject = minDistances.length - 1;
					if (minDistances[lastObject] > dist) {
						minDistances[lastObject] = dist;
						minObjects[lastObject] = q[j].number;
						sortArrayAsc(minDistances, minObjects);
					}
					// the first minKNN objects are added as initial minDistances
				} else if (j < (minKNN + 1)) {
					minDistances[j] = dist;
					minObjects[j] = q[j].number;
					if (j == minKNN) {
						// as soon as the first minKKN euclidean distances are determined the array is sorted ascending
						// with the smallest distance being the first element and the largest being the last
						sortArrayAsc(minDistances, minObjects);
					}
				}
			}
			ArrayList<Integer> tempObj = new ArrayList<Integer>();
			for (int j = 0; j < minObjects.length; j++) {
				tempObj.add(minObjects[j]);
			}
			o[i].NNList = tempObj;
		}

		return;
	}

	/**
	 * sorts an array with given double values and an array containing integer values somehow related to the values contained in data in descending
	 * order
	 * 
	 * @param data
	 *            an array containing double values which are considered for the sorting
	 * @param minObjects
	 *            an array containing integer values somehow related to the values contained in data
	 */
	public void sortArrayDesc(double[] data, int[] minObjects) {
		for (int i = 0; i < data.length; i++) {
			for (int j = 0; j < i; j++) {
				if (data[j] < data[j + 1]) {
					int tmpO = minObjects[j];
					double tmp = data[j];
					minObjects[j] = minObjects[j + 1];
					data[j] = data[j + 1];
					minObjects[j + 1] = tmpO;
					data[j + 1] = tmp;

					j = -1;
				}
			}
		}
		return;
	}

	/**
	 * sorts an array with given double values and an array containing integer values somehow related to the values contained in data in ascending
	 * order
	 * 
	 * @param data
	 *            an array containing double values which are considered for the sorting
	 * @param minObjects
	 *            an array containing integer values somehow related to the values contained in data
	 */
	public void sortArrayAsc(double[] data, int[] minObjects) {
		for (int i = 0; i < data.length; i++) {
			for (int j = 0; j < i; j++) {
				if (data[j] > data[j + 1]) {
					int tmpO = minObjects[j];
					double tmp = data[j];
					minObjects[j] = minObjects[j + 1];
					data[j] = data[j + 1];
					minObjects[j + 1] = tmpO;
					data[j + 1] = tmp;

					j = -1;
				}
			}
		}
		return;
	}

	/**
	 * determines the stepSize number of smallest cluster costs to the object of interest (objectNumber)
	 * 
	 * @param clusterCost
	 *            an array containing the stepSize number of smallest cluster costs
	 * @param costObjects
	 *            an array containing the integer values of the objects that are related to the values contained in clusterCost at the same position
	 * @param j
	 *            the number of the object considered at the moment
	 * @param temp_cost
	 *            the cluster cost of the object of interest
	 * @param objectNumber
	 *            the int number of the object of interest
	 */
	public void getClusterCost(double[] clusterCost, int[] costObjects, int j, double temp_cost) {
		// the last object of the clusterCost array is replaced if the cluster cost of interest is smaller
		if (j >= stepSize) {
			int lastObject = clusterCost.length - 1;
			if (clusterCost[lastObject] > temp_cost) {
				clusterCost[lastObject] = temp_cost;
				costObjects[lastObject] = j;
				sortArrayAsc(clusterCost, costObjects);
			}
			// the first stepSize objects are added as initial clusterCost and costObjects array
		} else if (j < stepSize) {
			clusterCost[j] = temp_cost;
			costObjects[j] = j;
			if (j == (stepSize - 1)) {
				sortArrayAsc(clusterCost, costObjects);
			}
		}
	}

	/**
	 * determines the stepSize number of largest outlier similarities of the object of interest (objectNumber)
	 * 
	 * @param similarities
	 *            an array containing the stepSize number of largest outlier similarities
	 * @param similarityObjects
	 *            an array containing the integer values of the objects that are related to the values contained in similarities at the same position
	 * @param j
	 *            the number of the object considered at the moment
	 * @param temp_similarity
	 *            the outlier similarity of the object of interest
	 * @param objectNumber
	 *            the int number of the object of interest
	 */
	public void getSimilarityValue(double[] similarities, int[] similarityObjects, int j, double temp_similarity) {
		// the last object of the similarityObjects array is replaced if the outlier similarity of interest is larger
		if (j >= stepSize) {
			int lastObject = similarities.length - 1;
			if (similarities[lastObject] < temp_similarity) {
				similarities[lastObject] = temp_similarity;
				similarityObjects[lastObject] = j;
				sortArrayDesc(similarities, similarityObjects);
			}
			// the first stepSize objects are added as initial similarities and the similarityObjects array
		} else if (j < stepSize) {
			similarities[j] = temp_similarity;
			similarityObjects[j] = j;
			if (j == (stepSize - 1)) {
				sortArrayDesc(similarities, similarityObjects);
			}
		}
	}

	
	
	
	// COST: writes coding cost using only EPD no ICA
	public IcaCluster mergedClusterNoRot(ArrayList<Integer> assignedObjects, double[][] d) {

		IcaCluster result = new IcaCluster();

		int currentNumberObjects = assignedObjects.size();
		result.codingCostObjects = new double[currentNumberObjects];
		result.assignedObjects = new int[currentNumberObjects];
		for (int i = 0; i < currentNumberObjects; i++) {
			result.assignedObjects[i] = assignedObjects.get(i);
		}

		if (currentNumberObjects == 0) {
			// empty cluster due to split
			System.err.println("empty");
		}

		double[] realCenter = new double[dim]; // center of cluster pdf
		double[][] realRotation = new double[dim][dim]; // rotation+scaling
		double[] beta = new double[dim];
		GenGauss gg = new GenGauss();
		// Ica ica = new Ica();
		/*
		 * double[][] d = new double[currentNumberObjects][]; for (int j = 0; j < currentNumberObjects; j++) { d[j] = o[j].coord.clone(); }
		 */
		// double[][] transform = new double[dim][dim];
		// double[][] retransform = new double[dim][dim];
		// double[] shift = new double[dim];
		// double centered[][] = ica.centerData(d, shift);
		// double whitened[][] = ica.whitenData(centered, transform, retransform);
		double bestcost = Double.MAX_VALUE;
		// Matrix localbestmw = new Matrix(1,1) ;
		double[][] localbestgenGaussParams = new double[0][];

		// double[][] w = ica.optimizeEntropy(whitened);
		// Matrix mwhitened = new Matrix(whitened);
		// Matrix mw = new Matrix(w);
		// double[][] projectedWhitened =( mwhitened.times(mw.transpose())).transpose().getArrayCopy();
		double[][] genGaussParams = new double[dim][];
		// double[] stdGauss = {0.0,1.0,0.0} ;
		// now determine the shape parameter for every dimension
		for (int j = 0; j < dim; j++) {
			double[] s = new double[currentNumberObjects];
			for (int k = 0; k < currentNumberObjects; k++) {
				s[k] = d[k][j];
			}
			genGaussParams[j] = EpdMineo.optimizeMSP(s);
		}
		double cost = 0.0;
		// determine cost for every object in d
		for (int j = 0; j < currentNumberObjects; j++) {
			for (int j2 = 0; j2 < dim; j2++) {
				double aktCost = gg.costV(d[j][j2], genGaussParams[j2]);
				cost += aktCost;
				result.codingCostObjects[j] += aktCost;
			}
		}
		if (bestcost > cost) {
			bestcost = cost;
			// localbestmw = mw ;
			localbestgenGaussParams = new double[dim][];
			for (int j = 0; j < dim; j++) {
				localbestgenGaussParams[j] = genGaussParams[j].clone();
			}
			double[] mu = new double[dim];
			for (int j = 0; j < mu.length; j++) {
				mu[j] = localbestgenGaussParams[j][0];
			}
			realCenter = mu;
			double[][] sigmaDiag = new double[dim][dim];
			for (int j = 0; j < dim; j++) {
				sigmaDiag[j][j] = 1.0 / localbestgenGaussParams[j][1];
			}
			realRotation = sigmaDiag;
			for (int j = 0; j < dim; j++) {
				beta[j] = localbestgenGaussParams[j][2];
			}
		}

		// double pCostNotRot = dim * 32 * 2;
		double pCostNotRot = 0;

		result.numObj = currentNumberObjects;
		result.dim = dim;
		result.realCenter = realCenter;
		result.realRotation = realRotation;
		result.beta = beta;
		result.cost = bestcost + pCostNotRot;

		result.codingCost = bestcost;
		return result;
	}

	/**
	 * Nikola & Kathrin: Bringing data into white space, given ICA parameter ...
	 * @param dataPoint
	 * @param bestResult
	 * @return
	 */
	public double[] whiten(double[] dataPoint, IcaCluster bestResult){
		
//		1) in den white space bringen, d.h. Zentrieren und PCA, 
//		2) nach ICA rotieren, und dann kann man den Punkt mit den geringsten Kosten bestimmen.
//		double centered[][] = ica.centerData(d, shift); //davon braucht ihr den Mittelwertvektor shift
//		double whitened[][] = ica.whitenData(centered, transform, retransform); //das ist das whitening: davon braucht ihr die Matrix transform, die die Daten in den white space überführt, also Schritt 1)
//		double[][] projectedWhitened =( mwhitened.times(mw.transpose())).transpose().getArrayCopy(); //das rotiert den Raum nach ICA, Schritt 2) man braucht also die Matrix mw. 
		
		
		// center data: (code from centerData()
		double[][] centered = new double [1][dim] ;
		for(int i = 0; i < dim; i++)
			centered[0][i] = dataPoint[i] - bestResult.shift[i] ;
		
		Matrix mdata = new Matrix(centered);
		
		// WHITEN: from some whitenData() code from ICA
		Matrix whitened = mdata.times(new Matrix(bestResult.transform));
		
		// ROTATE from claudia
		double[][] whitrot = whitened.times(bestResult.mw.transpose()).getArrayCopy();
		 					
		return whitrot[0];
		
	}
	
	// COST: writes coding cost using ICA and EPD
	public IcaCluster mergedCluster(ArrayList<Integer> assignedObjects, double[][] d) {

		IcaCluster result = new IcaCluster();

		int currentNumberObjects = assignedObjects.size();
		result.codingCostObjects = new double[currentNumberObjects];
		result.assignedObjects = new int[currentNumberObjects];
		for (int i = 0; i < currentNumberObjects; i++) {
			result.assignedObjects[i] = assignedObjects.get(i);
		}

		if (currentNumberObjects == 0) {
			// empty cluster due to split
			System.err.println("empty");
		}

		double[] realCenter = new double[dim]; // center of cluster pdf
		double[][] realRotation = new double[dim][dim]; // rotation+scaling
		double[] beta = new double[dim];
		GenGauss gg = new GenGauss();
		// gg = new GenGaussOld() ;
		Ica ica = new Ica();
		/*
		 * double[][] d = new double[currentNumberObjects][]; for (int j = 0; j < currentNumberObjects; j++) { d[j] = o[j].coord.clone(); }
		 */
		

		
		result.transform = new double[dim][dim];
		double[][] retransform = new double[dim][dim];
		result.shift = new double[dim];
		
		// (1/3) davon braucht ihr den Mittelwertvektor shift
		double centered[][] = ica.centerData(d, result.shift);
		
		// (2/3) das ist das whitening: davon braucht ihr die Matrix transform, die die Daten in den white space überführt, also Schritt 1)
		double whitened[][] = ica.whitenData(centered, result.transform, retransform);
		
		double bestcost = Double.MAX_VALUE;
		Matrix localbestmw = new Matrix(1, 1);
		double[][] localbestgenGaussParams = new double[0][];

		double[][] w = ica.optimizeEntropy(whitened);
		Matrix mwhitened = new Matrix(whitened); // associated with the dataPoints
		result.mw = new Matrix(w);
		
		// (3/3) das rotiert den Raum nach ICA, Schritt 2) man braucht also die Matrix mw.
		double[][] projectedWhitened = (mwhitened.times(result.mw.transpose())).transpose().getArrayCopy();
		
		double[][] genGaussParams = new double[dim][];
		double[] stdGauss = { 0.0, 1.0, 0.0 };
		// now determine the shape parameter for every dimension
		for (int j = 0; j < dim; j++) {
			try {
				// genGaussParams [j] = gg.optimizeMSB(projectedWhitened [j]) ; //GenGauss
				genGaussParams[j] = EpdMineo.optimizeMSP(projectedWhitened[j]); // GenGauss
				for (int k = 0; k < genGaussParams[j].length; k++) {
					if (Double.isNaN(genGaussParams[j][k])) {
						System.out.println("------genGaussParams[" + j + "][" + k + "] NaN-----------");
						genGaussParams[j] = stdGauss;
					}
				}
			} catch (java.lang.StackOverflowError e) {
				System.out.println("-----stack overflow error-------");
				genGaussParams[j] = stdGauss;
			}
		}
		
//		System.out.println("Obj 0 White: "+projectedWhitened[0][0]+"\t"+projectedWhitened[1][0]);
		
		double cost = 0.0;
		// determine cost for every object in d
		for (int j = 0; j < currentNumberObjects; j++) {
			for (int j2 = 0; j2 < dim; j2++) {
				double aktCost = gg.costV(projectedWhitened[j2][j], genGaussParams[j2]);
				cost += aktCost;
				result.codingCostObjects[j] += aktCost;
//				System.out.println(aktCost);
			}
			result.codingCostObjects[j] += ica.wscCost;
		}
		cost += ica.wscCost * currentNumberObjects;
		
		
		
		if (bestcost > cost) {
			bestcost = cost;
			localbestmw = result.mw;
			localbestgenGaussParams = new double[dim][];
			for (int j = 0; j < dim; j++) {
				localbestgenGaussParams[j] = genGaussParams[j].clone();
			}
			double[] mu = new double[dim];
			for (int j = 0; j < mu.length; j++) {
				mu[j] = localbestgenGaussParams[j][0];
			}
//			realCenter = (rowVector(result.shift).plus(rowVector(mu).times(localbestmw.times(new Matrix(retransform)).transpose()))).getRowPackedCopy();
			realCenter = mu;
			double[][] sigmaDiag = new double[dim][dim];
			for (int j = 0; j < dim; j++) {
//				sigmaDiag[j][j] = 1.0 / localbestgenGaussParams[j][1];
				sigmaDiag[j][j] = localbestgenGaussParams[j][1];
			}
//			realRotation = new Matrix(result.transform).times(localbestmw.transpose()).times(new Matrix(sigmaDiag)).getArrayCopy();
			realRotation = sigmaDiag;
			for (int j = 0; j < dim; j++) {
				beta[j] = localbestgenGaussParams[j][2];
			}
			
		}
		
		
		
		// double pCostRot = dim * 32 * 2 + dim * dim * 32;
		double pCostRot = 0;

		
		result.numObj = currentNumberObjects;
		result.dim = dim;
		result.realCenter = realCenter;
		result.realRotation = realRotation;
		result.beta = beta;
		result.cost = bestcost + pCostRot;
		result.codingCost = bestcost;

		return result;
	}

	public Matrix rowVector(double[] v) {
		return new Matrix(v, 1);
	}

	/**
	 * determine the stepSize using the size of the data set and the size of the NoNNList which is the variable size
	 * 
	 * @param size
	 *            the size of the NoNNList at the moment
	 */
	public void setStepSize(int size) {
		// only if the new stepSize is larger than the size of the NoNNList the stepSize object is
		// set to the exponent else it is set to the size of the NoNNList
		// the data set is devided into 6 equal parts the first 1*(numberObjects/6) outliers are calculated using a
		// setSize of one the second 2*(numberObjects/6) outliers are calculated using a stepSize of
		// exp(2) etc.
		if (size <= stepSize) {
			stepSize = size;
		} else if (size <= 1 * (numberObjects / 6)) {
			stepSize = (int) Math.exp(5);
			if (size <= stepSize) {
				stepSize = size;
			}
		} else if (size <= 2 * (numberObjects / 6)) {
			stepSize = (int) Math.exp(4);
			if (size <= stepSize) {
				stepSize = size;
			}
		} else if (size <= 3 * (numberObjects / 6)) {
			stepSize = (int) Math.exp(3);
			if (size <= stepSize) {
				stepSize = size;
			}
		} else if (size <= 4 * (numberObjects / 6)) {
			stepSize = (int) Math.exp(2);
			if (size <= stepSize) {
				stepSize = size;
			}
		} else if (size <= 5 * (numberObjects / 6)) {
			stepSize = (int) Math.exp(1);
			if (size <= stepSize) {
				stepSize = size;
			}
		}
	}

	public double mean(double[] values){
		double mean = 0d;
		for(double val : values)
			mean += val;
		mean /= values.length;
		return mean;
	}
	
	public double stddev(double[] values, double mean){
		double sigma = 0d;
		for(double val : values)
			sigma += Math.pow(val - mean, 2);
		sigma = Math.sqrt(sigma / values.length);
		return sigma;
	}
        
        public boolean isOutlier(DataObject o){
            
            boolean isOutlier = false;
            double outlierCount = 0;
            for (int i = 0; i < numberObjects; i++){
                if (o.clusterSize2Costs.containsKey(i) && o.clusterSize2Costs.get(i) < o.clusterSize2ClusterCostMean.get(i) - 3 * o.clusterSize2ClusterCostVariance.get(i)){
                    outlierCount++;
                }else if (o.clusterSize2Costs.containsKey(i) && o.clusterSize2Costs.get(i) > o.clusterSize2ClusterCostMean.get(i) + 3 * o.clusterSize2ClusterCostVariance.get(i)){
                    outlierCount++;
                }
            }
            if (outlierCount > 0){
                isOutlier = true;
            }
            
            o.outlierCount = outlierCount;
            
            return isOutlier;
        }
	

	public static void main(String[] args) {
		
		// Split down for outlier
		// dir = "datasynWithCl.txt";
		// dir = "data/default/nba20072008.txt";
		// dir = "data/synthetic.niko";
		// dir = "data/synthetic.2Clu.Niko.clu";
  
                String filePath ="/Users/khaegler/NetBeans/outlier-oci/trunk/data/default/synthetic_2Clu_niko.txt";
//		String filePath = "data/synthetic/2Clu.data";
//		String filePath = "data/synthetic/1Clu.EPD.data";
//		String filePath = "data/synthetic/1Clu.data";
//		 String filePath = "data/default/syn1d.txt";
		IO ea = new IO(filePath, 2); // SPECIFY DIMENSIONALITY
		// IO ea = new IO(dir, 4); // SPECIFY DIMENSIONALITY
		DataObject[] data = ea.readFile(); // NOT CLASSIFIED DATA: for kddExample
		// DataObject[] data = ea.readClassFile(); //CLASSIFIED DATA: for sigmodExample
		DB db = new DB(data);

		int dim = data[0].coord.length;

		KNN knn = new KNN(20, filePath);
		knn.getAllNN(db.store, dim);

		System.out.println(db.store);
	}

}
